/* DS configuration.
 * include/config.h
 * Copyright (C) 2000-2001 DarkSiderz Net.  All rights reserved.
 * Read the file doc/contact for contact information related to this software. */

/*************************************************************************/

#ifndef CONFIG_H
#define CONFIG_H

/*************************************************************************/

/* TS compatibility modes. */
#define TS_VERSION    "TS"
#define TS_MIN         1
#define TS_CURRENT     3

/* What timezone is DS located in? */
#define TIMEZONE			"EST"
#define TIMEOUT_CHECK			3

/* What is the maximum number of characters in a configuration value and option?  In
 * addition what is the actual configuration file name? */
#define CONF_OPTION_MAX		 	64
#define CONF_VALUE_MAX			512
#define CONF_FILE			"ds.conf"

/* Insecure proxy server settings.  Which ports do we connect to the client for socks 4
 * & 5 as well as wingate? */
#define SOCKS5PORT			1080
#define SOCKS4PORT			1080	
#define WINGATEPORT			23

/* Where are the helpfiles located in the data directory?  Also what are the exclusive
 * language help files directory under the base help file directory. */
#define HELPSERV_DIR			"helpfiles"
#define LANG_ENGLISH_HS_DIR		"helpfiles/english"
#define LANG_FRENCH_HS_DIR		"helpfiles/french"
#define LANG_SPANISH_HS_DIR		"helpfiles/spanish"
#define LANG_ITALIAN_HS_DIR		"helpfiles/italian"
#define LANG_GERMAN_HS_DIR		"helpfiles/german"

/**************************************************************************/

/* This is the definitions for detecting clones on the network, clone sizing restrictions and limitations
 * for connecting users. */

/* How many identical hosts do we allow connected to the network before we consider them clones? */
#define CLONES_ALLOWED	4
/* Do we examine @host or user@host for clone identification? (1 for user@host, 0 for @host) */
#define CLONES_TYPE	0
/* What is the base interval of detecting clones from the base clones allowed? Meaning, assuming 5 clients connect
 * with same hosts or user@hosts we send globops, then this number is mutiplied by the base, assume 2, so the
 * next globops will occur at 10 clones from the same host, this number is prefixed to int count as well. */
#define CLONES_BASE_INT	2


/* Enable code to check for clones? */
#define CHECK_CLONES

/* How big a hostname list do we keep for clone detection?  On large nets
 * (over 500 simultaneous users or so), you may want to increase this if
 * you want a good chance of catching clones. */
#define CLONE_DETECT_SIZE 16

/* How many successive connects do we have to see before we consider the
 * possibility of clones? */
#define CLONE_MIN_USERS	5

/* How much time (in seconds) has to elapse between successive users
 * before we decide they're not clones? */
#define CLONE_MAX_DELAY	10

/* How long do we wait (in seconds) between successive warnings for
 * clones between the same host? */
#define CLONE_WARNING_DELAY 30

/* Size of input buffer (note: this is different from BUFSIZ) */
#define BUFSIZE		1024

/* Delay (in seconds) before we time out on a read and do other stuff,
 * like checking timeouts.  The busier your server is, the lower this
 * should be set (though the busier your server is, the less likely that
 * this will make any difference). */
#define READ_TIMEOUT	10

/* Maximum amount of data from the network to buffer (bytes). */
// #define NET_BUFSIZE	65536
#define NET_BUFSIZE 8192
//#define NET_BUFSIZE 4096
//#define NET_BUFSIZE 1024
//#define NET_BUFSIZE 262144

/* Maximum length of a channel name */
#define CHANMAX		64
/* Maximum length of a nickname */
#define NICKMAX		32
/* Maximum length of a password */
#define PASSMAX		32
/* Maximum length of a host. */
#define HOSTMAX 	256
/* Maximum length of an ident. */
#define IDENTMAX	16
/* Maximum length of user modes. */
#define MODEMAX		16
/* Maximum length for a server name. */
#define SERVERMAX	128
/* Maximum length for a description of a server. */
#define DESCMAX		128
/* Maximum length for a services message. */
#define MSGMAX		512
/* Maximum length for a database file. */
#define DBMAX		128

/**************************************************************************/

#if !HAVE_STRICMP && HAVE_STRCASECMP
# define stricmp strcasecmp
# define strnicmp strncasecmp
#endif

/**************************************************************************/

#endif

/**************************************************************************/

