/* Prototypes and external variable declarations.
 * include/extern.h
 * Copyright (C) 2000-2001 DarkSiderz Net.  All rights reserved.
 * Read the file doc/contact for contact information related to this software. */

/*************************************************************************/

#ifndef   __extern_include__
#define __extern_include__

/*************************************************************************/

/* Handy simplifiers for prototypes. */
#define E extern
#define V void
#define I int
#define FUNCTION(x) ()
#define VAR(x) ()

/*************************************************************************/

/* actions.c */
E void kill_user 						FUNCTION((const char *source, const char *user, const char *reason));
E void bad_password 					FUNCTION((User *u));

/*************************************************************************/

/* channels.c */
E void get_channel_stats 				FUNCTION((long *nrec, long *memuse));
E Channel *findchan 					FUNCTION((const char *chan));
E void chan_adduser 					FUNCTION((User *user, const char *chan));
E void chan_deluser 					FUNCTION((User *user, Channel *c));
E void do_cmode 						FUNCTION((const char *source, int ac, char **av));
E void do_topic 						FUNCTION((const char *source, int ac, char **av));
E int only_one_user 					FUNCTION((const char *chan));
E void free_all_chan 					FUNCTION((void));
E void chan_adduser2 					FUNCTION((User *user, Channel *c));
E Channel *chan_create 					FUNCTION((const char *chan));

/*************************************************************************/

/* users */
E void do_sjoin 						FUNCTION((char *source, int ac, char **av));

/*************************************************************************/

/* chanserv.c */
E char s_ChanServ[NICKMAX];
E void get_chanserv_stats 				FUNCTION((long *nrec, long *memuse));
E void chanserv 						FUNCTION((const char *source, char *buf));
E int ret_level 						FUNCTION((int level));
E void load_cs_dbase 					FUNCTION((void));
E void save_cs_dbase 					FUNCTION((void));
E void check_modes 					FUNCTION((const char *chan));
E int check_valid_op 					FUNCTION((User *user, const char *chan, int newchan));
E int check_should_op 					FUNCTION((User *user, const char *chan));
E int check_should_voice 				FUNCTION((User *user, const char *chan));
E int check_kick 						FUNCTION((User *user, const char *chan));
E check_access 						FUNCTION((User *user, ChannelInfo *ci, int what));
E void record_topic 					FUNCTION((const char *chan));
E void restore_topic 					FUNCTION((const char *chan));
E int check_topiclock 					FUNCTION((const char *chan));
E void expire_chans 					FUNCTION((void));
E void cs_remove_nick 					FUNCTION((const char *nick));
E ChannelInfo *cs_findchan 				FUNCTION((const char *chan));
E int is_chan_aop 					FUNCTION((const char *nick, ChannelInfo *ci));
E int is_chan_sop 					FUNCTION((const char *nick, ChannelInfo *ci));
E int is_chan_vop 					FUNCTION((const char *nick, ChannelInfo *ci));
#ifdef V_UNREAL
   E int check_should_halfop				FUNCTION((User *user, const char *chan));
   E int is_chan_hop 					FUNCTION((const char *nick, ChannelInfo *ci));
#endif
E int check_should_srop					FUNCTION((User *user, const char *chan));
E ChannelInfo *chanlists[];

/*************************************************************************/

/* clients.c */
E int client_alias 					FUNCTION((char *client, char *alias));
E void create_clients 					FUNCTION((void));
E void recover_client 					FUNCTION((char *client));
E int is_services_client 				FUNCTION((char *client));
E int is_client 						FUNCTION((char *client));

/*************************************************************************/

/* conf.c */
E int load_conf						FUNCTION((char *conf));
E void setup_var_from_conf 				FUNCTION((void));
E int rehash_conf;

/*************************************************************************/

/* crash.c */
E short crash_id;
E int parse_crash_id 					FUNCTION((char *sigstr, int signum, char *buf));
	
/*************************************************************************/

/* hash.c */
E Hash *do_hash 						FUNCTION((const char *source, const char *cmd, Hash *table));
E SHash *do_shash 					FUNCTION((const char *source, const char *cmd, SHash *table));

/*************************************************************************/

/* datafiles.c */
E void check_file_version 				FUNCTION((FILE *f, const char *filename));
E int get_file_version 					FUNCTION((FILE *f, const char *filename));
E int write_file_version 				FUNCTION((FILE *f, const char *filename));
E FILE *open_db 						FUNCTION((const char *service, const char *filename, const char *mode));
E void close_db 						FUNCTION((FILE *dbfile, const char *filename));

/*************************************************************************/

/* helpserv.c */
E char s_HelpServ[NICKMAX];
E void helpserv 						FUNCTION((const char *whoami, const char *source, char *buf));

/*************************************************************************/

/* main.c */
E int remote_port;
E char *server_name, *services_dir, *time_zone, *quitmsg;
E int update_timeout, debug, readonly, nofork, quitting, delayed_quit;
E char inbuf[BUFSIZE];
E int servsock, save_data, got_alarm, dump_chan;
E time_t start_time;
E void open_log 						FUNCTION((void));
E void close_log 						FUNCTION((void));
E void log							(const char *fmt, ...);
E void log_perror						(const char *fmt, ...);
E void fatal						(const char *fmt, ...);
E void fatal_perror					(const char *fmt, ...);
E void introduce_user 					FUNCTION((const char *user));
E void join_channel 					FUNCTION((char *chan));

/*************************************************************************/

/* memory.c */
E void *smalloc						FUNCTION((long size));
E void *scalloc 						FUNCTION((long elsize, long els));
E void *srealloc 						FUNCTION((void *oldptr, long newsize));
E char *sstrdup 						FUNCTION((const char *s));

/*************************************************************************/

/* memoserv.c */
E char s_MemoServ[NICKMAX];
E void get_memoserv_stats 				FUNCTION((long *nrec, long *memuse));
E void memoserv 						FUNCTION((const char *source, char *buf));
E void load_ms_dbase 					FUNCTION((void));
E void save_ms_dbase 					FUNCTION((void));
E void check_memos 					FUNCTION((const char *nick));
E void clear_memos 					FUNCTION((const char *nick));
E void clear_cmemos 					FUNCTION((const char *chan));
E void check_cmemos 					FUNCTION((User *u, const char *chan));
E void ms_send_register 				FUNCTION((const char *source, const char *nick));
E void expire_memos 					FUNCTION((void));

/*************************************************************************/

/* misc.c */
E char *sgets2 						FUNCTION((char *buf, long size, int sock));
E char *strscpy 						FUNCTION((char *d, const char *s, size_t len));
#if !HAVE_SNPRINTF
# if BAD_SNPRINTF
#  define vsnprintf my_vsnprintf
#  define snprintf my_snprintf
# endif
E int vsnprintf 						FUNCTION((char *buf, size_t size, const char *fmt, va_list args));
E int snprintf						(char *buf, size_t size, const char *fmt, ...);
#endif
#if !HAVE_STRICMP && !HAVE_STRCASECMP
E int stricmp 						FUNCTION((const char *s1, const char *s2));
E int strnicmp 						FUNCTION((const char *s1, const char *s2, size_t len));
#endif
#if !HAVE_STRDUP
E char *strdup 						FUNCTION((const char *s));
#endif
#if !HAVE_STRSPN
E size_t strspn 						FUNCTION((const char *s, const char *accept));
#endif
E char *stristr 						FUNCTION((char *s1, char *s2));
#if !HAVE_STRERROR
E char *strerror 						FUNCTION((int errnum));
#endif
#if !HAVE_STRSIGNAL
char *strsignal 						FUNCTION((int signum));
#endif
E char *merge_args 					FUNCTION((int argc, char **argv));
E int match_wild 						FUNCTION((const char *pattern, const char *str));
E int match_wild_nocase					FUNCTION((const char *pattern, const char *str));
E char *month_name 					FUNCTION((int month));
E char *strupper 						FUNCTION((char *s));
E char *strlower						FUNCTION((char *s));
E char *read_string 					FUNCTION((FILE *f, const char *filename));
E void write_string 					FUNCTION((const char *s, FILE *f, const char *filename));

/*************************************************************************/

/* nickserv.c */
E char s_NickServ[NICKMAX];
E NickInfo *nicklists[256];
E void get_nickserv_stats 				FUNCTION((long *nrec, long *memuse));
E void nickserv 						FUNCTION((const char *source, char *buf));
E void load_ns_dbase 					FUNCTION((void));
E void save_ns_dbase 					FUNCTION((void));
E int validate_user 					FUNCTION((User *u));
E void cancel_user 					FUNCTION((User *u));
E void expire_nicks 					FUNCTION((void));
E NickInfo *findnick 					FUNCTION((const char *nick));
E int cr_encrypt_string 				FUNCTION((NickInfo *ni, ChannelInfo *ci, char *string, int lo));
E NickInfo *nicklists[256];

/*************************************************************************/

/* operserv.c */
E char s_OperServ[NICKMAX];
E char s_GlobalNoticer[NICKMAX];
E char route_server[HOSTMAX];
E void operserv 						FUNCTION((const char *source, char *buf));
E void load_akill 					FUNCTION((void));
E void save_akill 					FUNCTION((void));
E int check_akill 					FUNCTION((const char *nick, const char *username, const char *host));
E void expire_akills 					FUNCTION((void));
E void check_clones 					FUNCTION((User *user));
E int is_services_root 					FUNCTION((const char *nick));
E int is_services_op 					FUNCTION((const char *nick));
E int is_services_admin 				FUNCTION((const char *nick));
E int is_services_root_su				FUNCTION((const char *nick));
E void os_load_admin_db 				FUNCTION((void));
E void os_save_admin_db 				FUNCTION((void));
E void os_load_route_db 				FUNCTION((void));
E void os_save_route_db 				FUNCTION((void));
E int deladmin 						FUNCTION((AdminInfo *ai));
E AdminInfo *findadmin 					FUNCTION((char *nick));
E int delroute 						FUNCTION((RouteInfo *ri));
E int delclone 						FUNCTION((CloneList *cl));
E CloneList *findclone 					FUNCTION((char *host));
E RouteInfo *findroute 					FUNCTION((char *server));
E void set_akill 						FUNCTION((void));
E int get_admin_level 					FUNCTION((const char *nick));
E int is_serv_admin 					FUNCTION((const char *nick, const int level));
E void os_list_stats_cn					FUNCTION((const char *source));
E void os_list_stats_k					FUNCTION((const char *source));
E void os_list_stats_o					FUNCTION((const char *source));
E void os_list_stats_h					FUNCTION((const char *source));

/*************************************************************************/

/* process.c */
E void m_version 						FUNCTION((char *source, int ac, char **av));
E int split_buf 						FUNCTION((char *buf, char ***argv, int colon_special));
E void process 						FUNCTION((void));

/*************************************************************************/

/* send.c */
E void send_cmd						(const char *source, const char *fmt, ...);
E void vsend_cmd 						FUNCTION((const char *source, const char *fmt, va_list args));
E void wallops						(const char *source, const char *fmt, ...);
E void globops						(const char *source, const char *fmt, ...);
E void notice						(const char *source, const char *dest, const char *fmt, ...);
E void notice_list 					FUNCTION((const char *source, const char *dest, const char **text));
E void privmsg						(const char *source, const char *dest, const char *fmt, ...);

/*************************************************************************/

/* server.c */
E int delserver 						FUNCTION((ServerInfo *si));
E ServerInfo *findserver 				FUNCTION((char *server));
E ServerInfo *add_server 				FUNCTION((const char *server));
E void send_map 						FUNCTION((const char *source, char *client));
E void send_pings 					FUNCTION((void));
E void free_all_server 					FUNCTION((void));
E void set_hub 						FUNCTION((void));

/*************************************************************************/

/* socks.c */
E int32 total_read, total_written;
E int32 read_buffer_len 				FUNCTION((void));
E int32 write_buffer_len 				FUNCTION((void));
E int sgetc 						FUNCTION((int s));
E char *sgets 						FUNCTION((char *buf, int len, int s));
E int sread 						FUNCTION((int s, char *buf, int len));
E int sputs 						FUNCTION((char *str, int s));
E int sockprintf						(int s, char *fmt,...);
E int conn 							FUNCTION((char *host, int port));
E void disconn 						FUNCTION((int s));
E void scan_proxy 					FUNCTION((User *u));

/*************************************************************************/

/* users.c */
E int lu_users, lu_opers, lu_invisible, lu_opers_inv, lu_max_users, lu_max_opers;
E void get_user_stats 					FUNCTION((long *nusers, long *memuse));
E User *finduser 						FUNCTION((const char *nick));
E void do_nick 						FUNCTION((const char *source, int ac, char **av));
E void do_join 						FUNCTION((const char *source, int ac, char **av));
E void do_part 						FUNCTION((const char *source, int ac, char **av));
E void do_kick 						FUNCTION((const char *source, int ac, char **av));
E void do_umode 						FUNCTION((const char *source, int ac, char **av));
E void do_quit 						FUNCTION((const char *source, int ac, char **av));
E void do_kill 						FUNCTION((const char *source, int ac, char **av));
E void do_svinfo 						FUNCTION((const char *source, int ac, char **av));
E int is_oper 						FUNCTION((const char *nick));
E int is_sadmin 						FUNCTION((const char *nick));
E int is_invisible					FUNCTION((const char *nick));
E int is_on_chan 						FUNCTION((const char *nick, const char *chan));
E int is_chanop 						FUNCTION((const char *nick, const char *chan));
E int is_voiced 						FUNCTION((const char *nick, const char *chan));
E int is_halfop						FUNCTION((const char *nick, const char *chan));
E int match_usermask 					FUNCTION((const char *mask, User *u));
E void split_usermask 					FUNCTION((const char *mask, char **nick, char **user, char **host));
E char *create_mask 					FUNCTION((User *u));
E void send_akill 					FUNCTION((const char *source, char *user, char *host, const char *reason, int amount, time_t expires));
E void change_user_nick 				FUNCTION((User *u, const char *nick));
E void send_find						FUNCTION((const char *source, int type));
E void free_all_user 					FUNCTION((void));
E void update_last_seen					FUNCTION((void));
E int count_clones 					FUNCTION((User *u));

/*************************************************************************/

/* language.c */
E char *rpl_numeric_lang 				FUNCTION((const char *source, int numeric));

/*************************************************************************/

#endif

/*************************************************************************/

