/* Clients functions and creation.
 * src/clients.c
 * Copyright (C) 2000-2001 DarkSiderz Net.  All rights reserved.
 * Read the file doc/contact.html for contact information related to 
 * this software. */

/*************************************************************************/

#ifndef CLIENTS_C
#define CLIENTS_C

/*************************************************************************/

#include "services.h"

/*************************************************************************/

typedef struct {
   int pos;
   char nick[NICKMAX];
   char ident[IDENTMAX];
   char modes[MODEMAX];
   char name[DESCMAX];
   char alias[NICKMAX];
   char chan[CHANMAX];
} Clients;
Clients clients[] = {
   { 0, "ChanS",     "service",     "+i", "Channel Registration Service", 	"ChanServ", "#dsnoop"         },
   { 1, "NickS",     "service",     "+i", "Nick Registration Service",    	"NickServ", "#dsnoop"         },
   { 2, "MemoS",     "service",     "+i", "Memo Service",    	"MemoServ", "#dsnoop"         },
   { 3, "HelpS",     "stats",     "+i", "Help Service",    	"HelpServ", "#dsnoop"         },
   { 4, "OperS",     "stats",     "+io-fwgsaA", "Oper access port",	"OperServ", "#dsnoop"         },
   { 5, "RootServ",   "stats",   "+io-fwgsaA", "Oper access port",    	"OperServ", "#dsnoop"         },
   { 6, "RootS",     "stats",     "+io-fwgsaA", "Oper access port",    	"OperServ", "#dsnoop"         },
   { 7, "InfoServ",   "stats",   "+i", "Info Service",        "HelpServ", "#dsnoop"         },
   { 8, "InfoS",     "stats",     "+i", "Info Service",        "HelpServ", "#dsnoop"         },
   { -1 }
};

/*************************************************************************/

/* create_clients()
 * Creates and initializes all clients in the client array. */

void create_clients(void)
{
   int i;

   for (i = 0; clients[i].pos >= 0; i++) {
      if (clients[i].nick) {
#ifdef V_BAHAMUT
	send_cmd(NULL, "NICK %s 1 %lu %s %s %s %s 0 :%s", clients[i].nick, time(NULL), "+", clients[i].ident, 
		"darksiderz.net", conf.lserver, clients[i].name);
#else
         send_cmd(NULL, "NICK %s 1 %lu %s %s %s 0 :%s", clients[i].nick, time(NULL),
	     clients[i].ident, conf.lserver, conf.lserver, clients[i].name);
#endif
         send_cmd(clients[i].nick, "MODE %s %s", clients[i].nick, clients[i].modes);
	   if (clients[i].chan) {
#ifdef V_BAHAMUT
	      send_cmd(NULL, ":%s SJOIN %ld %ld %s + %s", conf.lserver, time(NULL), time(NULL),
			clients[i].chan, clients[i].nick);
#else
	      send_cmd(NULL, ":%s JOIN %s", clients[i].nick, clients[i].chan);
#endif
	      send_cmd(clients[i].nick, "MODE %s +o %s", clients[i].chan, clients[i].nick);
	   }
      }
   }
}

/*************************************************************************/

/* recover_client()
 * Recovers a client after being killed from the client array. */

void recover_client(char *client)
{
   int i;

   for (i = 0; clients[i].pos >= 0; i++) {
      if (clients[i].nick) {
#ifdef V_BAHAMUT
	send_cmd(NULL, "NICK %s 1 %lu %s %s %s %s 0 :%s", clients[i].nick, time(NULL), "+", clients[i].ident, 
		conf.lserver, conf.lserver, clients[i].name);
#else
         send_cmd(NULL, "NICK %s 1 %lu %s %s %s 0 :%s", clients[i].nick, time(NULL),
	    clients[i].ident, conf.lserver, conf.lserver, clients[i].name);
#endif
            send_cmd(clients[i].nick, "MODE %s %s", clients[i].nick, clients[i].modes);
	   if (clients[i].chan) {
#ifdef V_BAHAMUT
	      send_cmd(NULL, ":%s SJOIN %ld %ld %s + %s", conf.lserver, time(NULL), time(NULL),
			clients[i].chan, clients[i].nick);
#else
	      send_cmd(NULL, ":%s JOIN %s", clients[i].nick, clients[i].chan);
#endif
	      send_cmd(clients[i].nick, "MODE %s +o %s", clients[i].chan, clients[i].nick);
	   }
      }
   }
}

/*************************************************************************/

/* client_alias()
 * Is this client an alias of a true services client? */

int client_alias(char *client, char *alias)
{
   int i;

   for (i = 0; clients[i].pos >= 0; i++) {
      if (clients[i].nick) {
         if ((stricmp(clients[i].nick, client) == 0) && (stricmp(clients[i].alias, alias) == 0))
            return C_TRUE;
      }
   }
   return C_FALSE;
}

/*************************************************************************/

/* is_services_client()
 * Is this client a true services client? */

int is_services_client(char *client)
{
   if (stricmp(client, s_OperServ) == 0 || stricmp(client, s_NickServ) == 0 ||
       stricmp(client, s_ChanServ) == 0 || stricmp(client, s_MemoServ) == 0 ||
       stricmp(client, s_HelpServ) == 0 || stricmp(client, s_GlobalNoticer) == 0) {
	 return C_TRUE;
   }
   return C_FALSE;
}

/*************************************************************************/

/* is_client()
 * Is this client aa client from the client array? */

int is_client(char *client)
{
   int i;

   for (i = 0; clients[i].pos >= 0; i++)
      if (clients[i].nick)
         if (stricmp(clients[i].nick, client) == 0)
            return C_TRUE;
   return C_FALSE;
}

/*************************************************************************/

#endif

/*************************************************************************/
