/* Crash control
 * src/crash.c
 * Copyright (C) 2000-2001 DarkSiderz Net.  All rights reserved.
 * Read the file doc/contact.html for contact information related to 
 * this software. */

/*************************************************************************/

#ifndef CRASH_C
#define CRASH_C

/*************************************************************************/

#include "services.h"

/*************************************************************************/

/* Crash identifier (id) */
short crash_id;

/* Please note, the numerical assigments in this structure are crucial to
 * the functionality of the crash system, you must increment 2,3,4,5 not
 * 10,20,24, same applies for parsing the function. This feature is a debuggung
 * feature which traces by numerical id where the program went wrong 
 * and displays the function it crashed in. */

typedef struct {
   int id;
   char *func;
} Crash;
Crash crash[] = {
   { 0,          "undefined" 				},
   { 1,          "save_ns_dbase()" 			},
   { 2,          "save_cs_dbase()" 			},
   { 3,          "save_ms_dbase()" 			},
   { 4,          "os_save_admin_db()" 			},
   { 5,          "os_save_route_db()" 			},
   { 6,          "save_akill()" 			},
   { 7,          "(ns) do_getpass()"			},
   { 8,          "(ns) do_setpass()"			},
   { 9,          "(ns) do_forbid()"			},
   { 10,         "(ns) do_freeze()"			},
   { 11,         "(ns) do_mark()"			},
   { 12,         "(ns) do_svsnick()"			},
   { 13,         "(ns) do_bot()"			},
   { 14,         "(ns) do_expiry()"			},
   { 15,         "(ns) do_help()"			},
   { 16,         "(ns) do_register()"			},
   { 17,         "(ns) do_identify()"			},
   { 18,         "(ns) do_drop()"			},
   { 19,         "(ns) do_access()"			},
   { 20,         "(ns) do_info()"			},
   { 21,         "(ns) do_list()"			},
   { 22,         "(ns) do_recover()"			},
   { 23,         "(ns) do_release()"			},
   { 24,         "(ns) do_ghost()"			},
   { 25,         "(ns) do_find()"			},
   { 26,         "(ns) do_status()"			},
   { 27,         "(ns) do_set()"			},
   { 28,         "(ns) do_set_email()"			},
   { 29,         "(ns) do_set_url()"			},
   { 30,         "(ns) do_set_kill()"			},
   { 31,         "(ns) do_set_ircop()"			},
   { 32,         "(ns) do_set_nomemos()"		},
   { 33,         "(ns) do_set_guest()"			},
   { 34,         "(ns) do_set_noops()"			},
   { 35,         "(ns) do_set_private()"		},
   { 36,         "(ns) do_set_secure()"			},
   { 37,         "(ns) do_set_password()"		},
   { 38,         "process()"				},
   { 39,         "send_pings()"				},
   { 40,         "expire_nicks()"			},
   { 41,         "expire_chans()"			},
   { 42,         "check_timeouts()"			},
   { 43,         "expire_akills()"			},
   { 44,         "set_akill()"				},
   { 45,         "introduce_user()"			},
   { 46,         "create_clients()"			},
   { 47,         "join_dump_channel()"			},
   { 48,         "(ns) del_ns_timeout()"		},
   { 49,         "(ns) add_ns_timeout()"		},
   { 50,         "(ns) timeout_release()"		},
   { 51,         "(ns) timeout_collide()"		},
   { 52,         "(ns) rem_ns_timeout()"		},
   { 53,         "(ns) release()"			},
   { 54,         "(ns) collide()"			},
   { 55,         "(ns) delnick()"			},
   { 56,         "(ns) makenick()"			},
   { 57,         "(ns) alpha_insert_nick()"		},
   { 58,         "(ns) is_on_access()"			},
   { 59,         "(ns) findnick()"			},
   { 60,         "(ns) cancel_user()"			},
   { 61,         "(ns) validate_user()"			},
   { 62,         "load_ns_dbase()"			},
   { 63,         "nickserv()"				},
   { 64,         "get_nickserv_stats()"			},
   { 65,         "get_memoserv_stats()"			},
   { 66,         "memoserv()"				},
   { 67,         "load_ms_dbase()"			},
   { 68,         "(ms) check_memos()"			},
   { 69,         "(ms) clear_memos()"			},
   { 70,         "(ms) find_memolist()"			},
   { 71,         "(ms) alpha_inset_memolist()"		},
   { 72,         "(ms) del_memolist()"			},
   { 73,         "(ms) do_help()"			},
   { 74,         "(ms) do_send()"			},
   { 75,         "(ms) do_list()"			},
   { 76,         "(ms) do_read()"			},
   { 77,         "(ms) do_del()"			},
   { 78,         "(ms) do_set()"			},
   { 79,         "(ms) do_set_notify()"			},
   { 80,         "(ms) do_email()"			},
   { 81,         "(ms) do_send_register()"		},
   { 82,         "(ms) do_opersend()"			},
   { 83,         "(ms) do_usersend()"			},
   { 84,         "(ms) do_adminsend()"			},
   /* Expanision, increment by 1, all the time. */
   { -1 }
};
    
/*************************************************************************/

/* parse_crash_id()
 * Parses the crash_id and return the globops for the designated id. */

int parse_crash_id(char *sigstr, char *buf)
{
   int i, j, crash_key = 0;

   /* This function scans for the current crash_id and evaluates it.  If the crash_id
    * matches an id in the crash array, globops where services went wrong, then globops
    * the actual signal services recieved from the application kernel. */

   for (i = 0; crash[i].id >= 0; i++)
      if (crash[i].id == crash_id)
         crash_key = i;

   if (sigstr && crash_id)
      globops(NULL, "DS encountered a fatal exception error in function \2%s\2, crash id \2%d\2, DS is now shutting down.",
         crash[crash_key].func, crash[crash_key].id);
   else
      globops(NULL, "DS encountered fatal buffer %s exception error, DS is now shutting down.", buf);

   return C_TRUE;
}

/*************************************************************************/

#endif

/*************************************************************************/