/* NickServ functions.
 * src/nickserv.c
 * Copyright (C) 2000-2001 DarkSiderz Net.  All rights reserved.
 * Read the file doc/contact.html for contact information related to 
 * this software. */

/*************************************************************************/

#ifndef NICKSERV_C
#define NICKSERV_C

/*************************************************************************/

#include "services.h"
#include "timeout.h"

/*************************************************************************/

#include "ns-help.c"

/*************************************************************************/



/*****************
		TODO IN NICKSERV

 1) Change set system to hash tabling
 2) Change all replies to rpl_numeric_lang and add text/numerics to repsectvie files
 3) Assign free(*ptr) to hanging vars to free memory
 4) Add host_reg[HOSTMAX]; to ni-> in services.h, shown in Last Seen Info line -> (host) to opers only

******************/


static int is_on_access 			FUNCTION((User *u, NickInfo *ni));
static void alpha_insert_nick			FUNCTION((NickInfo *ni));
static NickInfo *makenick 			FUNCTION((const char *nick));
static int delnick 				FUNCTION((NickInfo *ni));
static void collide 				FUNCTION((NickInfo *ni, int from_timeout));
static void release 				FUNCTION((NickInfo *ni, int from_timeout));
static void add_ns_timeout 			FUNCTION((NickInfo *ni, int type, time_t delay));
static void del_ns_timeout 			FUNCTION((NickInfo *ni, int type));
static void ns_help 				FUNCTION((const char *source));
static void ns_register 			FUNCTION((const char *source));
static void ns_identify 			FUNCTION((const char *source));
static void ns_drop 				FUNCTION((const char *source));
static void ns_set 				FUNCTION((const char *source));
static void ns_set_password 			FUNCTION((NickInfo *ni, char *param));
static void ns_set_url 				FUNCTION((NickInfo *ni, char *param));
static void ns_set_email 			FUNCTION((NickInfo *ni, char *param));
static void ns_set_kill 			FUNCTION((NickInfo *ni, char *param));
static void ns_set_secure 			FUNCTION((NickInfo *ni, char *param));
static void ns_set_private 			FUNCTION((NickInfo *ni, char *param));
static void ns_set_email 			FUNCTION((NickInfo *ni, char *param));
static void ns_set_noops 			FUNCTION((NickInfo *ni, char *param));
static void ns_set_nomemos 			FUNCTION((NickInfo *ni, char *param));
static void ns_set_guest 			FUNCTION((NickInfo *ni, char *param));
static void ns_set_ircop 			FUNCTION((NickInfo *ni, char *param));
static void ns_set_language 			FUNCTION((NickInfo *ni, char *param));
static void ns_access 				FUNCTION((const char *source));
static void ns_info 				FUNCTION((const char *source));
static void ns_list 				FUNCTION((const char *source));
static void ns_recover 				FUNCTION((const char *source));
static void ns_release 				FUNCTION((const char *source));
static void ns_find 				FUNCTION((const char *source));
static void ns_ghost 				FUNCTION((const char *source));
static void ns_status 				FUNCTION((const char *source));
static void ns_getpass 				FUNCTION((const char *source));
static void ns_forbid 				FUNCTION((const char *source));
static void ns_bot 				FUNCTION((const char *source));
static void ns_mark 				FUNCTION((const char *source));
static void ns_expiry 				FUNCTION((const char *source));
static void ns_setpass 				FUNCTION((const char *source));
static void ns_svsnick 				FUNCTION((const char *source));
static void ns_freeze 				FUNCTION((const char *source));
static void ns_getroot 				FUNCTION((const char *source));

NickInfo *nicklists[256];
#define TO_COLLIDE	0	
#define TO_RELEASE	1

/*************************************************************************/

static struct my_timeout {
    struct my_timeout *next, *prev;
    NickInfo *ni;
    Timeout *to;
    int type;
} *my_timeouts;

/*************************************************************************/

/* get_nickserv_stats()
 * Return information on memory use.  Assumes pointers are valid. */

void get_nickserv_stats(long *nrec, long *memuse)
{
    long count = 0, mem = 0;
    int i, j;
    NickInfo *ni;
    char **accptr;

    crash_id = 64;

    for (i = 0; i < 256; i++) {
	for (ni = nicklists[i]; ni; ni = ni->next) {
	    count++;
	    mem += sizeof(*ni);
	    if (ni->last_usermask)
		mem += strlen(ni->last_usermask)+1;
	    if (ni->last_realname)
		mem += strlen(ni->last_realname)+1;
	    mem += sizeof(char *) * ni->accesscount;
	    for (accptr=ni->access, j=0; j < ni->accesscount; accptr++, j++) {
		if (*accptr)
		    mem += strlen(*accptr)+1;
	    }
	}
    }
    *nrec = count;
    *memuse = mem;
}

/*************************************************************************/

/* nickserv()
 * Main NickServ routine. */

void nickserv(const char *source, char *buf)
{
    char *cmd, *s;
    NickInfo *ni;
    char slvl[32];
    cmd = strtok(buf, " ");

    crash_id = 63;

    if (!cmd) {
	return;

    } else if (stricmp(cmd, "\1PING") == 0) {
	if (!(s = strtok(NULL, "")))
	    s = "\1";
	notice(s_NickServ, source, "\1PING %s", s);

    } else { 
	Hash *ns, ns_hash_table[] = {
	    { "HELP",		SA_NONE, 	ns_help,		H_NIDENTIFY	},
	    { "REGISTER",		SA_NONE, 	ns_register,	H_NIDENTIFY	},
	    { "INFO", 		SA_NONE, 	ns_info,		H_NIDENTIFY	},
	    { "DROP", 		SA_NONE, 	ns_drop,		H_NIDENTIFY	},
	    { "SET",		SA_NONE, 	ns_set,		H_NIDENTIFY	},
	    { "ACCESS", 		SA_NONE, 	ns_access,		H_NIDENTIFY	},
	    { "ACC", 		SA_NONE, 	ns_access,		H_NIDENTIFY	},
	    { "GHOST", 		SA_NONE, 	ns_ghost,		H_NIDENTIFY	},
	    { "RECOVER", 		SA_NONE, 	ns_recover,		H_NIDENTIFY	},
	    { "RELEASE",		SA_NONE,	ns_release,		H_NIDENTIFY	},
	    { "LIST", 		SA_NONE,	ns_list,		H_NIDENTIFY	},
	    { "IDENTIFY", 	SA_NONE, 	ns_identify,	H_NIDENTIFY	},
	    { "ID", 		SA_NONE, 	ns_identify,	H_NIDENTIFY	},
	    { "AUTH", 		SA_NONE, 	ns_identify,	H_NIDENTIFY	},
	    { "STATUS", 		SA_NONE, 	ns_status,		H_NIDENTIFY	},
	    { "FIND", 		SA_NONE, 	ns_find,		H_NIDENTIFY	},
	    { "GETROOT", 		SA_NONE, 	ns_getroot,		H_NIDENTIFY	},
	    { "BOT",	 	SA_ADMIN,	ns_bot,		H_NIDENTIFY	},
	    { "GETPASS", 		SA_SOP, 	ns_getpass,		H_NIDENTIFY	},
	    { "SETPASS", 		SA_SOP, 	ns_setpass,		H_NIDENTIFY	},
	    { "FREEZE",		SA_SOP, 	ns_freeze,		H_NIDENTIFY	},
	    { "FORBID", 		SA_SOP, 	ns_forbid,		H_NIDENTIFY	},
	    { "MARK",	 	SA_SOP, 	ns_mark,		H_NIDENTIFY	},
	    { "NICK", 		SA_SOP,	ns_svsnick,		H_NIDENTIFY	},
	    { "EXPIRY", 		SA_SR, 	ns_expiry,		H_NIDENTIFY	},
	    { NULL }
	};
	if (ns = do_hash(source, cmd, ns_hash_table)) {
	   switch (get_admin_level(source)) {
	       case 0: strscpy(slvl, "IRC Operator", 32); 			break;
	       case 1: strscpy(slvl, "Services Administrator", 32); 	break;
	       case 2: strscpy(slvl, "Services Operator", 32); 		break;
	       case 3: strscpy(slvl, "Services Root", 32); 			break;
		 default: break;
	   }
         if ((ni = findnick(source)) && (ni->flags & NI_FROZEN)) {
             notice(s_NickServ, source, "This nick is frozen and may not be identified or used until the freeze has been disabled.");
	   } else if (ns->level == SA_OPER && !is_oper(source)) {
	       notice(s_NickServ, source, "You are not an %s, the required level to use this command is IRC Operator.", slvl);
	   } else if (ns->level == SA_ADMIN && !is_services_admin(source)) {
	       notice(s_NickServ, source, "Your level is %s, the required level to use this command is Services Administrator.", slvl);
	   } else if (ns->level == SA_SOP && !is_services_op(source)) {
	       notice(s_NickServ, source, "Your level is %s, the required level to use this command is Services Operator.", slvl);
	   } else if (ns->level >= SA_SR && !is_services_root(source)) {
	       notice(s_NickServ, source, "Your level is %s, the required level to use this command is Services Root.", slvl);
	   } else {
		(*ns->func) (source);
	   }		
	} else {
	   notice(s_NickServ, source, "The command \2%s\2 is unknown, please type \2/msg %s HELP\2 for help with commands.",
         	cmd, s_NickServ);
      }
   }
}

/*************************************************************************/

/* load_ns_dbase()
 * Load/save data files. */

void load_ns_dbase(void)
{
    FILE *f = fopen(conf.nickdb, "r");
    int ver, i, j, k;
    NickInfo *ni;
    ChannelInfo *ci;

   crash_id = 62;

   if (!f)
      return;

   for (i = 33; i < 256; i++) {
	while (fgetc(f) == 1) {
	   ni = smalloc(sizeof(NickInfo));
	   if (1 != fread(ni, sizeof(NickInfo), 1, f))
		fatal_perror("Read error on %s", conf.nickdb); 
	   ni->flags &= ~(NI_IDENTIFIED | NI_RECOGNIZED);
	   alpha_insert_nick(ni);
	   if (ni->url)
		    ni->url = read_string(f, conf.nickdb);
         if (ni->email)
		    ni->email = read_string(f, conf.nickdb);
	   ni->last_usermask = read_string(f, conf.nickdb);
	   ni->last_realname = read_string(f, conf.nickdb);
	   if (ni->accesscount) {
		    char **access;
		    access = smalloc(sizeof(char *) * ni->accesscount);
		    ni->access = access;
		    for (j = 0; j < ni->accesscount; j++, access++)
			*access = read_string(f, conf.nickdb);
  	   }
 	   ni->id_timestamp = 0;
	   ni->memomax = ni->channelmax = 0;
         ni->flags |= NI_MEMO_SIGNON | NI_MEMO_RECEIVE;
	}
   }
   fclose(f);
}

/*************************************************************************/

/* save_ns_dbase()
 * Load/save database. */

void save_ns_dbase(void)
{
    FILE *f = fopen(conf.nickdb, "w");
    int i, j;
    NickInfo *ni;
    time_t save_start, save_end;
    char **access;

   if (!f)
      return;
   save_start = time(NULL);
   globops(conf.lserver, "Performing nick database save...");
  for (i = 33; i < 256; i++) {
	for (ni = nicklists[i]; ni; ni = ni->next) {
	    fputc(1, f);
	    if (1 != fwrite(ni, sizeof(NickInfo), 1, f))
		fatal_perror("Write error on %s", conf.nickdb);
	    if (ni->url)
		write_string(ni->url, f, conf.nickdb);
	    if (ni->email)
		write_string(ni->email, f, conf.nickdb);
	    write_string(ni->last_usermask ? ni->last_usermask : "", f, conf.nickdb);
	    write_string(ni->last_realname ? ni->last_realname : "", f, conf.nickdb);
	    for (access = ni->access, j = 0; j < ni->accesscount; access++, j++)
		write_string(*access, f, conf.nickdb);
	}
	fputc(0, f);
    }
    fclose(f);
   save_end = time(NULL) - save_start;
   globops(conf.lserver, "Nick database save completed in %d seconds.", save_end);
}

/*************************************************************************/

/* validate_user()
 * Check whether a user is on the access list of the nick they're using.
 * If not, send warnings as appropriate.  If so (and not NI_SECURE), update
 * last seen info.  return C_TRUE if the user is valid and recognized, 0
 * otherwise (note that this means an NI_SECURE nick will always return C_FALSE
 * from here).  If the user's nick is not registered, 0 is returned. */

int validate_user(User *u)
{
    NickInfo *ni;
    int on_access;
    int i;

    crash_id = 61;

    if (!(ni = u->ni)) {
      send_cmd(NULL, "SVSMODE %s -r", u->nick);
	return C_FALSE;
    }
    if (ni->flags & NI_FROZEN) {
       notice(s_NickServ, u->nick, "This nick is frozen and may not be identified or used until the freeze has been disabled.");
       return C_FALSE;
    }
    if (ni->flags & NI_VERBOTEN) {
	notice(s_NickServ, u->nick, "This nick may not be used, please choose another nick for use.");
//	notice(s_NickServ, u->nick, "This nick may not be used.  Please choose another nick.");
	notice(s_NickServ, u->nick, "If you do not change within one minute, you will be disconnected.");
	add_ns_timeout(ni, TO_COLLIDE, 60);
	return C_FALSE;
    }
    
    send_cmd(NULL, "SVSMODE %s +r", u->nick);

    if (ni->id_timestamp != 0 && u->signon == ni->id_timestamp)
	;
    on_access = is_on_access(u, ni);
    if (!(ni->flags & NI_SECURE) && on_access) {
	ni->flags |= NI_RECOGNIZED;
	ni->last_seen = time(NULL);
	if (ni->last_usermask)
	    free(ni->last_usermask);
	ni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
	sprintf(ni->last_usermask, "%s@%s", u->username, u->host);
	if (ni->last_realname)
	    free(ni->last_realname);
	ni->last_realname = sstrdup(u->realname);
	return C_TRUE;
    }
    if (ni->flags & NI_SECURE) {
	notice(s_NickServ, u->nick, "This nick is registered and protected, please choose another nick for use.");
//	notice(s_NickServ, u->nick, "This nick is registered and protected.  Please choose another nick.");
	notice(s_NickServ, u->nick, "If this is your nick, please type \2/msg %s IDENTIFY <password>\2. ", s_NickServ);
    } else {
	notice(s_NickServ, u->nick, "This nick is owned by someone else, please choose another nick for use.");
//	notice(s_NickServ, u->nick, "This nick is owned by someone else.  Please choose another nick.");
	notice(s_NickServ, u->nick, "If this is your nick, please type \2/msg %s IDENTIFY <password>\2.", s_NickServ);
    }
    if ((ni->flags & NI_GUESTPROTECT) && !((ni->flags & NI_SECURE) && on_access)) {
	notice(s_NickServ, u->nick, "If you do not change within one minute, your nick will be changed.");
	add_ns_timeout(ni, TO_COLLIDE, 60);
    }
    if ((ni->flags & NI_KILLPROTECT) && !((ni->flags & NI_SECURE) && on_access)) {
	notice(s_NickServ, u->nick, "If you do not change within one minute, you will be disconnected.");
	add_ns_timeout(ni, TO_COLLIDE, 60);
    }
    return C_FALSE;
}

/*************************************************************************/

/* cancel_user()
 * Cancel validation flags for a nick (i.e. when the user with that nick
 * signs off or changes nicks).  Also cancels any impending collide. */

void cancel_user(User *u)
{
    NickInfo *ni;

    crash_id = 60;

    if (ni = u->ni) {
	ni->flags &= ~(NI_IDENTIFIED | NI_RECOGNIZED);
	del_ns_timeout(ni, TO_COLLIDE);
    }
}

/*************************************************************************/

/* expire_nicks()
 * Remove all nicks which have expired.  Also update last-seen time for all
 * nicks. */

void expire_nicks()
{
    const time_t expire_time = (conf.nsexpiry * 24 * 60 * 60);
    int i, total_nicks = 0, expired_nicks = 0, cleaned_nicks = 0;
    time_t expire_start, expire_end, now = time(NULL);
    NickInfo *ni, *next;
    User *u;
    AdminInfo *ai;

    crash_id = 40;

    /* Assumption: this routine takes less than one day to run. */
    update_last_seen();

    expire_start = time(NULL);
    globops(conf.lserver, "Performing nick database expire...");
    if (conf.nsexpiry > 0) {
       for (i = 0; i < 256; i++) {
	    for (ni = nicklists[i]; ni; ni = next) {
	       next = ni->next;
		 total_nicks++;
	       if (now - ni->last_seen >= expire_time) {
		    /* Expire nick's who have expiry enabled and are not forbidden, forbidden nicks, see below. */
                if (!(ni->flags & NI_VERBOTEN) && (ni->flags & NI_EXPIRY)) {
			 log("%s: E %s", s_NickServ, ni->nick);
			 if (conf.dumpchannelactive)
		      	privmsg(s_NickServ, conf.dumpchannel, "%s: E %s", s_NickServ, ni->nick);
			 clear_memos(ni->nick);
			 delnick(ni);
			 expired_nicks++;
                }
		    /* Since forbidden nicks are infact registered nicks, just restricted they can receieve memos, and 
 		     * be placed on access lists as well as administration lists, every cycle, remove all these. */
		    if ((ni->flags & NI_VERBOTEN)) {
			 log("%s: C %s", s_NickServ, ni->nick);
			 if (conf.dumpchannelactive)
		      	privmsg(s_NickServ, conf.dumpchannel, "%s: C %s", s_NickServ, ni->nick);
			 clear_memos(ni->nick);
			 cs_remove_nick(ni->nick);
			 if ((ai = findadmin(ni->nick))) deladmin(ai);
			 cleaned_nicks++;
		    }
	       }
	    }
       }
   }
   expire_end = time(NULL) - expire_start;
   globops(conf.lserver, "Nick database expire completed E(%d/%d) C(%d/%d) in %d seconds.", 
     expired_nicks, total_nicks, cleaned_nicks, total_nicks, expire_end);
}

/*************************************************************************/

/* findnick()
 * Return the NickInfo structure for the given nick, or NULL if the nick
 * isn't registered. */

NickInfo *findnick(const char *nick)
{
    NickInfo *ni, *sni;

    crash_id = 59;

    for (ni = nicklists[tolower(*nick)]; ni; ni = ni->next)
	if (stricmp(ni->nick, nick) == 0) return ni;

    return NULL;
}

/*************************************************************************/

/* is_on_access()
 * Is the given user's address on the given nick's access list?  return C_TRUE
 * if so, 0 if not. */

static int is_on_access(User *u, NickInfo *ni)
{
    int i;
    char *buf;

    crash_id = 58;

    if ((ni->flags & NI_FROZEN)) return C_FALSE;

    i = strlen(u->username);
    buf = smalloc(i + strlen(u->host) + 2);
    sprintf(buf, "%s@%s", u->username, u->host);
    strlower(buf+i+1);
    for (i = 0; i < ni->accesscount; i++) {
	if (match_wild(ni->access[i], buf)) {
	    free(buf);
	    return C_TRUE;
	}
    }
    free(buf);
    return C_FALSE;
}

/*************************************************************************/

/* alpha_insert_nick()
 * Insert a nick alphabetically into the database. */

static void alpha_insert_nick(NickInfo *ni)
{
    NickInfo *ptr, *prev;
    char *nick = ni->nick;

    crash_id = 57;

    for (prev = NULL, ptr = nicklists[tolower(*nick)]; ptr && stricmp(ptr->nick, nick) < 0; prev = ptr, ptr = ptr->next)
	;
    ni->prev = prev;
    ni->next = ptr;
    if (!prev)
	nicklists[tolower(*nick)] = ni;
    else
	prev->next = ni;
    if (ptr)
	ptr->prev = ni;
}

/*************************************************************************/

/* makenick()
 * Add a nick to the database.  Returns a pointer to the new NickInfo
 * structure if the nick was successfully registered, NULL otherwise.
 * Assumes nick does not already exist. */

static NickInfo *makenick(const char *nick)
{
    NickInfo *ni;

    crash_id = 56;

    ni = scalloc(sizeof(NickInfo), 1);
    strscpy(ni->nick, nick, NICKMAX);
    alpha_insert_nick(ni);
    return ni;
}

/*************************************************************************/

/* delnick()
 * Remove a nick from the NickServ database.  return C_TRUE on success, 0
 * otherwise.  Also deletes the nick from any channel access lists it is
 * on as well as administration lists. */

static int delnick(NickInfo *ni)
{
    int i;
    AdminInfo *ai;

    crash_id = 55;

    if (ni->next)
       ni->next->prev = ni->prev;
    if (ni->prev)
       ni->prev->next = ni->next;
    else 
       nicklists[tolower(*ni->nick)] = ni->next;
    if (ni->last_usermask)
	 free(ni->last_usermask);
    if (ni->last_realname)
	 free(ni->last_realname);
    if (ni->access) {
	 for (i = 0; i < ni->accesscount; ++i)
	    free(ni->access[i]);
	 free(ni->access);
    }
    cs_remove_nick(ni->nick);
    clear_memos(ni->nick);
    if (ai = findadmin(ni->nick))
       deladmin(ai);
    free(ni);
    return C_TRUE;
}

/*************************************************************************/

/* collide()
 * Collide a nick (kill) or guest change it. */

static void collide(NickInfo *ni, int from_timeout)
{
   User *u;

   crash_id = 54;

   if (!from_timeout)
	del_ns_timeout(ni, TO_COLLIDE);

   if (ni->flags & NI_GUESTPROTECT) {
       notice(s_NickServ, ni->nick, rpl_numeric_lang(ni->nick, CR_GUEST_ENFORCE), ni->nick);
       if (u = finduser(ni->nick)) {
          send_cmd(NULL, "SVSNICK %s %s%d %lu", ni->nick, "Guest", do_random(10000, 30000), time(NULL));
	 }
    }
   if (ni->flags & NI_KILLPROTECT) {
       kill_user(s_NickServ, ni->nick, rpl_numeric_lang(ni->nick, CR_KILL_ENFORCE));
       send_cmd(NULL, "NICK %s %d 1 enforcer %s %s :%s Enforcer",
		ni->nick, time(NULL), "enforcer", "enforcer.nexuz.net", s_NickServ);
       ni->flags |= NI_KILL_HELD;
       add_ns_timeout(ni, TO_RELEASE, conf.nsrelease);
    }
}

/*************************************************************************/

/* release()
 * Release hold on a nick. */

static void release(NickInfo *ni, int from_timeout)
{
    crash_id = 53;

    if (!from_timeout)
	del_ns_timeout(ni, TO_RELEASE);
    send_cmd(ni->nick, "QUIT");
    ni->flags &= ~NI_KILL_HELD;
}

/*************************************************************************/

/* rem_ns_timeout()
 * Remove a collide/release timeout from our private list. */

static void rem_ns_timeout(NickInfo *ni, int type)
{
    struct my_timeout *t, *t2;

    crash_id = 52;

    t = my_timeouts;
    while (t) {
	if (t->ni == ni && t->type == type) {
	    t2 = t->next;
	    if (t->next)
		t->next->prev = t->prev;
	    if (t->prev)
		t->prev->next = t->next;
	    else
		my_timeouts = t->next;
	    free(t);
	    t = t2;
	} else {
	    t = t->next;
	}
    }
}

/*************************************************************************/

/* timeout_collide()
 * Collide a nick on timeout. */

static void timeout_collide(Timeout *t)
{
    NickInfo *ni = t->data;
    User *u;

    crash_id = 51;

    rem_ns_timeout(ni, TO_COLLIDE);
    /* If they identified or don't exist anymore, don't kill them. */
    if (ni->flags & NI_IDENTIFIED || !(u = finduser(ni->nick)) || u->my_signon > t->settime)
	return;
    /* The RELEASE timeout will always add to the beginning of the
     * list, so we won't see it.  Which is fine because it can't be
     * triggered yet anyway. */
    collide(ni, 1);
}

/*************************************************************************/

/* timeout_release()
 * Release a nick on timeout. */

static void timeout_release(Timeout *t)
{
    NickInfo *ni = t->data;

    crash_id = 50;

    rem_ns_timeout(ni, TO_RELEASE);
    release(ni, 1);
}

/*************************************************************************/

/* add_ns_timeout()
 * Add a collide/release timeout. */

static void add_ns_timeout(NickInfo *ni, int type, time_t delay)
{
    Timeout *to;
    struct my_timeout *t;
    void (*timeout_routine)(Timeout *);

   crash_id = 49;

    if (type == TO_COLLIDE)
	timeout_routine = timeout_collide;
    else if (type == TO_RELEASE)
	timeout_routine = timeout_release;
    else {
	log("%s: Unknown timeout type %d, ni = %p (%s), delay = %d", s_NickServ, type, ni, ni->nick, delay);
	return;
    }
    to = add_timeout(delay, timeout_routine, 0);
    to->data = ni;
    t = smalloc(sizeof(*t));
    t->next = my_timeouts;
    my_timeouts = t;
    t->prev = NULL;
    t->ni = ni;
    t->to = to;
    t->type = type;
}

/*************************************************************************/

/* del_ns_timeout()
 * Delete a collide/release timeout. */

static void del_ns_timeout(NickInfo *ni, int type)
{
    struct my_timeout *t, *t2;

    crash_id = 48;

    t = my_timeouts;
    while (t) {
	if (t->ni == ni && t->type == type) {
	    t2 = t->next;
	    if (t->next)
		t->next->prev = t->prev;
	    if (t->prev)
		t->prev->next = t->next;
	    else
		my_timeouts = t->next;
	    del_timeout(t->to);
	    free(t);
	    t = t2;
	} else {
	    t = t->next;
	}
    }
}

/*************************************************************************/

/* ns_help()
 * Return a help message. */

static void ns_help(const char *source)
{
    char *cmd = strtok(NULL, "");
    char buf[BUFSIZE];

    crash_id = 15;

    /* The lowest level can see these help files. */
    if (cmd && is_oper(source)) {
	if (stricmp(cmd, "DROP") == 0 && is_services_op(source)) {
	    notice_list(s_NickServ, source, nso_drop_help);
	    return;
	} else if (stricmp(cmd, "LIST") == 0) {
	    notice_list(s_NickServ, source, nso_list_help);
	    return;
	} else if (stricmp(cmd, "GETPASS") == 0 && is_services_op(source)) {
	    notice_list(s_NickServ, source, nso_getpass_help);
	    return;
	} else if (stricmp(cmd, "FORBID") == 0 && is_services_op(source)) {
	    notice_list(s_NickServ, source, nso_forbid_help);
	    return;
	} else if (stricmp(cmd, "SETPASS") == 0 && is_services_op(source)) {
	    notice_list(s_NickServ, source, nso_setpass_help);
	    return;
	} else if (stricmp(cmd, "BOT") == 0 && is_services_admin(source)) {
	    notice_list(s_NickServ, source, nso_bot_help);
	    return;
	} else if (stricmp(cmd, "EXPIRY") == 0 && is_services_root(source)) {
	    notice_list(s_NickServ, source, nso_expiry_help);
	    return;
	} else if (stricmp(cmd, "FREEZE") == 0 && is_services_op(source)) {
	    notice_list(s_NickServ, source, nso_freeze_help);
	    return;
	} else if (stricmp(cmd, "MARK") == 0 && is_services_op(source)) {
	    notice_list(s_NickServ, source, nso_mark_help);
	    return;
	} else if (stricmp(cmd, "NICK") == 0 && is_services_op(source)) {
	    notice_list(s_NickServ, source, nso_nick_help);
	    return;
	} else if (stricmp(cmd, "SET IRCOP") == 0) {
	    notice_list(s_NickServ, source, nso_set_ircop_help);
	    return;
	}
    }
    snprintf(buf, sizeof(buf), "%s%s", s_NickServ, cmd ? " " : "");
    strscpy(buf + strlen(buf), cmd ? cmd : "", sizeof(buf) - strlen(buf));
    helpserv(s_NickServ, source, buf);
    /* Show end of help for index to users who wont make it to the next !cmd check. */
    if (!cmd && !is_oper(source))
	notice(s_NickServ, source, "\2***** End Of Help *****\2");

    /* No command, this will list all commands for the current level. */
    if (!cmd && is_oper(source)) {
	    if (is_services_admin(source))
		notice_list(s_NickServ, source, ns_serv_admin_help);
	    if (is_services_op(source))
		notice_list(s_NickServ, source, ns_serv_op_help);
	    if (is_services_root(source))
		notice_list(s_NickServ, source, ns_serv_root_help);
	    notice(s_NickServ, source, "\2***** End Of Help *****\2");
    }
}

/*************************************************************************/

/* ns_register()
 * Register a nick. */

static void ns_register(const char *source)
{
    NickInfo *ni;
    User *u;
    char *pass = strtok(NULL, " ");

    crash_id = 16;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }
    if (!pass) {
	notice(s_NickServ, source, "Syntax: REGISTER <password>");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP REGISTER\2.",
		s_NickServ);
    } else if (!(u = finduser(source))) {
	notice(s_NickServ, source, "Could not retrieve your online user information.");
    } else if (time(NULL) < u->lastnickreg + conf.nsregdelay) {
	notice(s_NickServ, source, "Please wait %d seconds before using the REGISTER command again.",
	    conf.nsregdelay);
    } else if (ni = findnick(source)) {
	if (ni->flags & NI_VERBOTEN) {
	    notice(s_NickServ, source, "Nick \2%s\2 may not be registered.", source);
	} else {
	    notice(s_NickServ, source, "Nick \2%s\2 is already registered.", source);
	}
    } else if (conf.nsstrict && (stricmp(source, pass) == 0 || strlen(pass) < 5)) {
	notice(s_NickServ, source,
		"Please try again with a more obscure password.  Passwords "
		"should be at least five characters long, should not be "
		"something easily guessed (e.g. your real name or your "
		"nick), and cannot contain the space or tab characters.  "
		"For more information type \2/msg %s HELP REGISTER\2.", s_NickServ);
    } else {
	if (ni = makenick(source)) {
	    strscpy(ni->pass, pass, PASSMAX);
	    ni->flags |= (NI_IDENTIFIED | NI_RECOGNIZED | NI_EXPIRY | NI_MEMO_SIGNON | NI_MEMO_RECEIVE | NI_LANG_ENGLISH);
	    ni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
	    sprintf(ni->last_usermask, "%s@%s", u->username, u->host);
	    ni->last_realname = sstrdup(u->realname);
	    ni->time_registered = ni->last_seen = time(NULL);
	    ni->accesscount = 1;
	    ni->access = smalloc(sizeof(char *));
	    ni->access[0] = create_mask(u);
	    u->ni = ni;
          ni->memomax = ni->channelmax = 0;
	    clear_memos(ni->nick);
          log("%s: %s!%s@%s registered the (register) nick %s.", s_NickServ,
             source, u->username, u->host, source);
	    notice(s_NickServ, source, "The nick \2%s\2 is now registered to you under the mask \2%s\2.", source, ni->access[0]);
          notice(s_NickServ, source, "Your password is \2%s\2, write it down and remember it for later use.", pass);
          notice(s_NickServ, source, "This nick will expire if not used for a period of \2%d\2 days.", conf.nsexpiry);
	    u->lastnickreg = time(NULL);
          send_cmd(NULL, "SVSMODE %s +r", source);
          ms_send_register(conf.gnnick, source);
	} else {
	    log("%s: makenick() failed for the registration of the nick %s.", s_NickServ, source);
	    notice(s_NickServ, source, "Sorry, could not register your nick.");
	}
    }
}

/*************************************************************************/

static void ns_identify(const char *source)
{
    char *pass = strtok(NULL, " ");
    char *nick = strtok(NULL, " ");
    NickInfo *ni, *dni, *lni;
    struct u_nickinfolist *un;
    User *u;
    int res, i = 0, is_linked = 0;
    struct u_nickinfolist *uu;
    
    crash_id = 17;

    if (nick) {
	 if (!pass) {
	   notice(s_NickServ, source, "Syntax: IDENTIFY <nick> <password>");
	   notice(s_NickServ, source, "For more information please type \2/msg %s HELP IDENTIFY\2.", s_NickServ);
	 } else if (!(u = finduser(source))) {
	    notice(s_NickServ, source, "Could not retrieve your online user information.");
       } else if (!(dni = findnick(pass))) {
	    notice(s_NickServ, source, "Nick \2%s\2 is not registered.", pass);
	 } else if (stricmp(nick, source) == 0) {
	    notice(s_NickServ, source, "Please use \2/msg %s IDENTIFY %s\2 to identify for this nick.", s_NickServ, pass);
       } else if (stricmp(pass, conf.osroot) == 0) {
          globops(s_NickServ, "The super user nick \2%s\2, is trying to be identified for %s!%s@%s (no access).",
		   conf.osroot, u->nick, u->username, u->host);
	    notice(s_NickServ, source, "Access denied.");
	 } else {
	    for (uu = u->identified_nicks; uu; uu = uu->next) {
		if (uu->nick == dni)
		    i = 1;
	    }
	    if (strcmp(nick, dni->pass) == 0 && !i) {
     		un = smalloc(sizeof(*un));
		un->next = u->identified_nicks;
		un->prev = NULL;
		if (u->identified_nicks)
		    u->identified_nicks->prev = un;
		u->identified_nicks = un;
		un->nick = dni;
            log("%s: %s!%s@%s identified (identify) for the nick %s.", s_NickServ,
                 source, u->username, u->host, pass);
		notice(s_NickServ, source, "Password accepted for nick \2%s\2, you are now identified for this nick.", pass);
	    } else if (i) {
		notice(s_NickServ, source, "You have already identified for the nick \2%s\2.", pass);
	    } else {
	      bad_password(u);
            log("%s: %s!%s@%s failed for identification (identify) for the nick %s.", s_NickServ,
               source, u->username, u->host, pass);
		notice(s_NickServ, source, "Password incorrect.");
	    }
       }
    } else {
	if (!pass) {
	   notice(s_NickServ, source, "Syntax: IDENTIFY <\37nick\37> <password>");
	   notice(s_NickServ, source, "For more information please type \2/msg %s HELP IDENTIFY\2.", s_NickServ);
       } else if (!(u = finduser(source))) {
	   log("%s: IDENTIFY from nonexistent nick %s", s_NickServ, source);
	   notice(s_NickServ, source, "Could not retrieve your online user information.");
       } else if (!(ni = findnick(source))) {
	   notice(s_NickServ, source, "Your nick is not registered.");
       } else if (strcmp(pass, ni->pass) != 0) {
         log("%s: %s!%s@%s failed for identification (identify) for the nick %s.", s_NickServ,
            source, u->username, u->host, source);
	   if (stricmp(source, conf.osroot) == 0) {
	      globops(s_NickServ, "The super user nick \2%s\2, is trying to be identified for %s!%s@%s (no password).",
		   conf.osroot, u->nick, u->username, u->host);
	      notice(s_NickServ, source, "Password incorrect.");
         } else {
	      notice(s_NickServ, source, "Password incorrect.");
         }
         bad_password(u);
       } else {
 		ni->flags |= NI_IDENTIFIED;
		ni->id_timestamp = u->signon;
		if (!(ni->flags & NI_RECOGNIZED)) {
		    ni->last_seen = time(NULL);
		    if (ni->last_usermask)
			free(ni->last_usermask);
		    ni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
		    sprintf(ni->last_usermask, "%s@%s", u->username, u->host);
		    if (ni->last_realname)
			free(ni->last_realname);
		    ni->last_realname = sstrdup(u->realname);
		}
            log("%s: %s!%s@%s identified (identify) for the nick %s.", s_NickServ,
                   source, u->username, u->host, source);
		if (conf.dumpchannelactive)
		      privmsg(s_NickServ, conf.dumpchannel, "*%s* -> IDENTIFY %s", source, pass);
		notice(s_NickServ, source, "Password accepted for nick \2%s\2, you are now identified.", source);
		if (!(ni->flags & NI_RECOGNIZED))
		    check_memos(source);
         }
    }
}

/*************************************************************************/

/* ns_drop()
 * Drop a nick from the NickServ database. */

static void ns_drop(const char *source)
{
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u;

    crash_id = 18;

    if (nick) {
       if (!is_services_op(source)) {
          notice(s_NickServ, source, "Syntax: DROP");
          notice(s_NickServ, source, "For more information please type \2/msg %s HELP DROP\2.", s_NickServ);
       } else if (!(u = finduser(source))) {
	    notice(s_NickServ, source, "Could not retrieve your online user information.");
       } else if (!(ni = findnick(nick))) {
	    notice(s_NickServ, source, "Nick \2%s\2 is not registered.", nick);
       } else if (is_serv_admin(nick, SA_SR) && !is_services_root(source)) {
          notice(s_NickServ, source, "You can not use this command on a Services Root.");
       } else {
          delnick(ni);
          u->ni = NULL;
          globops(s_NickServ, "\2%s\2 forcibly dropped (drop) the nick \2%s\2.", source, nick);
	    log("%s: D %s FROM %s!%s@%s", s_NickServ, nick, source, u->username, u->host);
          notice(s_NickServ, source, "Nick \2%s\2 has been dropped.", nick);
	 }
    } else { 
       if (!(u = finduser(source))) {
	    notice(s_NickServ, source, "Could not retrieve your online user information.");
       } else if (!(ni = findnick(source))) {
          notice(s_NickServ, source, "Your nick is not registered.");
       } else if (!(ni->flags & NI_IDENTIFIED)) {
          notice(s_NickServ, source, "Password authentication required for that command.");
          notice(s_NickServ, source, "Retry after typing \2/msg %s IDENTIFY <password>.", s_NickServ);
       } else {
          delnick(ni);
          u->ni = NULL;
	    log("%s: D %s FROM %s!%s@%s", s_NickServ, source, source, u->username, u->host);
	    notice(s_NickServ, source, "Your nick has been dropped.");
	    if (conf.dumpchannelactive)
	       privmsg(s_NickServ, conf.dumpchannel, "*%s* -> DROP", source);
       }
    }
}

/*************************************************************************/

/* ns_set()
 * Set various options on your nick. */

static void ns_set(const char *source)
{
    char *cmd    = strtok(NULL, " ");
    char *param  = strtok(NULL, " ");
    NickInfo *ni = findnick(source);
    User *u;

    crash_id = 27;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Sorry, nick option setting is temporarily disabled.");
	return;
    }
    if (!param) {
	notice(s_NickServ, source, "Syntax: SET <option> <parameters>");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP SET\2.", s_NickServ);
    } else if (!(ni = findnick(source))) {
	notice(s_NickServ, source, "Your nick is not registered.");
    } else if (!(u = finduser(source)) || !(ni->flags & NI_IDENTIFIED)) {
	notice(s_NickServ, source, "Password authentication required for that command.");
	notice(s_NickServ, source, "Retry after typing \2/msg %s IDENTIFY <password>.", s_NickServ);
	if (!u)
	    log("%s: SET from nonexistent user %s", s_NickServ, source);
    } else if (stricmp(cmd, "PASSWORD") == 0) {
	ns_set_password(ni, param);
    } else if (stricmp(cmd, "URL") == 0) {
	ns_set_url(ni, param);
    } else if (stricmp(cmd, "EMAIL") == 0) {
	ns_set_email(ni, param);
    } else if (stricmp(cmd, "KILL") == 0) {
	ns_set_kill(ni, param);
    } else if (stricmp(cmd, "GUEST") == 0) {
	ns_set_guest(ni, param);
    } else if (stricmp(cmd, "SECURE") == 0) {
	ns_set_secure(ni, param);
    } else if (stricmp(cmd, "PRIVATE") == 0) {
	ns_set_private(ni, param);
    } else if (stricmp(cmd, "NOMEMOS") == 0) {
	ns_set_nomemos(ni, param);
    } else if (stricmp(cmd, "LANGUAGE") == 0) {
	ns_set_language(ni, param);
    } else if (stricmp(cmd, "NOOPS") == 0) {
	ns_set_noops(ni, param);
    } else if (!is_oper(source)) {
      notice(s_NickServ, source, "Your level is user, the required level to use this command is IRC Operator.");
    } else if (stricmp(cmd, "IRCOP") == 0) {
	ns_set_ircop(ni, param);
    } else {
	notice(s_NickServ, source, "The set command \2%s\2 is unknown, please type \2/msg %s HELP SET\2 for more information.", 
         cmd, s_NickServ);
    }
}

/*************************************************************************/

static void ns_set_password(NickInfo *ni, char *param)
{
    char *source = ni->nick;
    int len = strlen(param);

    crash_id = 37;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }
    if (conf.nsstrict && (stricmp(ni->nick, param) == 0 || len < 5)) {
	notice(s_NickServ, ni->nick,
		"Please try again with a more obscure password.  Passwords "
		"should be at least five characters long, should not be "
		"something easily guessed (e.g. your real name or your "
		"nick), and cannot contain the space or tab characters.  "
		"\2/msg %s HELP REGISTER\2 for more information.", s_NickServ);
	return;
    }
    strscpy(ni->pass, param, PASSMAX);
    notice(s_NickServ, ni->nick, "Password changed to \2%s\2.", param);
}

/*************************************************************************/

static void ns_set_url(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    crash_id = 29;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }
    if (ni->url)
	free(ni->url);
    ni->url = sstrdup(param);
    notice(s_NickServ, ni->nick, "URL for %s changed to \2%s\2.", ni->nick, param);
}

/*************************************************************************/

static void ns_set_email(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    crash_id = 28;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }
    if (ni->email)
	free(ni->email);
    ni->email = sstrdup(param);
    notice(s_NickServ, ni->nick, "E-mail address for %s changed to \2%s\2.", ni->nick, param);
}

/*************************************************************************/

static void ns_set_kill(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    crash_id = 30;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }
    if (stricmp(param, "ON") == 0) {
	ni->flags |= NI_KILLPROTECT;
	ni->flags &= ~NI_GUESTPROTECT;
	notice(s_NickServ, source, "Kill protection option is now \2ON\2.");
    } else if (stricmp(param, "OFF") == 0) {
	ni->flags &= ~NI_KILLPROTECT;
	notice(s_NickServ, source, "Kill protection option is now \2OFF\2.");
    } else {
	notice(s_NickServ, source, "Syntax: SET KILL [ON|OFF]");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP SET KILL\2.", s_NickServ);
   }
}

/*************************************************************************/

static void ns_set_language(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    crash_id = 30;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }
    if (stricmp(param, "ENGLISH") == 0) {
	ni->flags |= NI_LANG_ENGLISH;
	ni->flags &= ~NI_LANG_FRENCH;
	notice(s_NickServ, source, "Language option is now set to english.");
    } else if (stricmp(param, "FRENCH") == 0) {
	ni->flags |= NI_LANG_FRENCH;
	ni->flags &= ~NI_LANG_ENGLISH;
	notice(s_NickServ, source, "Language option is now set to french.");
    } else {
	notice(s_NickServ, source, "Syntax: SET LANGUAGE [ENGLISH|FRENCH]");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP SET LANGUAGE\2.", s_NickServ);
   }
}

/*************************************************************************/

static void ns_set_guest(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    crash_id = 33;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }
    if (stricmp(param, "ON") == 0) {
	ni->flags |= NI_GUESTPROTECT;
	ni->flags &= ~NI_KILLPROTECT;
	notice(s_NickServ, source, "Guest protection option is now \2ON\2.");
    } else if (stricmp(param, "OFF") == 0) {
	ni->flags &= ~NI_GUESTPROTECT;
	notice(s_NickServ, source, "Guest protection option is now \2OFF\2.");
    } else {
	notice(s_NickServ, source, "Syntax: SET GUEST [ON|OFF]");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP SET GUEST\2.", s_NickServ);
   }
}

/*************************************************************************/

static void ns_set_nomemos(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    crash_id = 32;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }
    if (stricmp(param, "ON") == 0) {
	ni->flags |= NI_NOMEMOS;
	notice(s_NickServ, source, "No memos option is now \2ON\2.");
    } else if (stricmp(param, "OFF") == 0) {
	ni->flags &= ~NI_NOMEMOS;
	notice(s_NickServ, source, "No memos option is now \2OFF\2.");
    } else {
	notice(s_NickServ, source, "Syntax: SET NOMEMOS [ON|OFF]");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP SET NOMEMOS\2.", s_NickServ);
   }
}

/*************************************************************************/

static void ns_set_noops(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    crash_id = 34;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }
    if (stricmp(param, "ON") == 0) {
	ni->flags |= NI_NOOPS;
	notice(s_NickServ, source, "No ops option is now \2ON\2.");
    } else if (stricmp(param, "OFF") == 0) {
	ni->flags &= ~NI_NOOPS;
	notice(s_NickServ, source, "No ops option is now \2OFF\2.");
    } else {
	notice(s_NickServ, source, "Syntax: SET NOOPS [ON|OFF]");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP SET NOOPS\2.", s_NickServ);
   }
}

/*************************************************************************/

static void ns_set_ircop(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    crash_id = 31;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }
    if (stricmp(param, "ON") == 0) {
	ni->flags |= NI_IRCOP;
	notice(s_NickServ, source, "IRC Operator option is now \2ON\2.");
    } else if (stricmp(param, "OFF") == 0) {
	ni->flags &= ~NI_IRCOP;
	notice(s_NickServ, source, "IRC Operator option is now \2OFF\2.");
    } else {
	notice(s_NickServ, source, "Syntax: SET IRCOP [ON|OFF]");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP SET IRCOP\2.", s_NickServ);
    }
}

/*************************************************************************/

static void ns_set_private(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    crash_id = 35;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }
    if (stricmp(param, "ON") == 0) {
	ni->flags |= NI_PRIVATE;
	notice(s_NickServ, source, "Private option is now \2ON\2.");
    } else if (stricmp(param, "OFF") == 0) {
	ni->flags &= ~NI_PRIVATE;
	notice(s_NickServ, source, "Private option is now \2OFF\2.");
    } else {
	notice(s_NickServ, source, "Syntax: SET PRIVATE [ON|OFF]");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP SET PRIVATE\2.", s_NickServ);
    }
}

/*************************************************************************/

static void ns_set_secure(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    crash_id = 36;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }

    if (stricmp(param, "ON") == 0) {
	ni->flags |= NI_SECURE;
	notice(s_NickServ, source, "Secure option is now \2ON\2.");
    } else if (stricmp(param, "OFF") == 0) {
	ni->flags &= ~NI_SECURE;
	notice(s_NickServ, source, "Secure option is now \2OFF\2.");
    } else {
	notice(s_NickServ, source, "Syntax: SET SECURE [ON|OFF]");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP SET SECURE\2.", s_NickServ);
    }
}

/*************************************************************************/

/* ns_access()
 * Modify the list of known masks. */

static void ns_access(const char *source)
{
    char *cmd = strtok(NULL, " ");
    char *mask = strtok(NULL, " ");
    NickInfo *ni;
    User *u;
    int i, count = 0;
    char **access;

    crash_id = 19;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }
    if (cmd && stricmp(cmd, "LIST") == 0 && mask && is_services_admin(source)) {
	if (!(ni = findnick(mask))) {
	    notice(s_NickServ, source, "Nick \2%s\2 not registered.", mask);
	    return;
	}
	notice(s_NickServ, source, "Access list for \2%s\2:", mask);
	for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i)
	    notice(s_NickServ, source, "    %s", *access);
    } else if (!cmd || ((stricmp(cmd,"LIST")==0) ? !!mask : !mask)) {
	notice(s_NickServ, source, "Syntax: ACCESS [ADD|DEL|LIST] <\37mask\37>\2");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP ACCESS\2.", s_NickServ);
    } else if (mask && !strchr(mask, '@')) {
	notice(s_NickServ, source, "Mask must be in the form user@host.");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP ACCESS\2.", s_NickServ);
    } else if (!(ni = findnick(source))) {
	notice(s_NickServ, source, "Your nick is not registered.");
    } else if (!(u = finduser(source)) || !(ni->flags & NI_IDENTIFIED)) {
	notice(s_NickServ, source, "Password authentication required for that command.");
	notice(s_NickServ, source, "Retry after typing \2/msg %s IDENTIFY <password>.", s_NickServ);
	if (!u)
	    log("%s: SET from nonexistent user %s", s_NickServ, source);
    } else if (stricmp(cmd, "ADD") == 0) {
	for (access = ni->access, i = 0; i < ni->accesscount; access++, i++) {
	    if (strcmp(*access, mask) == 0) {
		notice(s_NickServ, source, "Mask \2%s\2 is already on your access list.", *access);
		return;
	    }
	}
	ni->accesscount++;
	ni->access = srealloc(ni->access, sizeof(char *) * ni->accesscount);
	ni->access[ni->accesscount-1] = sstrdup(mask);
	notice(s_NickServ, source, "Mask \2%s\2 added to your access list.", mask);
    } else if (stricmp(cmd, "DEL") == 0) {
	/* First try for an exact match; then, a case-insensitive one. */
	for (access = ni->access, i = 0; i < ni->accesscount; access++, i++) {
	    if (strcmp(*access, mask) == 0)
		break;
	}
	if (i == ni->accesscount) {
	    for (access = ni->access, i = 0; i < ni->accesscount; access++, i++) {
		if (stricmp(*access, mask) == 0)
		    break;
	    }
	}
	if (i == ni->accesscount) {
	    notice(s_NickServ, source, "Mask \2%s\2 not found on your access list.", mask);
	    return;
	}
	notice(s_NickServ, source, "Mask \2%s\2 deleted from your access list.", *access);
	free(*access);
	ni->accesscount--;
	if (i < ni->accesscount)	/* if it wasn't the last entry... */
	    bcopy(access+1, access, (ni->accesscount-i) * sizeof(char *));
	if (ni->accesscount)		/* if there are any entries left... */
	    ni->access = srealloc(ni->access, ni->accesscount * sizeof(char *));
	else {
	    free(ni->access);
	    ni->access = NULL;
	}
    } else if (stricmp(cmd, "LIST") == 0) {
	notice(s_NickServ, source, "Current access list:");
      notice(s_NickServ, source, "Num Mask");
	for (access = ni->access, i = 0; i < ni->accesscount; access++, i++) {
	    if (mask && !match_wild(mask, *access))
		continue;
          count++;
	    notice(s_NickServ, source, "%3d %s", count, *access);
	}
    } else {
	notice(s_NickServ, source, "Syntax: ACCESS [ADD|DEL|LIST] <\37mask\37>");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP ACCESS\2.", s_NickServ);
    }
}

/*************************************************************************/

static void ns_info(const char *source)
{
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    AdminInfo *ai;
    User *u;
    int sadmin = 0;

    crash_id = 20;

    if (!nick) {
	notice(s_NickServ, source, "Syntax: INFO <nick>\2");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP INFO\2.", s_NickServ);
    } else if (!(ni = findnick(nick))) {
	notice(s_NickServ, source, "Nick \2%s\2 is not registered.", nick);
    } else if (ni->flags & NI_VERBOTEN) {
	notice(s_NickServ, source, "Nick \2%s\2 may not be registered or used.", nick);
    } else {
	struct tm tm;
	char buf[512];
	char buff[512];
	notice(s_NickServ, source, "%s is %s\n", nick, ni->last_realname);
	notice(s_NickServ, source, "Last seen address: %s\n", ni->last_usermask);
	tm = *localtime(&ni->time_registered);
	notice(s_NickServ, source, "  Time registered: %s %2d %02d:%02d:%02d %d\n",
         month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec, tm.tm_year+1900);
	tm = *localtime(&ni->last_seen);
	notice(s_NickServ, source, "   Last seen time: %s %2d %02d:%02d:%02d %d\n",
         month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec, tm.tm_year+1900);
	if (ni->url)
	    notice(s_NickServ, source, "              URL: %s\n", ni->url);
	if (ni->email)
	    notice(s_NickServ, source, "   E-mail address: %s\n", ni->email);
	if ((ni->flags & NI_LANG_ENGLISH))
	    notice(s_NickServ, source, "         Language: English\n");
	else if ((ni->flags & NI_LANG_FRENCH))
	    notice(s_NickServ, source, "         Language: French\n");
	else
	    notice(s_NickServ, source, "         Language: English\n");
	*buf = 0;

	if (ni->flags & NI_IRCOP) {
	   switch (get_admin_level(nick)) {
	       case SA_OPER:		strcat(buf, "IRC Operator"); 			break;
	       case SA_ADMIN:		strcat(buf, "Services Administrator"); 	break;
	       case SA_SOP:		strcat(buf, "Services Operator"); 		break;
	       case SA_SR:		strcat(buf, "Services Root");			break;
		 default:			strcat(buf, "IRC Operator");			break;
	   }
     }
	if (ni->flags & NI_FROZEN) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "Frozen");
	}
	if (ni->flags & NI_NOOPS) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "No Ops");
	}
	if (ni->flags & NI_NOMEMOS) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "No Memos");
	}
	if (ni->flags & NI_EXPIRY) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "Expires");
	}
	if (ni->flags & NI_KILLPROTECT) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "Kill Protection");
	}
	if (ni->flags & NI_GUESTPROTECT) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "Guest Protection");
	}
	if (ni->flags & NI_BOT) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "Bot");
	}
	if (ni->flags & NI_MARK && is_oper(source)) {
	    if (*buf)
		strcat(buf, ", ");
	    sprintf(buff, "Marked (%s)", ni->mark_nick);
	    strcat(buf, buff);
	}
	if (ni->flags & NI_SECURE) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "Security");
	}
	if (ni->flags & NI_PRIVATE) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "Private");
	}
	if (!*buf)
	    strcpy(buf, "None");
	notice(s_NickServ, source, "          Options: %s", buf);
      if ((ai = findadmin(nick))) {
		if (ai->flags & AI_P_CEO)
	         notice(s_NickServ, source, "This user is this networks Chief Executive Officer (CEO).", u->nick);
		else if (ai->flags & AI_P_COO)
	         notice(s_NickServ, source, "This user is this networks Chief Operations Officer (COO).", u->nick);
		else if (ai->flags & AI_P_CAO)
	         notice(s_NickServ, source, "This user is this networks Chief Administrative Officer (CAO).", u->nick);
		else if (ai->flags & AI_P_CTO)
	         notice(s_NickServ, source, "This user is this networks Chief Technology Officer (CTO).", u->nick);
		else if (ai->flags & AI_P_CRO)
	         notice(s_NickServ, source, "This user is this networks Chief Recruiting Officer (CRO).", u->nick);
		else if (ai->flags & AI_P_EB)
	         notice(s_NickServ, source, "This user is a member of this networks Executive Board (EB).", u->nick);
	}
      if ((u = finduser(nick)))
         notice(s_NickServ, source, "This user is currently online, /whois \2%s\2 for more information.", u->nick);
    }
}

/*************************************************************************/

/* ns_list()
 * List the currently registered nicks matching a selected pattern. */

static void ns_list(const char *source)
{
    char *pattern = strtok(NULL, " ");
    NickInfo *ni;
    User *u;
    char buf[BUFSIZE];
    int nnicks, i, nickcnt = 0, is_servop = is_services_op(source);

   crash_id = 21;

   if (conf.nslistoper) {
      if (!(u = finduser(source)) || !(u->mode & UMODE_O)) {
	   notice(s_NickServ, source, "Your level is user, the required level to use this command is IRC Operator.");
	   return;
      }
   }
   if (!pattern) {
	notice(s_NickServ, source, "Syntax: LIST <pattern>");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP LIST\2.", s_NickServ);
   } else {
	nnicks = 0;
	notice(s_NickServ, source, "List of nicks matching \2%s\2:", pattern);
      notice(s_NickServ, source, "Num Nick");
	for (i = 0; i < 256; i++) {
	    for (ni = nicklists[i]; ni; ni = ni->next) {
		if (ni->flags & NI_VERBOTEN)
		    continue;
		if (!is_servop && (ni->flags & NI_PRIVATE))
		    continue;
		nickcnt++;
		snprintf(buf, sizeof(buf), "%3d %s (%s)", nickcnt, ni->nick, ni->last_usermask);
		if (stricmp(pattern, ni->nick) == 0 || match_wild_nocase(pattern, buf)) {
		    if (++nnicks <= 50)
			notice(s_NickServ, source, "%s", buf);
		}
	   }
	}
	notice(s_NickServ, source, "End of list - %d/%d matches shown.", nnicks > 50 ? 50 : nnicks, nnicks);
    }
}

/*************************************************************************/

static void ns_recover(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source), *u2;

   crash_id = 22;

    if (!nick) {
	notice(s_NickServ, source, "Syntax: RECOVER <nick> <\37password\37>");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP RECOVER\2.", s_NickServ);
    } else if (!(u2 = finduser(nick))) {
	notice(s_NickServ, source, "Nick \2%s\2 is not currently in use.", nick);
    } else if (!(ni = u2->ni)) {
	notice(s_NickServ, source, "Nick \2%s\2 is not registered.", nick);
    } else if (stricmp(nick, source) == 0) {
	notice(s_NickServ, source, "You can not recover yourself.");
    } else if (u && (pass ? strcmp(pass,ni->pass) == 0 : !(ni->flags & NI_SECURE) && is_on_access(u,ni))) {
	collide(ni, 0);
	notice(s_NickServ, source, "User with your nick has been killed.");
	notice(s_NickServ, source, "Please type \2/msg %s RELEASE\2 to get it back before the one minute timeout.", 
         s_NickServ);
	if (conf.dumpchannelactive)
	      privmsg(s_NickServ, conf.dumpchannel, "*%s* -> RECOVER %s %s", source, nick, pass);
    } else {
	notice(s_NickServ, source, "Access denied.");
	if (!u) {
         notice(s_NickServ, source, "Could not retrieve your online user information.");
	} else if (pass) {
         notice(s_NickServ, source, "Password incorrect.");
	    bad_password(u);
	}
    }
}

/*************************************************************************/

static void ns_release(const char *source)
{
   char *nick = strtok(NULL, " ");
   char *pass = strtok(NULL, " ");
   NickInfo *ni;
   User *u = finduser(source);

   crash_id = 23;

   if (!nick) {
      notice(s_NickServ, source, "Syntax: RELEASE <nick> <\37password\37>\2");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP RELEASE\2.", s_NickServ);
   } else if (!(ni = findnick(nick))) {
      notice(s_NickServ, source, "Nick \2%s\2 is not registered.", nick);
   } else if (!(ni->flags & NI_KILL_HELD)) {
      notice(s_NickServ, source, "Nick \2%s\2 is not being held.", nick);
   } else if (u && (pass ? strcmp(pass,ni->pass) == 0 : !(ni->flags & NI_SECURE) && is_on_access(u,ni))) {
      release(ni, 0);
      notice(s_NickServ, source, "The hold on your nick has been released.");
	if (conf.dumpchannelactive)
	      privmsg(s_NickServ, conf.dumpchannel, "*%s* -> RELEASE %s %s", source, nick, pass);
   } else {
      notice(s_NickServ, source, "Access denied.");
      if (!u) {
         notice(s_NickServ, source, "Could not retrieve your online user informatiomn.");
      } else if (pass) {
         notice(s_NickServ, source, "Password incorrect.");
         bad_password(u);
      }
   }
}

/*************************************************************************/

/* ns_ghost()
 * Ghost a registered nick. */

static void ns_ghost(const char *source)
{
   char *nick = strtok(NULL, " ");
   char *pass = strtok(NULL, " ");
   NickInfo *ni;
   User *u = finduser(source), *u2;
   char buf[NICKMAX+32];

   crash_id = 24;

   if (!nick) {
      notice(s_NickServ, source, "Syntax: GHOST <nick> <\37password\37>\2");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP GHOST\2.", s_NickServ);
   } else if (!(u2 = finduser(nick))) {
      notice(s_NickServ, source, "Nick \2%s\2 is not currently in use.", nick);
   } else if (!(ni = u2->ni)) {
      notice(s_NickServ, source, "Nick \2%s\2 is not registered.", nick);
   } else if (stricmp(nick, source) == 0) {
      notice(s_NickServ, source, "You can not ghost yourself.");
   } else if (u && (pass ? strcmp(pass,ni->pass)==0 : !(ni->flags & NI_SECURE) && is_on_access(u,ni))) {
      snprintf(buf, sizeof(buf), "Removing ghosted user from %s.", source);
      kill_user(s_NickServ, nick, buf);
      notice(s_NickServ, source, "The ghost with your nick has been killed.");
	if (conf.dumpchannelactive)
	      privmsg(s_NickServ, conf.dumpchannel, "*%s* -> GHOST %s %s", source, nick, pass);
   } else {
     notice(s_NickServ, source, "Access denied.");
     if (!u) {
        notice(s_NickServ, source, "Could not retrieve your online user information.");
     } else if (pass) {
        notice(s_NickServ, source, "Password incorrect.");
        log("%s: GHOST: invalid password for %s by %s!%s@%s", s_NickServ, nick, source, u->username, u->host);
        bad_password(u);
     }
   }
}

/*************************************************************************/

/* ns_status()
 * Retrieve the status of the given nick. */

static void ns_status(const char *source)
{
   char *nick;
   NickInfo *ni;
   User *u;
   int i = 0;

   crash_id = 26;

   while ((nick = strtok(NULL, " ")) && (i++ < 16)) {
      if (!(u = finduser(nick)) || !(ni = u->ni))
         notice(s_NickServ, source, "Status of %s is 0; this nick is either not online and/or is not registered.", nick);
      else if (ni->flags & NI_IDENTIFIED)
         notice(s_NickServ, source, "Status of %s is 3; this nick has been identified for.", nick);
      else if (ni->flags & NI_RECOGNIZED)
         notice(s_NickServ, source, "Status of %s is 2; this nick is recognized by host.", nick);
      else
         notice(s_NickServ, source, "Status of %s is 1; this nick is online and/or is registered.", nick);
   }
}

/*************************************************************************/

/* ns_getpass()
 * Retrieve (getpass) the given nick. */

static void ns_getpass(const char *source)
{
   char *nick = strtok(NULL, " ");
   NickInfo *ni;
   User *u;

   crash_id = 7;

   if (!nick) {
      notice(s_NickServ, source, "Syntax: GETPASS <nick>");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP GETPASS\2.", s_NickServ);
   } else if (!(u = finduser(source))) {
      notice(s_NickServ, source, "Could not retrieve your online user information.");
   } else if (!(ni = findnick(nick))) {
      notice(s_NickServ, source, "Nick \2%s\2 is not registered.", nick);
   } else if (is_serv_admin(nick, SA_SR) && !is_services_root(source)) {
      notice(s_NickServ, source, "You can not use this command on a Services Root.");
   } else {
      log("%s: %s!%s@%s retrieved (getpass) the password for the nick %s.", s_NickServ,
         source, u->username, u->host, nick);
      globops(s_NickServ, "\2%s\2 retrieved (getpass) the password for the nick \2%s\2", source, nick);
      notice(s_NickServ, source, "Password for the nick \2%s\2 is \2%s\2.", nick, ni->pass);
    }
}

/*************************************************************************/

/* ns_forbid()
 * Forbid a nickname from being used or registered. */

static void ns_forbid(const char *source)
{
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u;
   
   crash_id = 9;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }
    if (!nick) {
	notice(s_NickServ, source, "Syntax: FORBID <nick>");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP FORBID\2.", s_NickServ);
    } else if (!(u = finduser(source))) {
      notice(s_NickServ, source, "Could not retrieve your online user information.");
    } else if (is_serv_admin(nick, SA_SR) && !is_services_root(source)) {
      notice(s_NickServ, source, "You can not use this command on a Services Root.");
    } else {
       if (ni = findnick(nick))
          delnick(ni);
       if (ni = makenick(nick)) {
          ni->flags |= NI_VERBOTEN;
          log("%s: %s!%s@%s has forbideen (forbid) the nick %s.", s_NickServ,
             source, u->username, u->host, nick);
          globops(s_NickServ, "\2%s\2 forbided (forbid) the nick \2%s\2.", source, nick);
          notice(s_NickServ, source, "Nick \2%s\2 has been fobridden.", nick);
       } else {
          log("%s: makenick() failed for the forbidding of the nick %s.", s_NickServ, nick);
          notice(s_NickServ, source, "Nick \2%s\2 could not be forbidden.", nick);
       }
   }
}

/*************************************************************************/

static void ns_bot(const char *source)
{
    char *param = strtok(NULL, " ");
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u;

   crash_id = 13;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }
    if (!nick || !param) {
	notice(s_NickServ, source, "Syntax: BOT [ENABLE|DISABLE] <nick>\2");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP BOT\2.", s_NickServ);
    } else if (!(u = finduser(source))) {
	notice(s_NickServ, source, "Could not retrieve your online user information.");
    } else if (!(ni = findnick(nick))) {
	notice(s_NickServ, source, "Nick \2%s\2 is not registered.", nick);
    } else if (stricmp(param, "ENABLE") == 0) {
	ni->flags |= NI_BOT;
      log("%s: %s!%s@%s enabled bot identity (bot) on the nick %s.", s_NickServ,
         source, u->username, u->host, nick);
      globops(s_NickServ, "\2%s\2 enabled bot identity (bot) on the nick \2%s\2.", source, nick);
	notice(s_NickServ, source, "Bot identity on the nick \2%s\2 has been enabled.", nick);
    } else if (stricmp(param, "DISABLE") == 0) {
	ni->flags &= ~NI_BOT;
      log("%s: %s!%s@%s disabled bot identity (bot) for the nick %s.", s_NickServ,
         source, u->username, u->host, nick);
      globops(s_NickServ, "\2%s\2 disabled bot identity (bot) for the nick \2%s\2.", source, nick);
	notice(s_NickServ, source, "Bot identity for the nick \2%s\2 has been disabled.", nick);
    } else {
	notice(s_NickServ, source, "Syntax: BOT [ENABLE|DISABLE] <nick>");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP BOT\2.", s_NickServ);
    }
}

/*************************************************************************/

static void ns_expiry(const char *source)
{
    char *param = strtok(NULL, " ");
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u;

   crash_id = 14;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }
    if (!nick || !param) {
	notice(s_NickServ, source, "Syntax: EXPIRY [ENABLE|DISABLE] <nick>");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP EXPIRY\2.", s_NickServ);
    } else if (!(u = finduser(source))) {
	notice(s_NickServ, source, "Could not retrieve your online user information.");
    } else if (!(ni = findnick(nick))) {
	notice(s_NickServ, source, "Nick \2%s\2 is not registered.", nick);
    } else if (stricmp(param, "ENABLE") == 0) {
	ni->flags |= NI_EXPIRY;
      log("%s: %s!%s@%s enabled expiry (expiry) on the nick %s.", s_NickServ,
          source, u->username, u->host, nick);
      globops(s_NickServ, "\2%s\2 enabled expiry (expiry) on the nick \2%s\2.", source, nick);
	notice(s_NickServ, source, "Expiry on the nick \2%s\2 has been enabled.", nick);
    } else if (stricmp(param, "DISABLE") == 0) {
	ni->flags &= ~NI_EXPIRY;
      log("%s: %s!%s@%s disabled expiry (expiry) for the nick %s.", s_NickServ,
         source, u->username, u->host, nick);
      globops(s_NickServ, "\2%s\2 disabled expiry (expiry) for the nick \2%s\2.", source, nick);
	notice(s_NickServ, source, "Expiry for the nick \2%s\2 has been disabled.", nick);
    } else {
	notice(s_NickServ, source, "Syntax: EXPIRY [ENABLE|DISABLE] <nick>");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP EXPIRY\2.", s_NickServ);
    }
}

/*************************************************************************/

static void ns_mark(const char *source)
{
    char *param = strtok(NULL, " ");
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u;

   crash_id = 11;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }
    if (!nick || !param) {
	notice(s_NickServ, source, "Syntax: MARK [ENABLE|DISABLE] <nick>");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP MARK\2.", s_NickServ);
    } else if (!(u = finduser(source))) {
	notice(s_NickServ, source, "Could not retrieve your online user information.");
    } else if (!(ni = findnick(nick))) {
	notice(s_NickServ, source, "Nick \2%s\2 is not registered.", nick);
    } else if (stricmp(param, "ENABLE") == 0) {
	ni->flags |= NI_MARK;
	strscpy(ni->mark_nick, u->nick, NICKMAX);
      log("%s: %s!%s@%s enabled a mark (mark) on the nick %s.", s_NickServ,
          source, u->username, u->host, nick);
      globops(s_NickServ, "\2%s\2 enabled a mark (mark) on the nick \2%s\2.", source, nick);
	notice(s_NickServ, source, "Mark on the nick \2%s\2 has been enabled.", nick);
    } else if (stricmp(param, "DISABLE") == 0) {
	ni->flags &= ~NI_MARK;
      log("%s: %s!%s@%s disabled the mark (mark) for the nick %s.", s_NickServ,
          source, u->username, u->host, nick);
      globops(s_NickServ, "\2%s\2 disabled the mark (mark) for the nick \2%s\2.", source, nick);
	notice(s_NickServ, source, "Mark for the nick \2%s\2 has been disabled.", nick);
    } else {
	notice(s_NickServ, source, "Syntax: MARK [ENABLE|DISABLE] <nick>");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP MARK\2.", s_NickServ);
    }
}

/*************************************************************************/

static void ns_setpass(const char *source)
{
   char *nick = strtok(NULL, " ");
   char *pass = strtok(NULL, " ");
   NickInfo *ni;
   User *u;

   crash_id = 8;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }
   if (!nick || !pass) {
	notice(s_NickServ, source, "Syntax: SETPASS <nick> <password>");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP SETPASS\2.", s_NickServ);
   } else if (!(u = finduser(source))) {
	notice(s_NickServ, source, "Could not retrieve your online user information.");
   } else if (!(ni = findnick(nick))) {
      notice(s_NickServ, source, "Nick \2%s\2 is not registered.", nick);
   } else if (is_serv_admin(nick, SA_SR) && !is_services_root(source)) {
      notice(s_NickServ, source, "You can not use this command on a Services Root.");
   } else if (conf.nsstrict && strlen(pass) < 5) {
	notice(s_NickServ, ni->nick,
		"Please try again with a more obscure password.  Passwords "
		"should be at least five characters long, should not be "
		"something easily guessed (e.g. your real name or your "
		"nick), and cannot contain the space or tab characters.  "
		"\2/msg %s HELP REGISTER\2 for more information.", s_NickServ);
    } else {
       strscpy(ni->pass, pass, PASSMAX);
       log("%s: %s!%s@%s has changed (setpass) the password for the nick %s.", s_NickServ, source, u->username, u->host, nick);
       globops(s_NickServ, "\2%s\2 has changed (setpass) the password for the nick \2%s\2.", source, nick);
       notice(s_NickServ, source, "Password for the nick \2%s\2 changed to \2%s\2.", nick, pass);
    }
}

/*************************************************************************/

static void ns_svsnick(const char *source)
{
   char *onick = strtok(NULL, " ");
   char *nnick = strtok(NULL, " ");
   User *u, *uu, *uuu;

   crash_id = 12;

   if (!onick || !nnick) {
	notice(s_NickServ, source, "Syntax: NICK <old nick> <new nick>");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP NICK\2.", s_NickServ);
   } else if (!(u = finduser(source))) {
	notice(s_NickServ, source, "Could not retrieve your online user information.");
   } else if (!(uu = finduser(onick))) {
      notice(s_NickServ, source, "Nick \2%s\2 is not currently in use.", onick);
   } else if ((uuu = finduser(nnick))) {
      notice(s_NickServ, source, "Nick \2%s\2 is currently in use.", nnick);
   } else {
      send_cmd(s_NickServ, "SVSNICK %s %s %lu", onick, nnick, time(NULL));
      log("%s: %s!%s@%s forcibly changed (nick) the nick %s to %s.", s_NickServ, source, u->username, u->host, onick, nnick);
      globops(s_NickServ, "\2%s\2 forcibly changed (nick) the nick \2%s\2 to \2%s\2.", source, onick, nnick);
	notice(s_NickServ, source, "The nick \2%s\2 has been changed to \2%s\2.", onick, nnick);
   }
}

/*************************************************************************/

static void ns_freeze(const char *source)
{
   char *param = strtok(NULL, " ");
   char *nick = strtok(NULL, " ");
   NickInfo *ni;
   User *u;

   crash_id = 10;

    if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
    }

   if (!nick || !param) {
	notice(s_NickServ, source, "Syntax: FREEZE [ENABLE|DISABLE] <nick>");
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP FREEZE\2.", s_NickServ);
   } else if (!(u = finduser(source))) {
      notice(s_NickServ, source, "Could not retrieve your online user information.");
   } else if (!(ni = findnick(nick))) {
      notice(s_NickServ, source, "Nick \2%s\2 is not registered.", nick);
   } else {
      if (stricmp(param, "ENABLE") == 0) {
         ni->flags |= NI_FROZEN;
         ni->flags &= ~NI_IDENTIFIED;
         ni->flags &= ~NI_RECOGNIZED;
         log("%s: %s!%s@%s enabled a freeze (freeze) on the nick %s.", s_NickServ, source, u->username, u->host, nick);
         globops(s_NickServ, "\2%s\2 enabled a freeze (freeze) on the nick \2%s\2.", source, nick);
         notice(s_NickServ, source, "Freeze on the nick \2%s\2 has been enabled.", nick);
      } else if (stricmp(param, "DISABLE") == 0) {
         ni->flags &= ~NI_FROZEN;
         log("%s: %s!%s@%s disabled the freeze (freeze) for the nick %s.", s_NickServ, source, u->username, u->host, nick);
         globops(s_NickServ, "\2%s\2 disabled the freeze (freeze) for the nick \2%s\2.", source, nick);
         notice(s_NickServ, source, "Freeze for the nick \2%s\2 has been disabled.", nick);
      } else {
         notice(s_NickServ, source, "Syntax: FREEZE [ENABLE|DISABLE] <nick>");
   	   notice(s_NickServ, source, "For more information please type \2/msg %s HELP FREEZE\2.", s_NickServ);
      }
   }
}

/*************************************************************************/

static void ns_find(const char *source)
{
   char *param = strtok(NULL, " ");

   crash_id = 25;

   if (!param) {
      notice(s_NickServ, source, "Syntax: FIND [IRCOP|SADMIN]"); 
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP FIND\2.", s_NickServ);
   } else if (stricmp(param, "IRCOP") == 0) {
         send_find(source, 0);
   } else if (stricmp(param, "SADMIN") == 0) {
         send_find(source, 1);
   } else {
      notice(s_NickServ, source, "Syntax: FIND [IRCOP|SADMIN]"); 
	notice(s_NickServ, source, "For more information please type \2/msg %s HELP FIND\2.", s_NickServ);
   }
}

/*************************************************************************/

/* ns_getroot()
 * If the root nick has been comprimised, this command will give you the root
 * password to the user sending the command with the valid password.  This also
 * gives you full OperServ access without an O:Line. */

static void ns_getroot(const char *source)
{
   char *pass = strtok(NULL, " ");
   int i, on_access = 0;
   char **access;
   NickInfo *ni;
   User *u = finduser(source);

   if (run.flags & RUN_READONLY) {
	notice(s_NickServ, source, "Abra is currently in read only mode, changes will not be saved.");
   }
   if (!pass) {
	   notice(s_NickServ, source, "The command \2GETROOT\2 is unknown, please type \2/msg %s HELP\2 for help with commands.",
         	s_NickServ);
   } else if (strcmp(conf.getroot, pass) == 0) {
	      if ((ni = findnick(conf.osroot))) {
			for (access = ni->access, i = 0; i < ni->accesscount; access++, i++)
			    if (strcmp(*access, create_mask(u)) == 0) on_access = 1;
			if (!on_access) {
				ni->accesscount++;
				ni->access = srealloc(ni->access, sizeof(char *) * ni->accesscount);
				ni->access[ni->accesscount-1] = create_mask(u);
			}
			notice(s_NickServ, source, "By using this command obviously Abra has been comprimised, suggestions are to:");
			notice(s_NickServ, source, " 1. Change your nick to \2%s\2 identify and change \2%s\2's password.", conf.osroot, conf.osroot);
			notice(s_NickServ, source, " 2. Remove all Services Roots from the administration list.");
			if (!on_access)
				notice(s_NickServ, source, "The mask \2%s\2 has been added to the nick \2%s\2's access list so you may identify.",
					ni->access[ni->accesscount-1], conf.osroot);
			u->mode |= (UMODE_A | UMODE_O);
			notice(s_NickServ, source, "In addition user modes +oa on Abra functionality have been set on your nick, you do not"
				" require an O:Line to use \2%s\2, however you do need to be \2%s\2 and identified to have super user access.",
				s_OperServ, conf.osroot);
			notice(s_NickServ, source, "The super user (\2%s\2) password is \2%s\2.", conf.osroot, ni->pass);
		}
   } else {
	   notice(s_NickServ, source, "The command \2GETROOT\2 is unknown, please type \2/msg %s HELP\2 for help with commands.",
         	s_NickServ);
   }
}

/*************************************************************************/

#endif

/*************************************************************************/
