/* OperServ functions.
 * src/operserv.c
 * Copyright (C) 2000-2001 DarkSiderz Net.  All rights reserved.
 * Read the file doc/contact.html for contact information related to 
 * this software. */

/*************************************************************************/

#ifndef OPERSERV_C
#define OPERSERV_C

/*************************************************************************/

#include "timeout.h"
#include "services.h"

/*************************************************************************/

#include "os-help.c"

/*************************************************************************/

static AdminInfo *add_admin 			FUNCTION((const char *nick));
static void alpha_insert_admin 		FUNCTION((AdminInfo *ai));
static RouteInfo *add_route 			FUNCTION((const char *server));
static void alpha_insert_route 		FUNCTION((RouteInfo *ri));
static CloneList *add_clone 			FUNCTION((const char *host));
static void alpha_insert_clone 		FUNCTION((CloneList *cl));
static void add_akill 				FUNCTION((const char *mask, const char *reason, const char *who, const char *expiry,
								    unsigned long tracking, const int block));
static int del_akill 				FUNCTION((const char *mask));
static void timeout_leave 			FUNCTION((Timeout *to));
static void os_help 				FUNCTION((const char *source));
static void os_admin 				FUNCTION((const char *source));
static void os_akill 				FUNCTION((const char *source));
static void os_readonly 			FUNCTION((const char *source));
static void os_stats 				FUNCTION((const char *source));
static void os_position 			FUNCTION((const char *source));
static void os_jupe 				FUNCTION((const char *source));
static void os_kick 				FUNCTION((const char *source));
static void os_mode 				FUNCTION((const char *source));
static void os_mdeop 				FUNCTION((const char *source));
static void os_mkick 				FUNCTION((const char *source));
static void os_sinvite 				FUNCTION((const char *source));
static void os_schan	 			FUNCTION((const char *source));
static void os_map 				FUNCTION((const char *source));
static void os_raw 				FUNCTION((const char *source));
static void os_activate 			FUNCTION((const char *source));
static void activate_client_notice 		FUNCTION((const char *source, char *client, int onoff));
static void os_shutdown 			FUNCTION((const char *source));
static void os_update 				FUNCTION((const char *source));
static void os_rehashconf 			FUNCTION((const char *source));
static void os_quit 				FUNCTION((const char *source));
static void os_restart 				FUNCTION((const char *source));
static void os_route 				FUNCTION((const char *source));
static void os_flog 				FUNCTION((const char *source));
static void os_noop 				FUNCTION((const char *source));
static void os_set				FUNCTION((const char *source));
static void os_set_aop				FUNCTION((AdminInfo *ai, const char *source, char *param));

/*************************************************************************/

static AdminInfo *adminlists[256];
static RouteInfo *routelists[256];
static CloneList *clonelists[256];
char route_server[HOSTMAX];

/*************************************************************************/

static int nakill = 0;
static int akill_size = 0;
static struct akill {
    char *mask;
    char *reason;
    char who[NICKMAX];
    time_t time;
    time_t expires;
    unsigned long tracking;
    long flags;
    long reserved[4];
} *akills = NULL;

#define AK_BLOCK	0x00000001
#define AK_AKILL	0x00000002

/*************************************************************************/

#ifdef CHECK_CLONES
struct clone {
   char *host;
   long time;
};
/* List of most recent users - statically initialized to zeros */
static struct clone clonelist[CLONE_DETECT_SIZE];
/* Which hosts have we warned about, and when?  This is used to keep us
 * from sending out notices over and over for clones from the same host. */
static struct clone warnings[CLONE_DETECT_SIZE];
#endif

/*************************************************************************/

/* operserv()
 * Main OperServ routine. */

void operserv(const char *source, char *buf)
{
   char *cmd =  strtok(buf, " ");
   User *u = finduser(source);
   char *s;
   int i;
   char slvl[32];
 
   log("%s: %s: %s", s_OperServ, source, buf);

   if (!cmd) {
      return;
   } else if (stricmp(cmd, "\1PING") == 0) {
	if (!(s = strtok(NULL, "")))
	    s = "\1";
	notice(s_OperServ, source, "\1PING %s", s);
   } else if (!u) {
	notice(s_OperServ, source, rpl_numeric_lang(source, CR_NO_ONLINE_USER_INFO));
   } else {
	Hash *os, os_hash_table[] = {
	    { "HELP",		SA_OPER, 	os_help,		H_NIDENTIFY	},
	    { "ADMIN", 		SA_OPER, 	os_admin,		H_NIDENTIFY	},
	    { "KICK", 		SA_OPER, 	os_kick,		H_NIDENTIFY	},
	    { "STATS", 		SA_OPER, 	os_stats,		H_NIDENTIFY	},
	    { "MAP",		SA_OPER, 	os_map,		H_NIDENTIFY	},
	    { "AKILL", 		SA_ADMIN, 	os_akill,		H_NIDENTIFY	},
	    { "MODE", 		SA_ADMIN, 	os_mode,		H_NIDENTIFY	},
	    { "FLOG", 		SA_ADMIN, 	os_flog,		H_NIDENTIFY	},
	    { "JUPE", 		SA_SOP, 	os_jupe,		H_NIDENTIFY	},
	    { "MDEOP", 		SA_SOP,	os_mdeop,		H_NIDENTIFY	},
	    { "MKICK",	 	SA_SOP,	os_mkick,		H_NIDENTIFY	},
	    { "READONLY", 	SA_SOP, 	os_readonly,	H_NIDENTIFY	},
	    { "RAW", 		SA_SOP, 	os_raw,		H_NIDENTIFY	},
	    { "RESTART", 		SA_SOP, 	os_restart,		H_NIDENTIFY	},
	    { "ROUTE", 		SA_SOP, 	os_route,		H_NIDENTIFY	},
	    { "SCHAN",	 	SA_SR, 	os_schan,		H_NIDENTIFY	},
	    { "QUIT", 		SA_SR, 	os_quit,		H_NIDENTIFY	},
	    { "SHUTDOWN", 	SA_SR, 	os_shutdown,	H_NIDENTIFY	},
	    { "SINVITE", 		SA_SR, 	os_sinvite,		H_NIDENTIFY	},
	    { "NOOP", 		SA_SR, 	os_noop,		H_NIDENTIFY	},
	    { "REHASHCONF", 	SA_SR, 	os_rehashconf,	H_NIDENTIFY	},
	    { "UPDATE", 		SA_SR, 	os_update,		H_NIDENTIFY	},
	    { "ACTIVATE", 	SA_SR, 	os_activate,	H_NIDENTIFY	},
	    { "POSITION",		SA_SR, 	os_position,	H_NIDENTIFY	},
	    { "SET", 		SA_SR, 	os_set,		H_NIDENTIFY	},
	    { NULL }
	};
	if (os = do_hash(source, cmd, os_hash_table)) {
	   switch (get_admin_level(source)) {
	       case SA_OPER:	strscpy(slvl, "IRC Operator", 32);			break;
	       case SA_ADMIN:	strscpy(slvl, "Services Administrator", 32);	break;
	       case SA_SOP:	strscpy(slvl, "Services Operator", 32);		break;
	       case SA_SR:	strscpy(slvl, "Services Root", 32);			break;
		 default:										break;
	   }
	   if (os->level == SA_ADMIN && !is_services_admin(source)) {
	       notice(s_OperServ, source, rpl_numeric_lang(source, CR_NO_ACCESS_LEVEL), slvl, "Services Administrator.");
	   } else if (os->level == SA_SOP && !is_services_op(source)) {
	       notice(s_OperServ, source, rpl_numeric_lang(source, CR_NO_ACCESS_LEVEL), slvl, "Services Operator.");
	   } else if (os->level >= SA_SR && !is_services_root(source)) {
	       notice(s_OperServ, source, rpl_numeric_lang(source, CR_NO_ACCESS_LEVEL), slvl, "Services Root.");
	   } else {
		(*os->func) (source);
	   }		
	} else {
	   notice(s_OperServ, source, rpl_numeric_lang(source, CR_CMD_UNKNOWN), cmd, s_OperServ);
      }
   }
}

/*************************************************************************/

/* os_position()
 * Set various position titles on a nick which are replied in a info nick command on NickServ. */

void os_position(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *param = strtok(NULL, "");
    NickInfo *ni;
    AdminInfo *ai;

    if (run.flags & RUN_READONLY) {
	notice(s_OperServ, source, rpl_numeric_lang(source, CR_CHANGES_NO_SAVE_RDONLY));
    }
    if (!nick || !param) {
	notice(s_OperServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "POSITION <nick> [CEO|COO|CAO|CTO|CRO|EB|NONE]");
	notice(s_OperServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_OperServ, "HELP POSITION");
    } else if (!(ni = findnick(nick))) {
	notice(s_OperServ, source, rpl_numeric_lang(source, CR_NICK_NOT_REGISTERED), nick);
    } else if (!(ai = findadmin(nick))) {
	notice(s_OperServ, source, "Nick \2%s\2 is not on the administration list.", nick);
    } else if (stricmp(param, "CEO") == 0) {
	ai->flags |= AI_P_CEO;
	ai->flags &= ~(AI_P_COO | AI_P_CAO | AI_P_CTO | AI_P_CRO | AI_P_EB);
	notice(s_OperServ, source, "Nick \2%s\2 now has the position Chief Exective Officer (CEO) assigned to their nick.", nick);
    } else if (stricmp(param, "COO") == 0) {
	ai->flags |= AI_P_COO;
	ai->flags &= ~(AI_P_CEO | AI_P_CAO | AI_P_CTO | AI_P_CRO | AI_P_EB);
	notice(s_OperServ, source, "Nick \2%s\2 now has the position Chief Operations Officer (COO) assigned to their nick.", nick);
    } else if (stricmp(param, "CAO") == 0) {
	ai->flags |= AI_P_CAO;
	ai->flags &= ~(AI_P_CEO | AI_P_COO | AI_P_CTO | AI_P_CRO | AI_P_EB);
	notice(s_OperServ, source, "Nick \2%s\2 now has the position Chief Administrative Officer (CAO) assigned to their nick.", nick);
    } else if (stricmp(param, "CTO") == 0) {
	ai->flags |= AI_P_CTO;
	ai->flags &= ~(AI_P_CEO | AI_P_COO | AI_P_CAO | AI_P_CRO | AI_P_EB);
	notice(s_OperServ, source, "Nick \2%s\2 now has the position Chief Technology Officer (CTO) assigned to their nick.", nick);
    } else if (stricmp(param, "CRO") == 0) {
	ai->flags |= AI_P_CRO;
	ai->flags &= ~(AI_P_CEO | AI_P_COO | AI_P_CAO | AI_P_CTO | AI_P_EB);
	notice(s_OperServ, source, "Nick \2%s\2 now has the position Chief Recruiting Officer (CRO) assigned to their nick.", nick);
    } else if (stricmp(param, "EB") == 0) {
	ai->flags |= AI_P_EB;
	ai->flags &= ~(AI_P_CEO | AI_P_COO | AI_P_CAO | AI_P_CTO | AI_P_CRO);
	notice(s_OperServ, source, "Nick \2%s\2 now has the position Execuitive Board (EB) assigned to their nick.", nick);
    } else if (stricmp(param, "NONE") == 0) {
	ai->flags &= ~(AI_P_CEO | AI_P_COO | AI_P_CAO | AI_P_CTO | AI_P_CRO | AI_P_EB);
	notice(s_OperServ, source, "Nick \2%s\2 now has no position assigned to their nick.", nick);
    }
}

/*************************************************************************/

/* os_flog()
 * Find in the log file a string matching the pattern provided. */

static void os_flog(const char *source)
{
   char *pattern = strtok(NULL, "");
   FILE *f = fopen(conf.logfile, "r");
   char buf[BUFSIZE];
   int i = 0;

   if (!pattern) {
	notice(s_OperServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "FLOG <*pattern*>");
	notice(s_OperServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_OperServ, "HELP FLOG");
   } else {
	if (f) {
		notice(s_OperServ, source, "Searching for \2%s\2 in the log file, please wait...", pattern);
      	while (fgets(buf, sizeof(buf), f)) {
      	   buf[strlen(buf) - 1] = 0;
		   if (match_wild_nocase(pattern, buf)) {
			i++;
			notice(s_OperServ, source, "%3d %s", i, buf);
	         }
            }
		fclose(f);
	} else {
		i++;
		notice(s_OperServ, source, "Could not open log file, search cancelled.");
	}
   }
   if (!i) {
	notice(s_OperServ, source, "Searched log file and found no results matching the pattern provided.");
   }
}

/*************************************************************************/

/* os_set()
 * Set various administration parameters, that affect an admin themselve or the functionality
 * of DS. */

static void os_set(const char *source)
{
    char *cmd = strtok(NULL, " ");
    char *param = strtok(NULL, "");
    NickInfo *ni;
    AdminInfo *ai;

    if (run.flags & RUN_READONLY) {
	notice(s_MemoServ, source, rpl_numeric_lang(source, CR_CHANGES_NO_SAVE_RDONLY));
    }
    if (!param) {
	notice(s_OperServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "SET <option> <parameters>");
	notice(s_OperServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_OperServ, "HELP SET");
    } else if (!(ni = findnick(source))) {
	notice(s_OperServ, source, rpl_numeric_lang(source, CR_YOUR_NICK_NOT_REGISTERED));
    } else if (!(ai = findadmin(ni->nick))) {
	notice(s_OperServ, source, "Your nick is not on the administration list.");
    } else if (stricmp(cmd, "AOP") == 0) {
	os_set_aop(ai, ni->nick, param);
    } else {
	 notice(s_OperServ, source, rpl_numeric_lang(source, CR_CMD_SET_UNKNOWN), cmd, s_OperServ);
    }
}

/*************************************************************************/

static void os_set_aop(AdminInfo *ai, const char *source, char *param)
{
    if (stricmp(param, "SERVER") == 0) {
	notice(s_OperServ, source, "The server \2%s\2 will now automatically grant you channel operator status in channels you have no access.",
		conf.lserver);
	ai->aop = AI_AOP_SERVER;
    } else if (stricmp(param, s_OperServ) == 0) {
	notice(s_OperServ, source, "\2%s\2 will now automatically grant you channel operator status in channels you have no access.",
		s_OperServ);
	ai->aop = AI_AOP_OS;
    } else if (stricmp(param, s_ChanServ) == 0) {
	notice(s_OperServ, source, "\2%s\2 will now automatically grant you channel operator status in channels you have no access.",
		s_ChanServ);
	ai->aop = AI_AOP_CS;
    } else {
	notice(s_OperServ, source, rpl_numeric_lang(source, CR_CMD_SYNTAX), "SET AOP [SERVER|%s|%s]", s_OperServ, s_ChanServ);
	notice(s_OperServ, source, rpl_numeric_lang(source, CR_CMD_MORE_INFO), s_MemoServ, "SET AOP");
    }
}

/*************************************************************************/

/* os_list_stats_cn()
 * List the /stats c and n commands. */

void os_list_stats_cn(const char *source)
{
   RouteInfo *ri;
   int i;

   for (i = 0; i < 256; i++) {
      for (ri = routelists[i]; ri; ri = ri->next) {
         send_cmd(NULL, "213 %s :C *@%s * %s %d 51", source, ri->server, ri->server, ri->port);
         send_cmd(NULL, "214 %s :N *@%s * %s 0 51", source, ri->server, ri->server);
      }
   }
}

/*************************************************************************/

/* os_list_stats_k()
 * List the /stats k command. */

void os_list_stats_k(const char *source)
{
   char *host;
   char user[256];
   int i;

   for (i = 0; i < nakill; i++) {
    strscpy(user, akills[i].mask, 256);
      if (host = strchr(user, '@')) {
         *host++ = 0;
         strlower(host);
      }
      send_cmd(NULL, "216 %s :A %s * %s 0 -1", source, host, user);
   }
}

/*************************************************************************/

/* os_list_stats_o()
 * List the /stats o command. */

void os_list_stats_o(const char *source)
{
   send_cmd(NULL, "243 %s :O %s@%s * %s 4293853183 2", source, conf.csident, conf.cshost, conf.csnick);
   send_cmd(NULL, "243 %s :O %s@%s * %s 4293853183 2", source, conf.nsident, conf.nshost, conf.nsnick);
   send_cmd(NULL, "243 %s :O %s@%s * %s 4293853183 2", source, conf.msident, conf.mshost, conf.msnick);
   send_cmd(NULL, "243 %s :O %s@%s * %s 4293853183 2", source, conf.osident, conf.oshost, conf.osnick);
   send_cmd(NULL, "243 %s :O %s@%s * %s 4293853183 2", source, conf.hsident, conf.hshost, conf.hsnick);
}

/*************************************************************************/

/* os_update()
 * Update (save) the database files. */

static void os_update(const char *source)
{
   globops(NULL, "\2%s\2 is forcibly saving (update) the datbases.", source);
   run.flags |= RUN_SAVING;
   notice(s_OperServ, source, "Update of databases is complete.");
}

/*************************************************************************/

/* os_rehashconf()
 * Rehash (reload) the services configuration file. */

static void os_rehashconf(const char *source)
{
   rehash_conf = 1;
   globops(NULL, "\2%s\2 is rehashing (rehashconf) the configuration file.", source);
   load_conf(CONF_FILE);
   setup_var_from_conf();
   notice(s_OperServ, source, "Rehash of services configuration file complete.");
}

/*************************************************************************/

/* os_quit()
 * Quit the servics program. */

static void os_quit(const char *source)
{
   quitmsg = malloc(28 + strlen(source));
   if (!quitmsg)
	quitmsg = "Attempted to quit services, but out of memory.";
   else
	sprintf(quitmsg, "\2%s\2 is quitting services.", source);
   run.flags &= ~RUN_LOOP;
   run.flags |= RUN_QUIT;
}

/*************************************************************************/

/* os_shutdown()
 * Shutdown the servics program. */

static void os_shutdown(const char *source)
{
   quitmsg = malloc(32 + strlen(source));
   if (!quitmsg)
      quitmsg = "Attempted to shutdown services, but out of memory.";
   else
      sprintf(quitmsg, "\2%s\2 is shutting down services.", source);

   /* Expire, save and update required elements, set flags and shutdown. */
	 expire_chans();
	 expire_nicks();

	 os_save_admin_db();
	 os_save_route_db();
	 save_akill();
	 save_ms_dbase();
	 save_cs_dbase();
	 save_ns_dbase();

   run.flags &= ~RUN_LOOP;
   run.flags |= RUN_QUIT;
}

/*************************************************************************/

/* os_restart()
 * Restart the services program. */

static void os_restart(const char *source)
{
#ifdef SERVICES_BIN
	quitmsg = malloc(31 + strlen(source));
	if (!quitmsg)
	    quitmsg = "Attempted to restart services, but out of memory.";
	else
	    sprintf(quitmsg, "\2%s\2 is restarting services.", source);
	 os_save_admin_db();
	 os_save_route_db();
	 save_akill();
	 save_ms_dbase();
	 save_cs_dbase();
	 save_ns_dbase();

   run.flags |= RUN_RESTART;
   run.flags &= ~RUN_LOOP;
#else
	notice(s_OperServ, source,
		"SERVICES_BIN not defined; cannot restart.  Rerun the "
		"\2configure\2 script and recompile Services to enable "
		"the RESTART command.");
#endif
}

/*************************************************************************/

/* os_load_db()
 * Load the OperServ services administration database. */

void os_load_admin_db(void)
{
    FILE *f = fopen(conf.admindb, "r");
    int i;
    AdminInfo *ai;

    if (!f) {
       return;
    }
    
    for (i = 33; i < 256; i++) {
	    while (fgetc(f) == 1) {
		ai = smalloc(sizeof(AdminInfo));
		if (1 != fread(ai, sizeof(AdminInfo), 1, f))
		    fatal_perror("Read error on %s", conf.admindb);
		alpha_insert_admin(ai);
	    }
	}
    fclose(f);
}

/*************************************************************************/

/* os_save_admin_db()
 * Save the OperServ services administration database. */

void os_save_admin_db(void)
{
    FILE *f = fopen(conf.admindb, "w");
    int i, j;
    AdminInfo *ai;

    if (!f) {
       return;
    }

    for (i = 33; i < 256; i++) {
	for (ai = adminlists[i]; ai; ai = ai->next) {
	    fputc(1, f);
	    if (1 != fwrite(ai, sizeof(AdminInfo), 1, f)) {
            fatal_perror("Read error on %s", conf.admindb);
          }
	}
	fputc(0, f);
    }
    fclose(f);
}

/*************************************************************************/

/* is_services_root_su()
 * Is the nick a services root su? */

int is_services_root_su(const char *nick)
{
   NickInfo *ni;

   /* Since they do not have the flag set, there is no point even checking, so return C_FALSE. */
#ifdef OS_A
   if (!is_sadmin(nick))
	   return SA_OPER;
#endif
   if (ni = findnick(conf.osroot)) {
         if ((stricmp(conf.osroot, nick) == 0) && (ni->flags & NI_IDENTIFIED) && is_oper(nick)) {
            return SA_HSR;
         }
   }
   return SA_OPER;
}

/*************************************************************************/

/* is_services_root()
 * Is the nick a services root or higher? */

int is_services_root(const char *nick)
{
   AdminInfo *ai;
   NickInfo *ni;
   int i, gotnick = 0;

   crash_id = 91;

   /* Since they do not have the flag set, there is no point even checking, so return C_FALSE. */
#ifdef OS_A
   if (!is_sadmin(nick))
	   return SA_OPER;
#endif
  for (i = 0; i < 256; i++) {
   for (ai = adminlists[i]; ai; ai = ai->next) {
      if (ni = findnick(ai->nick)) {
         if (stricmp(ai->nick, nick) == 0 && ai->level >= SA_SR && (ni->flags & NI_IDENTIFIED) && is_oper(nick)) {
            gotnick++;
            return (ai->level);
         }
      }
   }
  }
 if (!gotnick) {
      if (ni = findnick(conf.osroot)) {
         if ((stricmp(conf.osroot, nick) == 0) && (ni->flags & NI_IDENTIFIED) && is_oper(nick)) {
            return SA_HSR;
         }
      }
   }
   return SA_OPER;
}

/*************************************************************************/

/* is_services_admin()
 * Is the nick a services administrator or higher? */

int is_services_admin(const char *nick)
{
   AdminInfo *ai;
   NickInfo *ni;
   int i, gotnick = 0;

   crash_id = 92;

   /* Since they do not have the flag set, there is no point even checking, so return C_FALSE. */
#ifdef OS_A
   if (!is_sadmin(nick))
      return SA_OPER;
#endif
  for (i = 0; i < 256; i++) {
   for (ai = adminlists[i]; ai; ai = ai->next) {
      if (ni = findnick(ai->nick)) {
         if (stricmp(ai->nick, nick) == 0 && ai->level >= SA_ADMIN && (ni->flags & NI_IDENTIFIED) && is_oper(nick)) {
            gotnick++;
            return (ai->level);
         }
      }
   }
  }
   if (!gotnick) {
      if (ni = findnick(conf.osroot)) {
         if ((stricmp(conf.osroot, nick) == 0) && (ni->flags & NI_IDENTIFIED) && is_oper(nick)) {
            return SA_HSR;
         }
      }
   }
   return SA_OPER;
}

/*************************************************************************/

/* is_serv_admin()
 * Is the nick a services administrator at the defined level,
 * without the idents. */

int is_serv_admin(const char *nick, const int level)
{
   AdminInfo *ai;
   NickInfo *ni;
   int i, gotnick = 0;

   for (ai = adminlists[tolower(*nick)]; ai; ai = ai->next)
      if (ni = findnick(ai->nick)) {
         if (stricmp(ai->nick, nick) == 0 && ai->level == level) {
            gotnick++;
            return C_TRUE;
         }
      }
   if (!gotnick) {
      if (ni = findnick(conf.osroot)) {
         if (stricmp(conf.osroot, nick) == 0 && level >= SA_SR) {
            return C_TRUE;
         }
      }
   }
   return C_FALSE;
}

/*************************************************************************/

/* get_admin_level()
 * Get the level of the admin nick. */

int get_admin_level(const char *nick)
{
   AdminInfo *ai;
   NickInfo *ni;
   User *u = finduser(nick);
   int i, gotnick = 0;

   if (!u || !is_oper(nick)) return SA_NONE;

   /* Since they do not have the flag set, there is no point even checking, so return C_FALSE. */
#ifdef OS_A
   if (!is_sadmin(nick))
      return SA_OPER;
#endif
  for (i = 0; i < 256; i++) {
   for (ai = adminlists[i]; ai; ai = ai->next) {
      if (ni = findnick(ai->nick)) {
         if (stricmp(ai->nick, nick) == 0 && (ni->flags & NI_IDENTIFIED) && is_oper(nick)) {
            gotnick++;
            return ai->level;
         }
      }
   }
  }
   if (!gotnick) {
      if (ni = findnick(conf.osroot)) {
         if (stricmp(conf.osroot, nick) == 0 && (ni->flags & NI_IDENTIFIED) && is_oper(nick)) {
            return SA_SR;
         }
      }
   }
   return SA_OPER;
}

/*************************************************************************/

/* is_services_op()
 * Is the nick a services operator or higher? */

int is_services_op(const char *nick)
{
   AdminInfo *ai;
   NickInfo *ni;
   int i, gotnick = 0;

   /* Since they do not have the flag set, there is no point even checking, so return C_FALSE. */
#ifdef OS_A
   if (!is_sadmin(nick))
	return SA_OPER;
#endif
  for (i = 0; i < 256; i++) {
   for (ai = adminlists[i]; ai; ai = ai->next) {
      if (ni = findnick(ai->nick)) {
         if (stricmp(ai->nick, nick) == 0 && ai->level >= SA_SOP && (ni->flags & NI_IDENTIFIED) && is_oper(nick)) {
            gotnick++;
            return (ai->level);
         }
      }
   }
  }
   if (!gotnick) {
      if (ni = findnick(conf.osroot)) {
         if ((stricmp(conf.osroot, nick) == 0) && (ni->flags & NI_IDENTIFIED) && is_oper(nick)) {
            return SA_HSR;
         }
      }
   }
   return SA_OPER;
}

/*************************************************************************/

static void alpha_insert_admin(AdminInfo *ai)
{
   AdminInfo *ptr, *prev;
   char *nick = ai->nick;

   for (prev = NULL, ptr = adminlists[tolower(*nick)]; ptr && strcmp(ptr->nick, nick) < 0; prev = ptr, ptr = ptr->next)
   ;
   ai->prev = prev;
   ai->next = ptr;
   if (!prev)
      adminlists[tolower(*nick)] = ai;
   else
      prev->next = ai;
   if (ptr)
      ptr->prev = ai;
}

/*************************************************************************/

static AdminInfo *add_admin(const char *nick)
{
   AdminInfo *ai;

   ai = scalloc(sizeof(AdminInfo), 1);
   strscpy(ai->nick, nick, NICKMAX);
   alpha_insert_admin(ai);
   return (ai);
}

/*************************************************************************/

AdminInfo *findadmin(char *nick)
{
   AdminInfo *ai;
   int i;

   for (ai = adminlists[tolower(*nick)]; ai; ai = ai->next)
      if (stricmp(ai->nick, nick) == 0)
         return (ai);
   return NULL;
}

/*************************************************************************/

int deladmin(AdminInfo *ai)
{
    int i;

    if (ai->next)
	ai->next->prev = ai->prev;
    if (ai->prev)
	ai->prev->next = ai->next;
    else
	adminlists[tolower(*ai->nick)] = ai->next;
    free(ai);
    return C_TRUE;
}

/*************************************************************************/

int delroute(RouteInfo *ri)
{
    int i;

    if (ri->next)
	ri->next->prev = ri->prev;
    if (ri->prev)
	ri->prev->next = ri->next;
    else
	routelists[tolower(*ri->server)] = ri->next;
    free(ri);
    return C_TRUE;
}

/*************************************************************************/

/* delclone()
 * Remove a host from the clone list strucure, happens when we have only one client using this host. */

int delclone(CloneList *cl)
{
    int i;

    if (cl->next)
	cl->next->prev = cl->prev;
    if (cl->prev)
	cl->prev->next = cl->next;
    else
	clonelists[tolower(*cl->host)] = cl->next;
    free(cl);
    return C_TRUE;
}

/*************************************************************************/

static void alpha_insert_clone(CloneList *cl)
{
   CloneList *ptr, *prev;
   char *host = cl->host;

   for (prev = NULL, ptr = clonelists[tolower(*host)]; ptr && strcmp(ptr->host, host) < 0; prev = ptr, ptr = ptr->next)
   ;
   cl->prev = prev;
   cl->next = ptr;
   if (!prev)
      clonelists[tolower(*host)] = cl;
   else
      prev->next = cl;
   if (ptr)
      ptr->prev = cl;
}

/*************************************************************************/

static void alpha_insert_route(RouteInfo *ri)
{
   RouteInfo *ptr, *prev;
   char *server = ri->server;

   for (prev = NULL, ptr = routelists[tolower(*server)]; ptr && strcmp(ptr->server, server) < 0; prev = ptr, ptr = ptr->next)
   ;
   ri->prev = prev;
   ri->next = ptr;
   if (!prev)
      routelists[tolower(*server)] = ri;
   else
      prev->next = ri;
   if (ptr)
      ptr->prev = ri;
}

/*************************************************************************/

static CloneList *add_clone(const char *host)
{
   CloneList *cl;

   cl = scalloc(sizeof(CloneList), 1);
   strscpy(cl->host, host, HOSTMAX);
   alpha_insert_clone(cl);
   return (cl);
}

/*************************************************************************/

CloneList *findclone(char *host)
{
   CloneList *cl;
   int i;

   for (cl = clonelists[tolower(*host)]; cl; cl = cl->next)
      if (stricmp(cl->host, host) == 0)
         return (cl);
   return NULL;
}

/*************************************************************************/

static RouteInfo *add_route(const char *server)
{
   RouteInfo *ri;

   ri = scalloc(sizeof(RouteInfo), 1);
   strscpy(ri->server, server, HOSTMAX);
   alpha_insert_route(ri);
   return (ri);
}

/*************************************************************************/

RouteInfo *findroute(char *server)
{
   RouteInfo *ri;
   int i;

   for (ri = routelists[tolower(*server)]; ri; ri = ri->next)
      if (stricmp(ri->server, server) == 0)
         return (ri);
   return NULL;
}

/*************************************************************************/

void load_akill(void)
{
    FILE *f = fopen(conf.akilldb, "r");
    int i;

    if (!f) {
	return;
    }
	nakill = fgetc(f) * 256 + fgetc(f);
	if (nakill < 8)
	    akill_size = 16;
	else
	    akill_size = 2*nakill;
	akills = smalloc(sizeof(*akills) * akill_size);
	if (!nakill) {
	    fclose(f);
	    return;
	}
	if (nakill != fread(akills, sizeof(*akills), nakill, f))
	    fatal_perror("Read error on %s", conf.akilldb);
	for (i = 0; i < nakill; i++) {
	    akills[i].mask = read_string(f, conf.akilldb);
	    akills[i].reason = read_string(f, conf.akilldb);
	}
    fclose(f);
}

/*************************************************************************/

void save_akill(void)
{
    FILE *f = fopen(conf.akilldb, "w");
    int i;

    if (!f) {
       return;
    }
    fputc(nakill/256, f);
    fputc(nakill & 255, f);
    if (fwrite(akills, sizeof(*akills), nakill, f) != nakill)
	fatal_perror("Write error on %s", conf.akilldb);
    for (i = 0; i < nakill; i++) {
	write_string(akills[i].mask, f, conf.akilldb);
	write_string(akills[i].reason, f, conf.akilldb);
    }
    fclose(f);
}

/*************************************************************************/

/* os_load_route_db()
 * Load the OperServ routing database. */

void os_load_route_db(void)
{
    FILE *f = fopen(conf.routedb, "r");
    int i;
    RouteInfo *ri;

    if (!f) {
       return;
    }
    
    for (i = 33; i < 256; i++) {
	    while (fgetc(f) == 1) {
		ri = smalloc(sizeof(RouteInfo));
		if (1 != fread(ri, sizeof(RouteInfo), 1, f))
		    fatal_perror("Read error on %s", conf.routedb);
		alpha_insert_route(ri);
	    }
	}
    fclose(f);
}

/*************************************************************************/

/* os_save_route_db()
 * Save the OperServ routing database. */

void os_save_route_db(void)
{
    FILE *f = fopen(conf.routedb, "w");
    int i, j;
    RouteInfo *ri;

    if (!f) {
       return;
    }

    for (i = 33; i < 256; i++) {
	for (ri = routelists[i]; ri; ri = ri->next) {
	    fputc(1, f);
	    if (1 != fwrite(ri, sizeof(RouteInfo), 1, f)) {
            fatal_perror("Read error on %s", conf.routedb);
          }
	}
	fputc(0, f);
    }
    fclose(f);
}

/*************************************************************************/

/* check_akill()
 * User connecting, does their user@host match any Autokills currently set?
 * If so kill them and then set the Autokill again. */

int check_akill(const char *nick, const char *username, const char *host)
{
    char buf[512];
    int i;
    char *host2, *username2;
    strscpy(buf, username, sizeof(buf)-2);
    i = strlen(buf);
    buf[i++] = '@';
    strlower(strscpy(buf+i, host, sizeof(buf)-i));
    for (i = 0; i < nakill; i++) {
	if (match_wild_nocase(akills[i].mask, buf) && !(akills[i].flags & AK_BLOCK)) {
	    /* We have two possible messages for the kill and Autokill:
           * a) You are banned from this network
           * b) User has been banned */
            send_cmd(s_OperServ, "KILL %s :%s (User has been banned (%s. %s))",
                        nick, s_OperServ, akills[i].reason, conf.osakillmsg);
	    username2 = sstrdup(akills[i].mask);
	    host2 = strchr(username2, '@');
            if (!host2) {
		globops(NULL, "Autokill missing '@' in mask, \2%s\2.", akills[i].mask);
		continue;
	    }
	    *host2++ = 0;
#if defined(V_BAHAMUT)
	      send_cmd(conf.lserver, "AKILL %s %s %d %s %ld :User has been banned. (%s)",
               host2, username2, 86400*2, akills[i].who, akills[i].time, akills[i].reason);
#elif defined(V_DF) || (V_UNREAL)
	    send_cmd(conf.lserver, "AKILL %s %s :User has been banned (%s)",
             host2, username2, akills[i].reason);
#else
	    time_t now = time(NULL);
	    send_cmd(conf.lserver, "GLINE * +%ld %s@%s :User has been banned. (%s)",
		    akills[i].expires && (akills[i].expires > now) ? (akills[i].expires - time(NULL)) : 999999999,
		    username2, host2, akills[i].reason);
#endif
	    free(username2);
	    return C_TRUE;
	}
    }
    return C_FALSE;
}

/*************************************************************************/

/* set_akill()
 * When DS sucessfully connects to the network, set all the Autokills to all
 * the servers, incase they were rehashed or something. */

void set_akill(void)
{
    char buf[512];
    int i;
    char *host2, *username2;
    i = strlen(buf);
    buf[i++] = '@';
    for (i = 0; i < nakill; i++) {
	if (akills[i].flags & AK_AKILL) {
	    username2 = sstrdup(akills[i].mask);
	    host2 = strchr(username2, '@');
            if (!host2) {
		globops(NULL, "Autokill missing '@' in mask, \2%s\2.", akills[i].mask);
		continue;
	    }
	    *host2++ = 0;

#if defined(V_BAHAMUT)
	      send_cmd(conf.lserver, "AKILL %s %s %d %s %ld :User has been banned. (%s)",
               host2, username2, 86400*2, akills[i].who, akills[i].time, akills[i].reason);
#elif defined(V_DF) || (V_UNREAL)
	    send_cmd(conf.lserver, "AKILL %s %s :User has been banned (%s)",
             host2, username2, akills[i].reason);
#else
	    time_t now = time(NULL);
	    send_cmd(conf.lserver, "GLINE * +%ld %s@%s :User has been banned. (%s)",
		    akills[i].expires && (akills[i].expires > now) ? (akills[i].expires - time(NULL)) : 999999999,
		    username2, host2, akills[i].reason);
#endif
	    free(username2);
	}
    }
}

/*************************************************************************/

/* expire_akills()
 * Expire any Autokills which have reached their timeout, due to our timing
 * Autokills will actually last for 2-5 seconds longer then they actually are. */

void expire_akills(void)
{
    int i;
    time_t now = time(NULL);
    char *s;

    for (i = 0; i < nakill; i++) {
	if (akills[i].expires == 0 || akills[i].expires > now)
	    continue;
	if (s = strchr(akills[i].mask, '@')) {
	    *s++ = 0;
	    strlower(s);
	    send_cmd(conf.lserver, "RAKILL %s %s", s, akills[i].mask);
	    if (akills[i].flags & AK_BLOCK)
	          globops(NULL, "Removing timed Autokill block for \2%s@%s\2.", akills[i].mask, s);
	    else
	          globops(NULL, "Removing timed autokill for \2%s@%s\2.", akills[i].mask, s);
	}
	free(akills[i].mask);
	free(akills[i].reason);
	nakill--;
	if (i < nakill)
	    bcopy(akills + i + 1, akills + i, sizeof(*akills) * (nakill - i));
	i--;
    }
}

/*************************************************************************/

/* add_akill()
 * Note that all parameters except expiry are assumed to be non NULL.  A
 * value of NULL for expiry indicates that the AKILL should not expire. */

static void add_akill(const char *mask, const char *reason, const char *who, const char *expiry, unsigned long tracking, const int block)
{
    char *host;
    char user[256];
    int amount = 0, tamount = 0;

    if (nakill >= akill_size) {
	if (akill_size < 8)
	    akill_size = 8;
	else
	    akill_size *= 2;
	akills = srealloc(akills, sizeof(*akills) * akill_size);
    }
    akills[nakill].mask = sstrdup(mask);
    akills[nakill].reason = sstrdup(reason);
    akills[nakill].time = time(NULL);
    strscpy(akills[nakill].who, who, NICKMAX);
    akills[nakill].tracking = tracking;
    strscpy(user, mask, 256);

    if (is_services_root(who))	stats.sroot_akills++;
    else if (is_services_op(who))	stats.csop_akills++;
    else					stats.sadmin_akills++;

    if (block == C_TRUE) {
		akills[nakill].flags |= AK_BLOCK;
		akills[nakill].flags &= ~AK_AKILL;
    }
    if (block == C_FALSE) {
		akills[nakill].flags |= AK_AKILL;
		akills[nakill].flags &= ~AK_BLOCK;
    }

    if (host = strchr(user, '@')) {
       *host++ = 0;
       strlower(host);
    }
	amount = strtol(expiry, (char **)&expiry, 10);
      tamount = amount;
	if (amount == 0) {
	    akills[nakill].expires = 0;
	} else {
          amount *= 60;
	    akills[nakill].expires = amount + akills[nakill].time;
	}
    if (akills[nakill].flags & AK_AKILL) {
	    send_akill(who, user, host, reason, tamount, akills[nakill].expires);
    } else {
	if (tamount <= 0)
             globops(s_OperServ, "Autokill block for \2%s\2 by \2%s\2 (permanent), this mask can no longer be Autokilled.",
                mask, who);
	else
             globops(s_OperServ, "Autokill block for \2%s\2 by \2%s\2 for \2%d\2 minutes, this mask can not be Autokilled until it expires.",
                mask, who, tamount);
    }
    nakill++;
}

/*************************************************************************/

/* del_akill()
 * Delete the given Autokill from the target list, return whether or not it 
 * was found or not, and removed. */

static int del_akill(const char *mask)
{
    int i;

    for (i = 0; i < nakill && strcmp(akills[i].mask, mask) != 0; i++)
	;
    if (i < nakill) {
	free(akills[i].mask);
	free(akills[i].reason);
	nakill--;
	if (i < nakill)
	    bcopy(akills+i+1, akills+i, sizeof(*akills) * (nakill-i));
	return C_TRUE;
    } else {
	return C_FALSE;
    }
}

/*************************************************************************/

/* check_clones()
 * Check to see if the connecting user is a clone. */

void check_clones(User *user)
{
#ifdef CHECK_CLONES
   int i, clone_count;
   long last_time;

   if (clonelist[0].host)
      free(clonelist[0].host);
   i = CLONE_DETECT_SIZE-1;
   memmove(clonelist, clonelist+1, sizeof(struct clone) * i);
   clonelist[i].host = sstrdup(user->host);
   last_time = clonelist[i].time = time(NULL);
   clone_count = 1;

   while (--i >= 0 && clonelist[i].host) {
      if (clonelist[i].time < last_time - CLONE_MAX_DELAY)
         break;
      if (stricmp(clonelist[i].host, user->host) == 0) {
         ++clone_count;
         last_time = clonelist[i].time;
         if (clone_count >= CLONE_MIN_USERS)
            break;
      }
   }
   if (clone_count >= CLONE_MIN_USERS) {
      for (i = CLONE_DETECT_SIZE - 1; i >= 0 && warnings[i].host; --i) {
         if (stricmp(warnings[i].host, user->host) == 0)
            break;
      }
      if (i < 0 || warnings[i].time < user->signon - CLONE_WARNING_DELAY) {
         globops(s_OperServ, "Possible clones, \2%d\2 clients from \2%s\2.",
            clone_count, user->host);
         i = CLONE_DETECT_SIZE - 1;
         if (warnings[0].host)
            free(warnings[0].host);
         bcopy(warnings+1, warnings, sizeof(struct clone) * i);
         warnings[i].host = sstrdup(user->host);
         warnings[i].time = clonelist[i].time;
      }
   }
#endif
}

/*************************************************************************/

/* os_help()
 * Return help messages. */

static void os_help(const char *source)
{
   const char *cmd = strtok(NULL, " ");
   char *param;

   /* Handle no comman indexes first, outout all allowed commands for the current level. */
   if (!cmd) {
      notice_list(s_OperServ, source, os_serv_oper_help);
	if (is_services_admin(source))							notice_list(s_OperServ, source, os_serv_admin_help);
	if (is_services_op(source))								notice_list(s_OperServ, source, os_serv_op_help);
	if (is_services_root(source))								notice_list(s_OperServ, source, os_serv_root_help);
	notice_list(s_OperServ, source, os_serv_end_help);   
   } else if (stricmp(cmd, "STATS") == 0)							notice_list(s_OperServ, source, oso_stats_help);
   else if (stricmp(cmd, "KILL") == 0)							notice_list(s_OperServ, source, oso_kill_help);
   else if (stricmp(cmd, "KICK") == 0)							notice_list(s_OperServ, source, oso_kick_help);
   else if (stricmp(cmd, "ADMIN") == 0)							notice_list(s_OperServ, source, oso_admin_help);
   else if (stricmp(cmd, "MAP") == 0)							notice_list(s_OperServ, source, oso_map_help);
   else if (stricmp(cmd, "AKILL") == 0 && is_services_admin(source))		notice_list(s_OperServ, source, oso_akill_help);
   else if (stricmp(cmd, "GLOBAL") == 0 && is_services_admin(source))		notice_list(s_OperServ, source, oso_global_help);
   else if (stricmp(cmd, "MODE") == 0 && is_services_admin(source))		notice_list(s_OperServ, source, oso_mode_help);
   else if (stricmp(cmd, "SQLINE") == 0 && is_services_admin(source))		notice_list(s_OperServ, source, oso_sqline_help);
   else if (stricmp(cmd, "MDEOP") == 0 && is_services_op(source))			notice_list(s_OperServ, source, oso_mdeop_help);
   else if (stricmp(cmd, "MKICK") == 0 && is_services_op(source))			notice_list(s_OperServ, source, oso_mkick_help);
   else if (stricmp(cmd, "JUPE") == 0 && is_services_op(source))			notice_list(s_OperServ, source, oso_jupe_help);
   else if (stricmp(cmd, "RAW") == 0 && is_services_op(source))			notice_list(s_OperServ, source, oso_raw_help);
   else if (stricmp(cmd, "ROUTE") == 0 && is_services_op(source))			notice_list(s_OperServ, source, oso_route_help);
   else if (stricmp(cmd, "READONLY") == 0 && is_services_op(source))		notice_list(s_OperServ, source, oso_readonly_help);
   else if (stricmp(cmd, "RESTART") == 0 && is_services_op(source))		notice_list(s_OperServ, source, oso_restart_help);
   else if (stricmp(cmd, "QUIT") == 0 && is_services_root(source))		notice_list(s_OperServ, source, oso_quit_help);
   else if (stricmp(cmd, "NOOP") == 0 && is_services_root(source))		notice_list(s_OperServ, source, oso_noop_help);
   else if (stricmp(cmd, "UPDATE") == 0 && is_services_root(source))		notice_list(s_OperServ, source, oso_update_help);
   else if (stricmp(cmd, "SHUTDOWN") == 0 && is_services_root(source))		notice_list(s_OperServ, source, oso_shutdown_help);
   else if (stricmp(cmd, "SCHAN") == 0 && is_services_root(source))		notice_list(s_OperServ, source, oso_schan_help);
   else if (stricmp(cmd, "REHASHCONF") == 0 && is_services_root(source))	notice_list(s_OperServ, source, oso_rehashconf_help);
   else if (stricmp(cmd, "SINVITE") == 0 && is_services_root(source))		notice_list(s_OperServ, source, oso_sinvite_help);
   else if (stricmp(cmd, "ACTIVATE") == 0 && is_services_root(source))		notice_list(s_OperServ, source, oso_activate_help);
   else if (stricmp(cmd, "FLOG") == 0 && is_services_root(source)) {		notice_list(s_OperServ, source, oso_flog_help);
   } else if (stricmp(cmd, "SET") == 0 && is_services_root(source)) {
	notice(s_OperServ, source, "Sorry, there is no help available for the set command \2%s\2.", param);
   } else
	notice(s_OperServ, source, "Sorry, there is no help available for the command \2%s\2.", cmd);
}

/*************************************************************************/

/* os_admin()
 * Manipulate the services administrator list. */

static void os_admin(const char *source)
{
   char *cmd, *nick, *level;
   int i, lvl = 0, sacount = 1;
   AdminInfo *ai;
   NickInfo *ni;
   cmd = strtok(NULL, " ");

   if (!cmd)
      cmd = "";

    if (run.flags & RUN_READONLY) {
	notice(s_OperServ, source, "DS is currently in read only mode, changes will not be saved.");
   }
   if (stricmp(cmd, "ADD") == 0) {
      nick = strtok(NULL, " ");
      level = strtok(NULL, " ");
      if (!is_services_root(source)) {
         notice(s_OperServ, source, "Access denied.");
         return;
      }
      if (nick && level) {
	    if (!(ni = findnick(nick))) {
		notice(s_OperServ, source, "Nick \2%s\2 is not registered.", nick);
	    } else if (!(lvl = atoi(level))) {
		notice(s_OperServ, source, "Level must be an integer.");
	    } else if (lvl == SA_SR && !is_services_root_su(source)) {
	      notice(s_OperServ, source, "Access denied.");
          } else if (lvl >= SA_ADMIN && lvl <= SA_SR) {
             if (!(ai = findadmin(nick))) {
                if (ai = add_admin(ni->nick)) {
                   ai->level = atoi(level);
                   notice(s_OperServ, source, "\2%s\2 has been added to the administration list.", ai->nick);
                }
             } else {
                notice(s_OperServ, source, "\2%s\2 is already on the administration list.", ai->nick);
             }
          }	else {
		notice(s_OperServ, source, "Level must be an integer between %d and %d inclusive.", SA_ADMIN, SA_SR);
          }
	} else {
	    notice(s_OperServ, source, "Syntax: ADMIN ADD <nick> <level>");
          notice(s_OperServ, source, "For more information type \2/msg %s HELP ADMIN.", s_OperServ);
	}

    } else if (stricmp(cmd, "DEL") == 0) {
	if (!is_services_root(source)) {
	    notice(s_OperServ, source, "Access denied.");
	    return;
	}
      nick = strtok(NULL, " ");
	if (nick) {
        if (ni = findnick(nick)) {
          if (ai = findadmin(ni->nick)) {
                notice(s_OperServ, source, "\2%s\2 has been removed from the administration list.", 
                   ni->nick);
                deladmin(ai);
             } else {
                notice(s_OperServ, source, "\2%s\2 is not on the administration list.", 
                   ni->nick);
             }
         }
	} else {
	    notice(s_OperServ, source, "Syntax: ADMIN DEL <nick>");
          notice(s_OperServ, source, "For more information type \2/msg %s HELP ADMIN.", s_OperServ);
	}
    } else if (stricmp(cmd, "LIST") == 0) {
         notice(s_OperServ, source, "Current administration list:");
         notice(s_OperServ, source, "Num Level Nick");
         if (ni = findnick(conf.osroot)) {
           notice(s_OperServ, source, "  1     4 %s (%s)", ni->nick, ni->last_usermask);
         }
         for (i = 0; i < 256; i++) {
             for (ai = adminlists[i]; ai; ai = ai->next) {
                 if (ni = findnick(ai->nick)) {
                    sacount++;
                       notice(s_OperServ, source, "%3d %5d %s (%s)", 
                          sacount, ai->level, ai->nick, ni->last_usermask);
                }
             }
          }
    } else {
	notice(s_OperServ, source, "Syntax: ADMIN [ADD|DEL|LIST] <nick> <\37level\37>");
      notice(s_OperServ, source, "For more information type \2/msg %s HELP ADMIN.", s_OperServ);
    }
}

/*************************************************************************/

/* os_akill()
 * Autokill list manipulation. */

static void os_akill(const char *source)
{
   char *cmd, *scmd, *mask, *reason, *expiry, *s, *ss;
   int i, akcount = 1;
   unsigned long tracking;
#if HAVE_GETTIMEOFDAY
   struct timeval tv;
#endif   
   cmd = strtok(NULL, " ");
   if (!cmd)
      cmd = "";

    if (run.flags & RUN_READONLY) {
	notice(s_OperServ, source, "DS is currently in read only mode, changes will not be saved.");
   }
   if (stricmp(cmd, "ADD") == 0) {
      expiry = strtok(NULL, " ");
	mask = strtok(NULL, " ");
      reason = strtok(NULL, "");
	if (mask && reason && expiry) {
	    if (s = strchr(mask, '@')) {
		strlower(s);
	    } else {
		notice(s_OperServ, source, "Autokill must contain the `@' character.");
		return;
 	    }
          if (!is_services_root(source) && (stricmp(mask, "*@*") == 0 || stricmp(mask, "*@*.*") == 0 ||
               stricmp(mask, "*!*@*") == 0 || stricmp(mask, "*!*@*.*") == 0 || stricmp(mask, "*") == 0)) {
                  notice(s_OperServ, source, "Access denied.");
                  globops(s_OperServ, "\2%s\2 is attempting to add a network wide autokill for \2%s\2", source, mask);
			return;
           }
          if (!is_services_root(source) && strlen(mask) <= 7) {
             globops(s_OperServ, "\2%s\2 is attempting to add a top level domain Autokill for \2%s\2", source, mask);
             notice(s_OperServ, source, "Access denied.");
		 return;
          }

          for (i = 0; i < nakill; i++) {
             if (stricmp(mask, akills[i].mask) == 0) {
		   if (akills[i].flags & AK_BLOCK) {
 		      if (!is_services_root(source))
		            globops(s_OperServ, "\2%s\2 is attempting to add a Autokill matching \2%s\2 which is currrently blocked.",
					source, mask);
                  notice(s_OperServ, source, "\2%s\2 is already on the Autokill block list.", mask);
		   } else {
                  notice(s_OperServ, source, "\2%s\2 is already on the Autokill list.", mask);
		   }
  		   return;
 		 }
          }

#if HAVE_GETTIMEOFDAY
	   gettimeofday(&tv, NULL);
	   tracking = tv.tv_usec % 100001;
	   if (tracking < 0) tracking += 100001;
#else
	tracking = os_random(10000, 30000);
#endif
 	  add_akill(mask, reason, source, expiry, tracking, C_FALSE);

	    notice(s_OperServ, source, "\2%s\2 has been added to the autokill list, tracking number is \2%ld\2.", mask, tracking);
	    notice(s_OperServ, source, "Please email %s with this number as the subject and any other", conf.klineemail);
	    notice(s_OperServ, source, "information you may have in regards to this autokill in the body.");
	} else {
	    notice(s_OperServ, source, "Syntax: AKILL ADD <expiry> <user@host> <reason>");
          notice(s_OperServ, source, "For more information type \2/msg %s HELP AKILL.", s_OperServ);
	}
    } else if (stricmp(cmd, "BADD") == 0) {
      expiry = strtok(NULL, " ");
	mask = strtok(NULL, " ");
      if (!is_services_root(source)) {
		notice(s_OperServ, source, "Access denied.");
		return;
	}

	if (mask && expiry) {
	    if (s = strchr(mask, '@')) {
		strlower(s);
	    } else {
		notice(s_OperServ, source, "Autokill must contain the `@' character.");
		return;
 	    }

          if ((stricmp(mask, "*@*") == 0 || stricmp(mask, "*@*.*") == 0 || stricmp(mask, "*!*@*") == 0 ||
              stricmp(mask, "*!*@*.*") == 0 || stricmp(mask, "*") == 0 || strlen(mask) <= 7)) {
             notice(s_OperServ, source, "Network wide masks, top level domains and Autokills smaller then 7 characters"
			" in length are blocked by default.");
		 return;
           }

          for (i = 0; i < nakill; i++) {
             if (stricmp(mask, akills[i].mask) == 0) {
			if (akills[i].flags & AK_BLOCK)
	                notice(s_OperServ, source, "\2%s\2 is already on the Autokill block list.", mask);
			if (akills[i].flags & AK_AKILL) {
	                notice(s_OperServ, source, "\2%s\2 is already on the Autokill list, changed to blocked list.", mask);
				akills[i].flags &= ~AK_AKILL;
				akills[i].flags |= AK_BLOCK;
			}
			return;
 		}
          }

#if HAVE_GETTIMEOFDAY
	   gettimeofday(&tv, NULL);
	   tracking = tv.tv_usec % 100001;
	   if (tracking < 0) tracking += 100001;
#else
	tracking = os_random(10000, 30000);
#endif

 	  add_akill(mask, "blocked", source, expiry, tracking, C_TRUE);
	    notice(s_OperServ, source, "\2%s\2 has been added to the Autokill block list, tracking number is \2%ld\2.", mask, tracking);
	    notice(s_OperServ, source, "The mask noted above will now be blocked from being Autokill.  Keep in the mind", conf.klineemail);
	    notice(s_OperServ, source, "this mask is not a wildward match, only an exact match.");
	} else {
	    notice(s_OperServ, source, "Syntax: AKILL BADD <expiry> <user@host> <reason>");
          notice(s_OperServ, source, "For more information type \2/msg %s HELP AKILL.", s_OperServ);
	}
    } else if (stricmp(cmd, "BLIST") == 0) {
	s = strtok(NULL, " ");
	notice(s_OperServ, source, "Current Autokill block list:");
      notice(s_OperServ, source, "Num Tracking Mask");
	for (i = 0; i < nakill; i++) {
	  if (akills[i].flags & AK_BLOCK) {
          if (s) {
		if (stricmp(s, "ME") == 0) {
                if (stricmp(akills[i].who, source) == 0) {
  			    notice(s_OperServ, source, "%3d %8ld %-32s", akcount, akills[i].tracking, akills[i].mask);
			}
		} else {
		   if ((match_wild_nocase(s, akills[i].mask)) || (atoi(s) == akills[i].tracking)) {
  			    notice(s_OperServ, source, "%3d %8ld %-32s", akcount, akills[i].tracking, akills[i].mask);
               }
		}
          } else {
		    notice(s_OperServ, source, "%3d %8ld %-32s", akcount, akills[i].tracking, akills[i].mask);
          }
        }
     }
    } else if (stricmp(cmd, "DEL") == 0) {
	if (mask = strtok(NULL, " ")) {
	    if (s = strchr(mask, '@'))
		strlower(s);
	    if (del_akill(mask)) {
            globops(NULL, "\2%s\2 removed Autokill for \2%s\2.", source, mask);
		notice(s_OperServ, source, "\2%s\2 has been removed from the Autokill list.", mask);
		
		if (s) {
		    *s++ = 0;
		    send_cmd(conf.lserver, "RAKILL %s %s", s, mask);
		} 
	    } else {
		notice(s_OperServ, source, "\2%s\2 is not on the Autokill list.", mask);
	    }
	} else {
 	    notice(s_OperServ, source, "Syntax: AKILL DEL <user@host>");
          notice(s_OperServ, source, "For more information type \2/msg %s HELP AKILL.", s_OperServ);
	}
    } else if (stricmp(cmd, "BDEL") == 0) {
	if (!is_services_root(source)) {
		notice(s_OperServ, source, "Access denied.");
		return;
	}
	if (mask = strtok(NULL, " ")) {
	    if (s = strchr(mask, '@'))
		strlower(s);
	    if (del_akill(mask)) {
            globops(NULL, "\2%s\2 removed Autokill block for \2%s\2.", source, mask);
		notice(s_OperServ, source, "\2%s\2 has been removed from the Autokill block list.", mask);
		if (s) {
		    *s++ = 0;
		    send_cmd(conf.lserver, "RAKILL %s %s", s, mask);
		} 
	    } else {
		notice(s_OperServ, source, "\2%s\2 is not on the Autokill block list.", mask);
	    }
	} else {
 	    notice(s_OperServ, source, "Syntax: AKILL BDEL <user@host>");
          notice(s_OperServ, source, "For more information type \2/msg %s HELP AKILL.", s_OperServ);
	}

    } else if (stricmp(cmd, "LIST") == 0) {
	s = strtok(NULL, " ");
	notice(s_OperServ, source, "Current Autokill list:");
      notice(s_OperServ, source, "Num Tracking Mask");
	for (i = 0; i < nakill; i++) {
	  if (akills[i].flags & AK_AKILL) {
          if (s) {
		if (stricmp(s, "ME") == 0) {
                if (stricmp(akills[i].who, source) == 0) {
  			    notice(s_OperServ, source, "%3d %8ld %-32s (%s)", akcount, akills[i].tracking, akills[i].mask, akills[i].reason);
      	          akcount++;
			}
		} else {
		   if ((match_wild_nocase(s, akills[i].mask)) || (atoi(s) == akills[i].tracking)) {
  			    notice(s_OperServ, source, "%3d %8ld %-32s (%s)", akcount, akills[i].tracking, akills[i].mask, akills[i].reason);
      	          akcount++;
               }
		}
          } else {
		    notice(s_OperServ, source, "%3d %8ld %-32s (%s)", akcount, akills[i].tracking, akills[i].mask, akills[i].reason);
                akcount++;
          }
       }
     }
    } else if (stricmp(cmd, "VIEW") == 0) {
      s = strtok(NULL, " ");
	notice(s_OperServ, source, "Current Autokill list:");
      notice(s_OperServ, source, "Num Tracking Mask");
	for (i = 0; i < nakill; i++) {
       if (akills[i].flags & AK_AKILL) {
        if (s) {
	    if (stricmp(s, "ME") == 0) {
		if (stricmp(akills[i].who, source) == 0) {
  			char timebuf[32], expirebuf[64];
			struct tm tm;
			time_t t = time(NULL);

			tm = *localtime(akills[i].time ? &akills[i].time : &t);
			strftime(timebuf, sizeof(timebuf), "%d %b %Y", &tm);
			if (akills[i].expires == 0) {
			    strcpy(expirebuf, "permanent");
			} else {
			    time_t t2 = akills[i].expires - t;
			    t2 += 59;
				t2 /= 60;
				snprintf(expirebuf, sizeof(expirebuf), "expires in %d minute%s", 
	                    t2, t2==1 ? "" : "s");
			}
			notice(s_OperServ, source, "%3d %8ld %s (by %s on %s; %s)", akcount, akills[i].tracking, akills[i].mask,
      	         *akills[i].who ? akills[i].who : "<unknown>", timebuf, expirebuf);
			notice(s_OperServ, source, " %s", akills[i].reason);
      	      akcount++;
            } else {
		   if ((match_wild_nocase(s, akills[i].mask)) || (atoi(s) == akills[i].tracking)) {
  			char timebuf[32], expirebuf[64];
			struct tm tm;
			time_t t = time(NULL);

			tm = *localtime(akills[i].time ? &akills[i].time : &t);
			strftime(timebuf, sizeof(timebuf), "%d %b %Y", &tm);
			if (akills[i].expires == 0) {
			    strcpy(expirebuf, "permanent");
			} else {
			    time_t t2 = akills[i].expires - t;
			    t2 += 59;
				t2 /= 60;
				snprintf(expirebuf, sizeof(expirebuf), "expires in %d minute%s", 
	                    t2, t2==1 ? "" : "s");
			}
			notice(s_OperServ, source, "%3d %8ld %s (by %s on %s; %s)", akcount, akills[i].tracking, akills[i].mask,
      	         *akills[i].who ? akills[i].who : "<unknown>", timebuf, expirebuf);
			notice(s_OperServ, source, " %s", akills[i].reason);
      	      akcount++;
               }
		}
	    }
        } else {
		char timebuf[32], expirebuf[64];
		struct tm tm;
		time_t t = time(NULL);

		tm = *localtime(akills[i].time ? &akills[i].time : &t);
		strftime(timebuf, sizeof(timebuf), "%d %b %Y", &tm);

		if (akills[i].expires == 0) {
		    strcpy(expirebuf, "permanent");
		} else {
		    time_t t2 = akills[i].expires - t;
		    t2 += 59;
			t2 /= 60;
			snprintf(expirebuf, sizeof(expirebuf), "expires in %d minute%s", 
                     t2, t2==1 ? "" : "s");
		}

		notice(s_OperServ, source, "%3d %8ld %s (by %s on %s; %s)", akcount, akills[i].tracking, akills[i].mask,
               *akills[i].who ? akills[i].who : "<unknown>", timebuf, expirebuf);
		notice(s_OperServ, source, " %s", akills[i].reason);
            akcount++;
	  }
      }
     }


    } else if (stricmp(cmd, "BVIEW") == 0) {
      s = strtok(NULL, " ");
	notice(s_OperServ, source, "Current Autokill block list:");
      notice(s_OperServ, source, "Num Tracking Mask");
	for (i = 0; i < nakill; i++) {
       if (akills[i].flags & AK_BLOCK) {
        if (s) {
	    if (stricmp(s, "ME") == 0) {
		if (stricmp(akills[i].who, source) == 0) {
  			char timebuf[32], expirebuf[64];
			struct tm tm;
			time_t t = time(NULL);

			tm = *localtime(akills[i].time ? &akills[i].time : &t);
			strftime(timebuf, sizeof(timebuf), "%d %b %Y", &tm);
			if (akills[i].expires == 0) {
			    strcpy(expirebuf, "permanent");
			} else {
			    time_t t2 = akills[i].expires - t;
			    t2 += 59;
				t2 /= 60;
				snprintf(expirebuf, sizeof(expirebuf), "expires in %d minute%s", 
	                    t2, t2==1 ? "" : "s");
			}
			notice(s_OperServ, source, "%3d %8ld %s (by %s on %s; %s)", akcount, akills[i].tracking, akills[i].mask,
      	         *akills[i].who ? akills[i].who : "<unknown>", timebuf, expirebuf);
			notice(s_OperServ, source, " %s", akills[i].reason);
      	      akcount++;
            } else {
		   if ((match_wild_nocase(s, akills[i].mask)) || (atoi(s) == akills[i].tracking)) {
  			char timebuf[32], expirebuf[64];
			struct tm tm;
			time_t t = time(NULL);

			tm = *localtime(akills[i].time ? &akills[i].time : &t);
			strftime(timebuf, sizeof(timebuf), "%d %b %Y", &tm);
			if (akills[i].expires == 0) {
			    strcpy(expirebuf, "permanent");
			} else {
			    time_t t2 = akills[i].expires - t;
			    t2 += 59;
				t2 /= 60;
				snprintf(expirebuf, sizeof(expirebuf), "expires in %d minute%s", 
	                    t2, t2==1 ? "" : "s");
			}
			notice(s_OperServ, source, "%3d %8ld %s (by %s on %s; %s)", akcount, akills[i].tracking, akills[i].mask,
      	         *akills[i].who ? akills[i].who : "<unknown>", timebuf, expirebuf);
			notice(s_OperServ, source, " %s", akills[i].reason);
      	      akcount++;
               }
		}
	    }
        } else {
		char timebuf[32], expirebuf[64];
		struct tm tm;
		time_t t = time(NULL);

		tm = *localtime(akills[i].time ? &akills[i].time : &t);
		strftime(timebuf, sizeof(timebuf), "%d %b %Y", &tm);
		if (akills[i].expires == 0) {
		    strcpy(expirebuf, "permanent");
		} else {
		    time_t t2 = akills[i].expires - t;
		    t2 += 59;
			t2 /= 60;
			snprintf(expirebuf, sizeof(expirebuf), "expires in %d minute%s", 
                     t2, t2==1 ? "" : "s");
		}
		notice(s_OperServ, source, "%3d %8ld %s (by %s on %s; %s)", akcount, akills[i].tracking, akills[i].mask,
               *akills[i].who ? akills[i].who : "<unknown>", timebuf, expirebuf);
		notice(s_OperServ, source, " %s", akills[i].reason);
            akcount++;
	  }
      }
     }
    } else {
	notice(s_OperServ, source, "Syntax: AKILL [ADD|DEL|LIST|VIEW|BADD|BLIST|BDEL|BLIST] <expiry> <user@host> <\37reason\37> [\37ME\37]");
      notice(s_OperServ, source, "For more information type \2/msg %s HELP AKILL.", s_OperServ);
    }
}

/*************************************************************************/

/* os_readonly()
 * Set services in read only or read write mode. */

static void os_readonly(const char *source)
{
    char *option = strtok(NULL, " ");

    if (!option) {
      notice(s_OperServ, source, "Syntax: READONLY [ON|OFF]");
      notice(s_OperServ, source, "For more information type \2/msg %s HELP READONLY.", s_OperServ);
    } else if (stricmp(option, "ON") == 0) {
	    run.flags |= RUN_READONLY;	    
	    log("DS is now in read only mode.");
	    close_log();
	    notice(s_OperServ, source, "DS is now in read only mode, changes will not be saved.");
    } else if (stricmp(option, "OFF") == 0) {
	    run.flags &= ~RUN_READONLY;
	    open_log();
	    log("DS is not in read write mode.");
	    notice(s_OperServ, source, "DS is now in read write mode, changes will be saved.");
    } else {
      notice(s_OperServ, source, "Syntax: SET READONLY [ON|OFF]");
      notice(s_OperServ, source, "For more information type \2/msg %s HELP READONLY.", s_OperServ);
    }
}

/*************************************************************************/

/* os_stats() 
 * Statistics on services and the network. */

static void os_stats(const char *source)
{
   time_t uptime = time(NULL) - run.started;
   char *extra = strtok(NULL, " ");
   AdminInfo *ai;
   long count, mem;
   int i;

   if (!extra) {
      notice(s_OperServ, source, "Syntax: STATS [MEM|CLIENTS|NET|AKILL|CMDS]");
      notice(s_OperServ, source, "For more information type \2/msg %s HELP STATS.", s_OperServ);
   } else if (stricmp(extra, "AKILL") == 0) {
      notice(s_OperServ, source, "Autokill statistics:");
      notice(s_OperServ, source, "Autokills: \2%d\2", nakill);
      notice(s_OperServ, source, "There have been %d autokills by Services Administrators (1).", stats.sadmin_akills);
      notice(s_OperServ, source, "There have been %d autokills by Services Operators (2).", stats.csop_akills);
      notice(s_OperServ, source, "There have been %d autokills by Services Roots (3).", stats.sroot_akills);
   } else if (stricmp(extra, "CMDS") == 0) {
   } else if (stricmp(extra, "MEM") == 0) {
      if (!is_services_admin(source)) {
         notice(s_OperServ, source, "Your level is IRC Operator, the required level to use this command is Services Administrator.");
      } else {
         notice(s_OperServ, source, "Memory usage statistics:");
         get_user_stats(&count, &mem);
         notice(s_OperServ, source, "User    : \2%6d\2 records, \2%5d\2 kB", count, (mem + 512) / 1024);
         get_channel_stats(&count, &mem);
         notice(s_OperServ, source, "Channel : \2%6d\2 records, \2%5d\2 kB", count, (mem + 512) / 1024);
         get_nickserv_stats(&count, &mem);
         notice(s_OperServ, source, "NickServ: \2%6d\2 records, \2%5d\2 kB", count, (mem + 512) / 1024);
         get_chanserv_stats(&count, &mem);
         notice(s_OperServ, source, "ChanServ: \2%6d\2 records, \2%5d\2 kB", count, (mem + 512) / 1024);
         get_memoserv_stats(&count, &mem);
         notice(s_OperServ, source, "MemoServ: \2%6d\2 records, \2%5d\2 kB", count, (mem + 512) / 1024);
         count = 0;
         mem = sizeof(struct clone) * CLONE_DETECT_SIZE * 2;
         for (i = 0; i < CLONE_DETECT_SIZE; i++) {
            if (clonelist[i].host) {
               count++;
               mem += strlen(clonelist[i].host) + 1;
            }
            if (warnings[i].host) {
               count++;
               mem += strlen(warnings[i].host) + 1;
            }
         }
         for (i = 0; i < 256; i++) {
            for (ai = adminlists[i]; ai; ai = ai->next) {
               mem += strlen(ai->nick) + 1;
            }
         }
         mem += sizeof(struct akill) * akill_size;
         count += nakill;
         for (i = 0; i < nakill; i++)
            mem += strlen(akills[i].mask) + 1 + strlen(akills[i].reason) + 1;
         notice(s_OperServ, source, "OperServ: \2%6d\2 records, \2%5d\2 kB", count, (mem + 512) / 1024);
      }
   } else if (stricmp(extra, "CLIENTS") == 0) {
      notice(s_OperServ, source, "Client (active/deactive) statistics:");
      if (conf.servicesactive) {
         notice(s_OperServ, source, "%s is currently %sactivated.",
            s_NickServ, conf.nsactive == 1 ? "" : "de");
         notice(s_OperServ, source, "%s is currently %sactivated.",
            s_ChanServ, conf.csactive == 1 ? "" : "de");
         notice(s_OperServ, source, "%s is currently %sactivated.",
            s_MemoServ, conf.msactive == 1 ? "" : "de");
         notice(s_OperServ, source, "%s is currently %sactivated.",
            s_HelpServ, conf.hsactive == 1 ? "" : "de");
         notice(s_OperServ, source, "%s is currently %sactivated.",
            s_OperServ, conf.osactive == 1 ? "" : "de");
      } else {
         notice(s_OperServ, source, "All DS clients are currently deactivated.");
      }
   } else if (stricmp(extra, "NET") == 0) {
      notice(s_OperServ, source, "Network statistics:");
      notice(s_OperServ, source, "Current users network wide: \2%d\2, invisible \2%d\2 of which \2%d\2 are operators", lu_users, lu_invisible, lu_opers);
      notice(s_OperServ, source, "Maximum users network wide: \2%d\2", lu_max_users);
      if (uptime > 86400)
         notice(s_OperServ, source, "DS up \2%d\2 day%s, \2%02d:%02d\2",
            uptime / 86400, (uptime / 86400 == 1) ? "" : "s", (uptime / 3600) % 24, (uptime / 60) % 60);
      else if (uptime > 3600)
         notice(s_OperServ, source, "DS up \2%d hour%s, %d minute%s\2",
            uptime / 3600, uptime/3600==1 ? "" : "s", (uptime / 60) % 60, (uptime / 60)%60 == 1 ? "" : "s");
      else
         notice(s_OperServ, source, "DS up \2%d minute%s, %d second%s\2",
            uptime / 60, uptime / 60 == 1 ? "" : "s", uptime%60, uptime%60 == 1 ? "" : "s");
   } else {
      notice(s_OperServ, source, "Syntax: STATS [MEM|CLIENTS|NET|AKILL|CMDS]");
      notice(s_OperServ, source, "For more information type \2/msg %s HELP STATS.", s_OperServ);
   }
}

/*************************************************************************/

/* os_jupe()
 * Jupiter a server. */

static void os_jupe(const char *source)
{
   char *server = strtok(NULL, " ");
   ServerInfo *si;

   if (!server) {
      notice(s_OperServ, source, "Syntax: JUPE <server>");
      notice(s_OperServ, source, "For more information type \2/msg %s HELP JUPE.", s_OperServ);
   } else {
      if (si = findserver(server) ) {
         notice(s_OperServ, source, "You must /squit %s before you are able to jupiter it.", server);
      } else {
         if (si = add_server(server)) {
            send_cmd(NULL, "SERVER %s 2 :Jupitered (%s)", server, source);
            globops(s_OperServ, "\2%s\2 has jupitered (jupe) the server \2%s\2.", source, server);
            notice(s_OperServ, source, "Server \2%s\2 has been jupitered.", server); 
            si->usrcnt = 0;
            si->hops = 1;
		si->flags = SI_JUPE;
            si->ping = 0;
         }
      }
   }
}

/*************************************************************************/

/* os_mode()
 * Change the modes on a channel or user on a channel. */

static void os_mode(const char *source)
{
   char *modestr = strtok(NULL, "");
   int argc;
   char **argv;

   if (!modestr) {
      notice(s_OperServ, source, "Syntax: MODE <channel> <modes> <\37nick\37>");
      notice(s_OperServ, source, "For more information type \2/msg %s HELP MODE.", s_OperServ);
   } else {
      send_cmd(s_OperServ, "MODE %s", modestr);
      argc = split_buf(modestr, &argv, 1);
      do_cmode(s_OperServ, argc, argv);
   }
}

/*************************************************************************/

/* os_kick()
 * Kick a user from a channel. */

static void os_kick(const char *source)
{
   char *chan = strtok(NULL, " ");
   char *nick = strtok(NULL, " ");
   char *reason = strtok(NULL, "");
   char *av[3];
   Channel *c;

   if (!chan || !nick || !reason) {
      notice(s_OperServ, source, "Syntax: KICK <channel> <nick> <reason>");
      notice(s_OperServ, source, "For more information type \2/msg %s HELP KICK.", s_OperServ);
   } else if (!(c = findchan(chan))) {
      notice(s_OperServ, source, "Channel \2%s\2 does not exist.", chan);
   } else if (!is_on_chan(nick, chan)) {
      notice(s_OperServ, source, "Nick \2%s\2 is not currently in \2%s\2.", nick, chan);
   } else {
      send_cmd(s_OperServ, "KICK %s %s :%s (%s)", chan, nick, source, reason);
      av[0] = sstrdup(chan);
      av[1] = sstrdup(nick);
      av[2] = sstrdup(reason);
      do_kick(s_OperServ, 3, av);
      free(av[2]);
      free(av[1]);
      free(av[0]);
   }
}

/*************************************************************************/

static void timeout_leave(Timeout *to)
{
    char *chan = to->data;
    send_cmd(s_OperServ, "MODE %s -isb *!*@*", chan);
    send_cmd(s_OperServ, "PART %s", chan);
    free(to->data);
}

/*************************************************************************/

/* os_mdeop()
 * Massdeop all users in a channel. */

static void os_mdeop(const char *source)
{
   char *chan = strtok(NULL, " ");
   Channel *c;
   Timeout *t;
   char *av[3];
   struct c_userlist *cu, *next;
   int users = 0;
   User *u = finduser(source);

   if (!chan) {
      notice(s_OperServ, source, "Syntax: MDEOP <channel>");
      notice(s_OperServ, source, "For more information type \2/msg %s HELP MDEOP.", s_OperServ);
   } else if (!(c = findchan(chan))) {
      notice(s_OperServ, source, "Channel \2%s\2 does not exist.", chan);
   } else {
      for (cu = c->chanops; cu; cu = next) {
         next = cu->next;
         users++;
         av[0] = sstrdup(chan);
         av[1] = sstrdup("-o");
         av[2] = sstrdup(cu->user->nick);
         send_cmd(s_OperServ, "MODE %s %s :%s", av[0], av[1], av[2]);
         do_cmode(s_OperServ, 3, av);
         free(av[2]);
         free(av[1]);
         free(av[0]);
      }
      send_cmd(s_OperServ, "JOIN %s", chan);
      send_cmd(s_OperServ, "MODE %s +o %s", chan, s_OperServ);
      t = add_timeout(conf.csinhabit, timeout_leave, 0);
      t->data = sstrdup(chan);
      globops(s_OperServ, "\2%s\2 massdeoped (mdeop) the channel \2%s\2, affecting \2%d\2 users.",
         u->nick, chan, users);
      notice(s_OperServ, source, "Channel \2%s\2 has been massdeoped.", chan);
   }
}

/*************************************************************************/

/* os_mkick()
 * Masskick all users from a channel. */

static void os_mkick(const char *source)
{
   char *chan = strtok(NULL, " ");
   char *reason = strtok(NULL, "");
   Channel *c;
   Timeout *t;
   char *av[3];
   struct c_userlist *cu, *next;
   char buf[256];
   int users = 0;
   User *u = finduser(source);

   if (!chan || !reason) {
      notice(s_OperServ, source, "Syntax: MKICK <channel> <reason>");
      notice(s_OperServ, source, "For more information type \2/msg %s HELP MKICK.", s_OperServ);
   } else if (!(c = findchan(chan))) {
      notice(s_OperServ, source, "Channel \2%s\2 does not exist.", chan);
   } else {
      for (cu = c->users; cu; cu = next) {
         users++;
         next = cu->next;
         av[0] = sstrdup(chan);
         av[1] = sstrdup(cu->user->nick);
         av[2] = sstrdup(reason);
         send_cmd(s_OperServ, "KICK %s %s :%s", av[0], av[1], av[2]);
         do_kick(s_OperServ, 3, av);
         free(av[2]);
         free(av[1]);
         free(av[0]);
      }
      send_cmd(s_OperServ, "JOIN %s", chan);
      send_cmd(s_OperServ, "MODE %s +o %s", chan, s_OperServ);
      send_cmd(s_OperServ, "MODE %s +isb *!*@*", chan);
      t = add_timeout(conf.csinhabit, timeout_leave, 0);
      t->data = sstrdup(chan);

      globops(s_OperServ, "\2%s\2 masskicked (mkick) the channel \2%s\2, affecting \2%d\2 users.",
         u->nick, chan, users);
      notice(s_OperServ, source, "Channel \2%s\2 has been masskicked.", chan);
   }
}

/*************************************************************************/

/* os_sinvite()
 * Invite roots into the dump channel and special channel. */

static void os_sinvite(const char *source)
{
   char *s = strtok(NULL, " ");

   if (!s) {
      notice(s_OperServ, source, "Syntax: SINVITE [DUMP|SPC]");
          notice(s_OperServ, source, "For more information type \2/msg %s HELP SINVITE.", s_OperServ);
   } else if (stricmp(s, "DUMP") == 0) {
      send_cmd(s_OperServ, "INVITE %s %s", source, conf.dumpchannel);
   } else if (stricmp(s, "SPC") == 0) {
      send_cmd(s_OperServ, "INVITE %s %s", source, conf.spcchannel);
   } else {
      notice(s_OperServ, source, "Syntax: SINVITE [DUMP|SPC]");
          notice(s_OperServ, source, "For more information type \2/msg %s HELP SINVITE.", s_OperServ);
   }
}

/*************************************************************************/

/* os_schan()
 * Enable or disable the use of the dump channel or special. */

static void os_schan(const char *source)
{
   char *chan = strtok(NULL, " ");
   char *s = strtok(NULL, " ");

   if (!s || !chan) {
      notice(s_OperServ, source, "Syntax: SCHAN [DUMP|SPC] [ENABLE|DISABLE]");
          notice(s_OperServ, source, "For more information type \2/msg %s HELP SCHAN.", s_OperServ);
   } else if (stricmp(chan, "DUMP") == 0) {
      if (stricmp(s, "ENABLE") == 0) {
	      conf.dumpchannelactive = 1;
	      notice(s_OperServ, source, "Channel dumping for DS in \2%s\2 has been enabled.", conf.dumpchannel);
      } else if (stricmp(s, "DISABLE") == 0) {
	      conf.dumpchannelactive = 0;
	      notice(s_OperServ, source, "Channel dumping for DS in \2%s\2 has been disabled.", conf.dumpchannel);
      } else {
      	notice(s_OperServ, source, "Syntax: SCHAN DUMP [ENABLE|DISABLE]");
            notice(s_OperServ, source, "For more information type \2/msg %s HELP SCHAN.", s_OperServ);
      }
   } else if (stricmp(chan, "SPC") == 0) {
      if (stricmp(s, "ENABLE") == 0) {
	      conf.spcchannelactive = 1;
	      notice(s_OperServ, source, "Special channel for DS in \2%s\2 has been enabled.", conf.spcchannel);
      } else if (stricmp(s, "DISABLE") == 0) {
	      conf.spcchannelactive = 0;
	      notice(s_OperServ, source, "Special channel for DS in \2%s\2 has been enabled.", conf.spcchannel);
      } else {
	      notice(s_OperServ, source, "Syntax: SCHAN SPC [ENABLE|DISABLE]");
            notice(s_OperServ, source, "For more information type \2/msg %s HELP SCHAN.", s_OperServ);
      }
   } else {
      notice(s_OperServ, source, "Syntax: SCHAN [DUMP|SPC] [ENABLE|DISABLE]");
          notice(s_OperServ, source, "For more information type \2/msg %s HELP SCHAN.", s_OperServ);
   }
}

/*************************************************************************/

/* os_noop()
 * Prevent a user from opering or all users on a server. */

static void os_noop(const char *source)
{
   char *s = strtok(NULL, " ");
   ServerInfo *si;
   User *u;
   int noop_set = 0;

   /* We look for a user then a server, for simplistic reasons. */
   if (!s) {
      notice(s_OperServ, source, "Syntax: NOOP [nick|server]");
      notice(s_OperServ, source, "For more information type \2/msg %s HELP NOOP.", s_OperServ);
   } else if ((u = finduser(s))) {
      u->flags |= U_NOOP;
      notice(s_OperServ, source, "The user \2%s!%s@%s\2 is now unable to oper.", u->nick, u->username, u->host);
   } else if ((si = findserver(s))) {
      si->flags |= SI_NOOP;
      notice(s_OperServ, source, "All users on the server \2%s\2 are now unable to oper.", s);
   } else {
      notice(s_OperServ, source, "Syntax: NOOP [nick|server]");
      notice(s_OperServ, source, "For more information type \2/msg %s HELP NOOP.", s_OperServ);
   }
}

/*************************************************************************/

/* os_raw()
 * Send a raw string to the server. */

static void os_raw(const char *source)
{
   char *text = strtok(NULL, "");

   if (!text) {
      notice(s_OperServ, source, "Syntax: RAW <command>");
      notice(s_OperServ, source, "For more information type \2/msg %s HELP RAW.", s_OperServ);
   } else {
      send_cmd(NULL, text);
   }
}

/*************************************************************************/

/* os_activate()
 * Activate or deactivate services clients. */

static void os_activate(const char *source)
{
   char *what = strtok(NULL, " ");
   char *param = strtok(NULL, " ");
   int allflag = 0, onoffflag = 0, normflag = 0;

   if (!what || !param) {
      notice(s_OperServ, source, "Syntax: ACTIVATE [%s|%s|%s|%s|%s|ALL] [ON|OFF]",
         s_ChanServ, s_NickServ, s_MemoServ, s_HelpServ, s_OperServ);
      notice(s_OperServ, source, "For more information type \2/msg %s HELP ACTIVATE.", s_OperServ);
   } else if (stricmp(what, s_ChanServ) == 0) {
      if (stricmp(param, "ON") == 0) {
         conf.csactive = 1;
         onoffflag = 1;
         normflag = 1;
      }
      if (stricmp(param, "OFF") == 0) {
         conf.csactive = 0;
         onoffflag = 0;
         normflag = 1;
     } else {
        notice(s_OperServ, source, "Syntax: ACTIVATE %s [ON|OFF]", s_ChanServ);
        notice(s_OperServ, source, "For more information type \2/msg %s HELP ACTIVATE.", s_OperServ);
     }
     if (normflag)
         activate_client_notice(source, s_ChanServ, onoffflag);
   } else if (stricmp(what, s_NickServ) == 0) {
      if (stricmp(param, "ON") == 0) {
         conf.nsactive = 1;
         onoffflag = 1;
         normflag = 1;
      } else if (stricmp(param, "OFF") == 0) {
         conf.nsactive = 0;
         onoffflag = 0;
         normflag = 1;
     } else {
        notice(s_OperServ, source, "Syntax: ACTIVATE %s [ON|OFF]", s_NickServ);
        notice(s_OperServ, source, "For more information type \2/msg %s HELP ACTIVATE.", s_OperServ);
     }
     if (normflag)
         activate_client_notice(source, s_NickServ, onoffflag);
   } else if (stricmp(what, s_MemoServ) == 0) {
      if (stricmp(param, "ON") == 0) {
         conf.msactive = 1;
         onoffflag = 1;
         normflag = 1;
      } else if (stricmp(param, "OFF") == 0) {
         conf.msactive = 0;
         onoffflag = 0;
         normflag = 1;
     } else {
        notice(s_OperServ, source, "Syntax: ACTIVATE %s [ON|OFF]", s_MemoServ);
        notice(s_OperServ, source, "For more information type \2/msg %s HELP ACTIVATE.", s_OperServ);
     }
     if (normflag)
         activate_client_notice(source, s_MemoServ, onoffflag);
   } else if (stricmp(what, s_OperServ) == 0) {
      if (stricmp(param, "ON") == 0) {
         conf.osactive = 1;
         onoffflag = 1;
         normflag = 1;
      } else if (stricmp(param, "OFF") == 0) {
         conf.osactive = 0;
         onoffflag = 0;
         normflag = 1;
     } else {
        notice(s_OperServ, source, "Syntax: ACTIVATE %s [ON|OFF]", s_OperServ);
        notice(s_OperServ, source, "For more information type \2/msg %s HELP ACTIVATE.", s_OperServ);
     }
     if (normflag)
         activate_client_notice(source, s_OperServ, onoffflag);
   } else if (stricmp(what, s_HelpServ) == 0) {
      if (stricmp(param, "ON") == 0) {
         conf.hsactive = 1;
         onoffflag = 1;
         normflag = 1;
      } else if (stricmp(param, "OFF") == 0) {
         conf.hsactive = 0;
         onoffflag = 0;
         normflag = 1;
     } else {
        notice(s_OperServ, source, "Syntax: ACTIVATE %s [ON|OFF]", s_HelpServ);
        notice(s_OperServ, source, "For more information type \2/msg %s HELP ACTIVATE.", s_OperServ);
     }
     if (normflag)
         activate_client_notice(source, s_HelpServ, onoffflag);
   } else if (stricmp(what, "ALL") == 0) {
      if (stricmp(param, "ON") == 0) {
         conf.servicesactive = 1;
         allflag = 1;
         onoffflag = 1;
      } else if (stricmp(param, "OFF") == 0) {
         conf.servicesactive = 0;
         allflag = 1;
         onoffflag = 0;
      } else {
         notice(s_OperServ, source, "Syntax: ACTIVATE ALL [ON|OFF]");
         notice(s_OperServ, source, "For more information type \2/msg %s HELP ACTIVATE.", s_OperServ);
      }
      if (allflag)
         activate_client_notice(source, "ALL", onoffflag);
   } else {
      notice(s_OperServ, source, "Syntax: ACTIVATE [%s|%s|%s|%s|%s|ALL] [ON|OFF]\2",
         s_ChanServ, s_NickServ, s_MemoServ, s_HelpServ, s_OperServ);
      notice(s_OperServ, source, "For more information type \2/msg %s HELP ACTIVATE.", s_OperServ);
   }
}

/*************************************************************************/

/* activate_client_notice()
 * Send the notices for activation and deactivation of services clients. */

static void activate_client_notice(const char *source, char *client, int onoff)
{
   if (stricmp(client, "ALL") == 0) {
      globops(s_OperServ, "\2%s\2 has %sactivated (activate) all services clients.",
         source, onoff == 1 ? "" : "de");
      notice(s_OperServ, source, "All DS clients are now %sactivated.",
         onoff == 1 ? "" : "de");
   } else {
      globops(s_OperServ, "\2%s\2 has %sactivated (activate) \2%s\2.",
         source, client, onoff == 1 ? "" : "de");
      notice(s_OperServ, source, "\2%s\2 is now %sactivated.", 
         client, onoff == 1 ? "" : "de");
   }
}

/*************************************************************************/

static void os_map(const char *source)
{
   send_map(source, s_OperServ);
}

/*************************************************************************/

/* os_route()
 * Manipulate the routing list. */

static void os_route(const char *source)
{
   char *cmd, *server, *pass, *port;
   int i, rcount = 1, lprt = 0;
   RouteInfo *ri;
   cmd = strtok(NULL, " ");

   if (!cmd)
      cmd = "";

    if (run.flags & RUN_READONLY) {
	notice(s_OperServ, source, "DS is currently in read only mode, changes will not be saved.");
   }

   if (stricmp(cmd, "ADD") == 0) {
      server = strtok(NULL, " ");
      pass = strtok(NULL, " ");
      port = strtok(NULL, " ");
      if (!is_services_op(source)) {
         notice(s_OperServ, source, "Access denied.");
         return;
      }
      if (server && pass && port) {
	    if (!(lprt = atoi(port))) {
		notice(s_OperServ, source, "Port must be an integer.");
          } else if (lprt <= 9999 && lprt >= 5000) {
             if (!(ri = findroute(server))) {
                if (ri = add_route(server)) {
                   ri->port = atoi(port);
			 strscpy(ri->pass, pass, PASSMAX);
			 globops(NULL, "\2%s\2 has added a route (route) for %s:%d.", source, ri->server, ri->port); 
                   notice(s_OperServ, source, "\2%s\2 has been added to the routing list.", ri->server);
                }
             } else {
                notice(s_OperServ, source, "\2%s\2 is already on the routing list.", ri->server);
             }
          }	else {
		notice(s_OperServ, source, "Port must be an integer between %d and %d inclusive.", 5000, 9999);
          }
	} else {
	    notice(s_OperServ, source, "Syntax: ROUTE ADD <server> <password> <port>");
          notice(s_OperServ, source, "For more information type \2/msg %s HELP ROUTE.", s_OperServ);
	}

    } else if (stricmp(cmd, "DEL") == 0) {
	if (!is_services_op(source)) {
	    notice(s_OperServ, source, "Access denied.");
	    return;
	}
      server = strtok(NULL, " ");
	if (server) {
          if (ri = findroute(server)) {
			 globops(NULL, "\2%s\2 has removed the route (route) for %s:%d.", source, ri->server, ri->port); 
                notice(s_OperServ, source, "\2%s\2 has been removed from the routing list.", 
                   ri->server);
                delroute(ri);
             } else {
                notice(s_OperServ, source, "\2%s\2 is not on the routing list.", 
                   server);
             }
	} else {
	    notice(s_OperServ, source, "Syntax: ROUTE DEL <server>");
          notice(s_OperServ, source, "For more information type \2/msg %s HELP ROUTE.", s_OperServ);
	}
    } else if (stricmp(cmd, "LIST") == 0) {
         notice(s_OperServ, source, "Current routing list:");
         notice(s_OperServ, source, "Num Port Password         Server");
         for (i = 0; i < 256; i++) {
             for (ri = routelists[i]; ri; ri = ri->next) {
                       notice(s_OperServ, source, "%3d %4d %-16s %s", 
                          rcount, ri->port, ri->pass, ri->server);
                    rcount++;
             }
          }
        notice(s_OperServ, source, "Use \2/msg %s ROUTE CONNECT <server>\2 to reroute DS.", s_OperServ);
    } else if (stricmp(cmd, "CONNECT") == 0) {
	 server = strtok(NULL, "");
       if (server) {
		if (ri = findroute(server)) {
		 /* Puts into reroute mode. */
	 	 globops(NULL, "\2%s\2 is rerouting DS from %s:%d to %s:%d.", source, conf.rserver, conf.rport, ri->server, ri->port);
	       // save_data = 4;
	       strscpy(route_server, server, HOSTMAX);
  	       log("route: Freeing structures...");
	       log("route: Freeing server structures...");
	       free_all_server();
	       log("route: Freeing channel structures...");
	       free_all_chan();
	       log("route: Freeing user structures...");
	       free_all_user();
	       log("route: Freeing structures complete.");	       
	       run.flags |= (RUN_SAVING | RUN_ROUTING); 
		 run.flags &= ~(RUN_SYNCING | RUN_LINKED);
                 run.flags &= ~(RUN_SYNCING1 | RUN_LINKED1);
	       /* The rest is done in main.c */
		} else {
                notice(s_OperServ, source, "\2%s\2 is not on the routing list.", server);
		}
	 } else {
	    notice(s_OperServ, source, "Syntax: CONNECT <server>");
          notice(s_OperServ, source, "For more information type \2/msg %s HELP ROUTE.", s_OperServ);
	}
    } else {
	notice(s_OperServ, source, "Syntax: ROUTE [ADD|DEL|LIST|CONNECT] <server> <password> <port>");
      notice(s_OperServ, source, "For more information type \2/msg %s HELP ROUTE.", s_OperServ);
    }
}

/*************************************************************************/

#endif

/*************************************************************************/
