/* Functions to nanipulate online user information, commands, user settings and modes.
 * src/users.c
 * Copyright (C) 2000-2001 Lucas Balemba.  All rights reserved.
 * Read the file doc/contact for contact information related to this software. */

/*************************************************************************/

#ifndef    __users_source__
#define __users_source__

#ifndef lint
   static char sccsid[] = "@(#)users.c		7.0 18/03/01 (C) 2001 Lucas Balemba $exp 0";
#endif

/*************************************************************************/

#include "services.h"

/*************************************************************************/

static User *userlist = NULL;
int lu_users = 0, lu_invisible = 0, lu_opers = 0, lu_opers_inv = 0, lu_max_users = 0, lu_max_opers = 0;

/*************************************************************************/

/* new_user()
 * Allocate a new User structure, fill in basic values, link it to the overall list,
 * and return it.  Always successful, returns *user or NULL. */

static User *new_user(const char *nick)
{
   User *user;

   user = scalloc(sizeof(User), 1);
   if (!nick) {
      nick = "";
   }
   strscpy(user->nick, nick, NICKMAX);
   user->next = userlist;
   if (userlist) {
      userlist->prev = user;
   }
   userlist = user;
   user->ni = findnick(nick);
   lu_users++;
   /* We do not need to test for mode +i here because user info is set before modes,
    * thus we would create the user then assign its modes. - Drakan */
   if (lu_users > lu_max_users) {
      lu_max_users = lu_users;
   }
   /* This should never happen, but we can not be sure we will always have correct
    * pointer values - Drakan */
   return (user ? user : NULL);
}

/*************************************************************************/

/* update_last_seen()
 * Update the last seen time field on registered nicks. */

void update_last_seen(void)
{
   User *u;

   for (u = userlist; u; u = u->next) {
      if (u->ni) {
         u->ni->last_seen = time(NULL);
      }
   }
}

/*************************************************************************/

/* change_user_nick()
 * Change the nick of a online user, and move pointers as necessary. */

void change_user_nick(User *u, const char *nick)
{
    strscpy (u->nick, nick, NICKMAX);
    u->ni = findnick(nick);
}

/*************************************************************************/

/* delete_user()
 * Remove a user structure from the linked user list and update counters. */

static void delete_user(User *u)
{
    struct u_chanlist *c, *c2;
    struct u_chaninfolist *ci, *ci2;
    struct u_nickinfolist *id, *id2;
    ServerInfo *si;

    /* Check for +i, then non +i decrease them accordingly. */
    if (is_invisible(u->nick)) {
       lu_invisible--;
    } else {
       lu_users--;
    }
    /* Then for IRC Operators, decrease this count as well. */
    if (is_oper(u->nick)) {
	lu_opers--;
    }
    /* And finally keep server counts up to date, and decrease this count. */
    if (si = findserver(u->server)) {
          si->usrcnt--;
    }
    cancel_user(u);
    free(u->username);
    free(u->host);
    free(u->realname);
    free(u->server);
    c = u->chans;
    while (c) {
	c2 = c->next;
	chan_deluser(u, c->chan);
	free(c);
	c = c2;
    }
    ci = u->founder_chans;
    while (ci) {
	ci2 = ci->next;
	free(ci);
	ci = ci2;
    }
    id = u->identified_nicks;
    while (id) {
	id2 = id->next;
	free(id);
	id = id2;
    }
    if (u->prev) {
	u->prev->next = u->next;
    } else {
	userlist = u->next;
    }
    if (u->next) {
	u->next->prev = u->prev;
    }
    free(u);
}

/*************************************************************************/

/* get_user_stats()
 * Return memory usage statistics on the linked user structure, all pointers are
 * assumed to be valid and non NULL or accepted as non NULL. */

void get_user_stats(long *nusers, long *memuse)
{
   long count = 0, mem = 0;
   int i;
   User *u;
   struct u_chanlist *uc;
   struct u_chaninfolist *uci;

   for (u = userlist; u; u = u->next) {
      count++;
	mem += sizeof(*u);
	if (u->username) {
	   mem += strlen(u->username)+1;
      }
	if (u->host) {
	   mem += strlen(u->host)+1;
      }
	if (u->realname) {
         mem += strlen(u->realname)+1;
      }
	if (u->server) {
         mem += strlen(u->server)+1;
      }
	for (uc = u->chans; uc; uc = uc->next) {
         mem += sizeof(*uc);
      }
	for (uci = u->founder_chans; uci; uci = uci->next) {
         mem += sizeof(*uci);
      }
   }
   *nusers = count;
   *memuse = mem;
}

/*************************************************************************/

/* free_all_user()
 * Remove and free all users on the linked user list. */

void free_all_user(void)
{
   User *u;

   for (u = userlist; u; u = u->next) {
      delete_user(u);
   }
   lu_opers = lu_users = lu_invisible = 0;
}

/*************************************************************************/

/* finduser()
 * Iterate over the linked user list for a specific user, if found return it's
 * data otherwise return NULL. */

User *finduser(const char *nick)
{
    User *u;

    for (u = userlist; u && stricmp(u->nick, nick) != 0; u = u->next);
       return (u ? u : NULL);
}

/*************************************************************************/

/* send_akill()
 * Send a kill message to all users who's mask matches the autokilled mask. */

void send_akill(const char *source, char *user, char *host, const char *reason, int amount, time_t expires)
{
   User *u;
   int i, kills = 0, users = 0;
      char treason[256];
   time_t now = time(NULL);

   for (u = userlist; u; u = u->next) {
        if (match_wild_nocase(user, u->username) && match_wild_nocase(host, u->host)) {
/*#ifdef OS_BSD
            snprintf(treason, sizeof(treason), "User has been banned (%s. %s)",
               reason, conf.osakillmsg);
            kill_user(NULL, u->nick, treason);
#else
*/

            notice(s_OperServ, u->nick, "User has been banned. (%s. %s)", reason, conf.osakillmsg);
            
	    send_cmd(NULL, "KILL %s :(User has been banned (%s. %s))",
 	    u->nick, reason, conf.osakillmsg);
#if defined(V_BAHAMUT)
	      send_cmd(conf.lserver, "AKILL %s %s %d %s %ld :User has been banned (%s. %s)",
               host, user, 86400*2, source, time(NULL), reason, conf.osakillmsg);
#elif defined(V_DF) || defined(V_UNREAL)
	      send_cmd(conf.lserver, "AKILL %s %s :User has been banned (%s. %s)",
               host, user, reason, conf.osakillmsg);
#else
	      send_cmd(conf.lserver, "GLINE * +%ld %s@%s :User has been banned (%s)",
		    expires && (expires > now) ? (expires - time(NULL)) : 999999999,
		    username2, host2, reason);
#endif
            kills++;
         }
         users++;
   }
   if (amount <= 0) {
     globops(NULL, "Autokill added for \2%s@%s\2 by \2%s\2 (permanent), affecting %d users, %3.4f%% of the network.",
       user, host, source, kills, (float) (kills * 100) / (float)users);
   } else {
     globops(NULL, "Autokill added for \2%s@%s\2 by \2%s\2 for %d minutes, affecting %d users, %3.4f%% of the network.",
       user, host, source, amount, kills, (float) (kills * 100) / (float)users);
   }
   /* Here, there are no users on the network?  Not likely to happen, but have to do this.  Or we had no kills, so
    * we need to Autokill the host anyways. */
   if (users == 0 || kills == 0) {
#if defined(V_BAHAMUT)
	      send_cmd(conf.lserver, "AKILL %s %s %d %s %ld :User has been banned (%s. %s)",
               host, user, 86400*2, source, time(NULL), reason, conf.osakillmsg);
#elif defined(V_DF) || defined(V_UNREAL)
	      send_cmd(conf.lserver, "AKILL %s %s :User has been banned (%s. %s)",
               host, user, reason, conf.osakillmsg);
#else
	      send_cmd(conf.lserver, "GLINE * +%ld %s@%s :User has been banned (%s)",
		    expires && (expires > now) ? (expires - time(NULL)) : 999999999,
		    username2, host2, reason);
#endif
   }
   kills = users = 0;
}

/*************************************************************************/

/* send_find()
 * Find all the users online who are +o or +a and or are services
 * administrators. */

void send_find(const char *source, int type)
{
   User *u;
   int i, j = 0;

   notice(s_NickServ, source, rpl_numeric_lang(source, NS_PLEASE_ADVISE_FIND));
   notice(s_NickServ, source, "Num Nick");
      for (u = userlist; u; u = u->next) {
          switch (type) {
             case C_FALSE:
                if (is_oper(u->nick)) { 
                    j++;
                   notice(s_NickServ, source, "%3d %s (%s) %s", j, u->nick, u->host, IsAway(u) ? "(away)" : "");
                }
                break;
             case C_TRUE:
                if (is_sadmin(u->nick)) { 
                   j++;
                   notice(s_NickServ, source, "%3d %s (%s) %s", j, u->nick, u->host, IsAway(u) ? "(away)" : "");
                }
                break;
             default:
                if (is_oper(u->nick)) { 
                    j++;
                   notice(s_NickServ, source, "%3d %s (%s) %s", j, u->nick, u->host, IsAway(u) ? "(away)" : "");
                }
                break;
         }
     }
} 

/*************************************************************************/

/* do_nick()
 * Handle a server NICK command.
 *	av[0] = nick
 *	If a new user:
 *		av[1] = hop count
 *		av[2] = signon time
 *		av[3] = username
 *		av[4] = hostname
 *		av[5] = user's server
 *		av[6] = user's real name
 *	Else:
 *		av[1] = time of change */

void do_nick(const char *source, int ac, char **av)
{
   User *u;
   NickInfo *ni;
   ServerInfo *si;
   char *uv[3];

   if (!*source) {
	/* This is a new user; create a User structure for it. */

	/* We used to ignore the ~ which a lot of ircd's use to indicate no
	 * identd response.  That caused channel bans to break, so now we
	 * just take what the server gives us.  People are _still_
	 * encouraged to read the RFCs and stop doing anything to usernames
	 * depending on the result of an identd lookup. */

      /* Does this user cross the allowed user limit based on user Y:line 1? */
	if ((lu_users + 1) > conf.yumax) {
	   send_cmd(NULL, "KILL %s :%s %s", av[0], conf.lserver, rpl_numeric_lang(av[0], CR_KILL_YLINE_USER));
	   return;
	}
#ifdef V_BAHAMUT
      if (check_akill(av[0], av[4], av[5])) {
         return;
      }
#else
	if (check_akill(av[0], av[3], av[4])) {
         return;
      }
#endif
	u = new_user(av[0]);
	u->signon = atol(av[2]);
#ifdef V_BAHAMUT
      u->username = sstrdup(av[4]);
      u->host = sstrdup(av[5]);
      u->server = sstrdup(av[6]);
      u->realname = sstrdup(av[8]);

      if (conf.logonmsg1) {
         notice(s_GlobalNoticer, u->nick, "%s", conf.logonmsg1);
      }
      if (conf.logonmsg2) {
         notice(s_GlobalNoticer, u->nick, "%s", conf.logonmsg2);
      }
      if (conf.logonmsg3) {
         notice(s_GlobalNoticer, u->nick, "%s", conf.logonmsg3);
      }
      uv[0] = sstrdup(u->nick);
      uv[1] = sstrdup(av[3]);
      do_umode(u->nick, 2, uv);
#else
	u->username = sstrdup(av[3]);
	u->host = sstrdup(av[4]);
	u->server = sstrdup(av[5]);
	u->realname = sstrdup(av[6]);
#endif
      if (conf.servscan) {
         /* First check username, then realname. */
         if (match_wild_nocase("*CHANSERV*", u->username) || match_wild_nocase("*MEMOSERV*", u->username) ||
	       match_wild_nocase("*NICKSERV*", u->username) || match_wild_nocase("*OPERSERV*", u->username) ||
	       match_wild_nocase("*SERV*", u->username) || match_wild_nocase("*MEMOSERV*", u->username)) {
	       if (conf.spcchannelactive) {
	          privmsg(s_NickServ, conf.spcchannel, "User connecting has a username matching or referencing a services client -> (%s!%s@%s).", 
                   u->nick, u->username, u->host);
             }
         } else if (match_wild_nocase("*IRC*OP", u->username) || match_wild_nocase("*ADMIN*", u->username) ||
	              match_wild_nocase("*SERVICE*", u->username) || match_wild_nocase("*STAFF*", u->username) ||
		        match_wild_nocase("*OPER*", u->username)) {
	      if (conf.spcchannelactive) {
	         privmsg(s_NickServ, conf.spcchannel, "User connecting has a username matching or referencing a staff title or position -> (%s!%s@%s).", 
                  u->nick, u->username, u->host);
            }
         }
         if (match_wild_nocase("*CHANSERV*", u->realname) || match_wild_nocase("*MEMOSERV*", u->realname) ||
	       match_wild_nocase("*NICKSERV*", u->realname) || match_wild_nocase("*OPERSERV*", u->realname) ||
	       match_wild_nocase("*SERV*", u->realname) || match_wild_nocase("*HELPSERV*", u->realname)   ) {
	       if (conf.spcchannelactive) {
	          privmsg(s_NickServ, conf.spcchannel, "User connecting has a realname matching or referencing a services client -> (%s!%s@%s).", 
		       u->nick, u->username, u->host);
             }
         } else if (match_wild_nocase("*IRC*OP", u->realname) || match_wild_nocase("*ADMIN*", u->realname) ||
	              match_wild_nocase("*SERVICE*", u->realname) || match_wild_nocase("*STAFF*", u->realname) ||
		        match_wild_nocase("*OPER*", u->realname)) {
            if (conf.spcchannelactive) {
	         privmsg(s_NickServ, conf.spcchannel, "User connecting has a realname matching or referencing a staff title or position -> (%s!%s@%s).", 
		      u->nick, u->username, u->host);
            }
         }
      }
      /* Add user count to the server user count. */
      if (si = findserver(u->server)) {
         si->usrcnt++;
      }
#ifndef V_BAHAMUT
      if (conf.logonmsg1) {
         notice(s_GlobalNoticer, user->nick, "%s", conf.logonmsg1);
      }
      if (conf.logonmsg2) {
         notice(s_GlobalNoticer, user->nick, "%s", conf.logonmsg2);
      }
      if (conf.logonmsg3) {
         notice(s_GlobalNoticer, user->nick, "%s", conf.logonmsg3);
      }
#endif
   } else {
      /* This is an old user changing nicks.  Simply unset the reg mode, and change the nick in the structure. */
    	if (!(u = finduser(source))) {
         return;
	}
      cancel_user(u);

      /* If this is a registered nick set mode +r on the nick using SVSMODE. */
      if (!(ni = findnick(av[0]))) {
         send_cmd(NULL, "SVSMODE %s -r", av[0]);
      }
#ifdef ELF
	/* Remove the access flags for the services status on the nick. */
      send_cmd(NULL, "SVSMODE %s -xy", av[0]);
#endif
	change_user_nick(u, av[0]);
	u->signon = atol(av[1]);
   }
   u->my_signon = time(NULL);

   /* Check this nick's memos. */
   if (validate_user(u)) {
	check_memos(u->nick);
   }
}

/*************************************************************************/

/* do_join()
 * Handle a JOIN command.
 *	av[0] = channels to join */

void do_join(const char *source, int ac, char **av)
{
   User *u;
   char *s, *t;
   struct u_chanlist *c, *nextc;
   ChannelInfo *ci;

   if (!(u = finduser(source))) {
      return;
   }
   t = av[0];
   while (*(s=t)) {
      t = s + strcspn(s, ",");
	if (*t) {
	    *t++ = 0;
      }
	if (*s == '0') {
	   c = u->chans;
	   while (c) {
		nextc = c->next;
		chan_deluser(u, c->chan);
		free(c);
		c = nextc;
	   }
	   u->chans = NULL;
	   continue;
	}
	/* Make sure check_kick comes before chan_adduser, so banned users
	 * don't get to see things like channel keys. */
	if (check_kick(u, s)) {
	    continue;
      }
	/* Check this channel for new memos, or existing memos. */
      check_cmemos(u, s);

      if ((ci = cs_findchan(s)) && ci->join_message) {
	    notice(s_ChanServ, u->nick, "%s", ci->join_message);
      }
	chan_adduser(u, s);

	c = smalloc(sizeof(*c));
	c->next = u->chans;
	c->prev = NULL;
	if (u->chans) {
	    u->chans->prev = c;
      }
	u->chans = c;
	c->chan = findchan(s);
   }
}

/*************************************************************************/

void join_user_update(User *u, Channel *chan)
{
   struct u_chanlist *c;
   ChannelInfo *ci;

   if (!chan) {
      return;
   }
   /* Make sure check_kick comes before chan_adduser, so banned users
    * don't get to see things like channel keys. */

   if (check_kick(u, chan->name)) {
      return;
   }

   /* Check this channel for new memos, or existing memos. */
   check_cmemos(u, chan->name);

   if ((ci = cs_findchan(chan->name)) && ci->join_message) {
      notice(s_ChanServ, u->nick, "%s", ci->join_message);
   }
   chan_adduser2(u, chan);

   c = smalloc(sizeof(*c));
   c->next = u->chans;
   c->prev = NULL;
   if (u->chans) {
      u->chans->prev = c;
   }
   u->chans = c;
   c->chan = chan;
}

/*************************************************************************/

#ifdef V_BAHAMUT

/* Handle a SVINFO command. 
 * On server connect:
 *   av[0] = current
 *   av[1] = minimum
 *   av[2] = timestamp */

/* do_svinfo()
 * Server information handling. */

void do_svinfo(const char *source, int ac, char **av)
{
   hub.tscur = atoi(av[0]);
   hub.tsmin = atoi(av[1]);
   hub.svinfo = 1;
}

/*************************************************************************/

/* Handle a SJOIN command. 
 * On channel creation, syntax is:
 *   av[0] = timestamp
 *   av[1] = channel name
 *   av[2|3|4] = modes   \   depends whether the modes k and l
 *   av[3|4|5] = users   /   are set or not.
 * When a single user joins an (existing) channel, it is:
 *   av[0] = timestamp
 *   av[1] = user */

/* do_sjoin()
 * Prototype joining system for bahamut servers. */

void do_sjoin(const char *source, int ac, char **av)
{
	Channel *c;
	User *u;
	char *bmodes[ac-2], *cumodes[4];
	char *s, *end;
	
	/*
	if (run.debug == 1)
         
         globops(NULL, "Entering do_sjoin() ac=%d, av[0]=%s, av[1]=%s, av[2]=%s, av[3]=%s, av[4]=%s, av[5]=%s, av[6]=%s",
		ac, av[0], av[1], av[2], av[3], av[4], av[5], av[6]);
         
      }
	*/
	/* Double check to avoid unknown modes that need parameters */
	if ((ac >= 4) && (ac <= 7)) {
		if (!(c = findchan(av[2]))) { // 1
			/* It's a new channel, so we need to create it first. */
			c = chan_create(av[2]); // 1
		}
		
		/* We make all the users join */
		s = av[ac-1];	
		/* Users are always the last argument */
		while (*s) {
			int is_op = 0, is_voice = 0;
			int must_break = 0;
			
			end = strchr(s, ' ');
			if (end) *end = 0;
			else must_break = 1;
			
			while (*s == '@' || *s == '+') {
				if (*s == '@') is_op = 1;
				if (*s++ == '+') is_voice = 1;
			}
			
			u = finduser(s);
	    		if (!u) {
				log("user: SJOIN for nonexistent user %s on %s", s, av[1]);
				return;
    			}
			
			/* We make the user join */
			join_user_update(u, c);

			/* We update user mode on the channel */
			if (is_op || is_voice) {
				cumodes[0] = c->name;
				cumodes[1] = smalloc(8);
				sprintf(cumodes[1], "+%s%s", (is_op ? "o" : ""), (is_voice ? "v" : ""));
				cumodes[2] = u->nick;
				cumodes[3] = ((is_op && is_voice) ? u->nick : NULL);
				do_cmode(source, ((is_op && is_voice) ? 4 : 3), cumodes);
				free(cumodes[1]);
			}
			
			if (must_break) break;
			s = end+1;	
		}
		
		/* We now update the channel mode. */
		bmodes[0] = c->name;
		bmodes[1] = av[3]; // 2
		if (ac>=5) {
			bmodes[2] = av[4]; // 3
			if (ac==7) bmodes[3] = av[5]; // 4
		}
		do_cmode(source, ac-2, bmodes);
	} else if (ac == 2) {
		u = finduser(source);
    		if (!u) {
			log("user: SJOIN for nonexistent user %s on %s", source, av[1]);
			return;
    		}
		if (!(c = findchan(av[2]))) {
			/* This may be the case when a bot of BotServ is still on the channel
			   when the server sends the SJOIN. Grmbl :P */
			c = chan_create(av[2]);
		}
		join_user_update(u, c);
	}
}

#endif

/*************************************************************************/

/* do_part()
 * Handle a PART command.
 *	av[0] = channels to leave
 *	av[1] = reason (optional) */

void do_part(const char *source, int ac, char **av)
{
    User *u;
    char *s, *t;
    struct u_chanlist *c;
    ChannelInfo *ci;

    if (!(u = finduser(source))) {
       return;
    }
    t = av[0];
    while (*(s=t)) {
	t = s + strcspn(s, ",");
	if (*t)
	    *t++ = 0;
	for (c = u->chans; c && stricmp(s, c->chan->name) != 0; c = c->next)
	    ;
	if (c) {
	    chan_deluser(u, c->chan);
	    if ((ci = cs_findchan(s)) && ci->part_message) {
		 notice(s_ChanServ, u->nick, "%s", ci->part_message);
          }
	    if (c->next)
		c->next->prev = c->prev;
	    if (c->prev)
		c->prev->next = c->next;
	    else
		u->chans = c->next;
	    free(c);
	}
    }
}

/*************************************************************************/

/* do_kick()
 * Handle a KICK command.
 *	av[0] = channel
 *	av[1] = nick(s) being kicked
 *	av[2] = reason */

void do_kick(const char *source, int ac, char **av)
{
    User *u;
    char *s, *t;
    struct u_chanlist *c;

    t = av[1];
    while (*(s=t)) {
	t = s + strcspn(s, ",");
	if (*t)
	    *t++ = 0;
	if (!(u = finduser(s)))
         continue;
	for (c = u->chans; c && stricmp(av[0], c->chan->name) != 0; c = c->next)
	    ;
	if (c) {
	    chan_deluser(u, c->chan);
	    if (c->next)
		c->next->prev = c->prev;
	    if (c->prev)
		c->prev->next = c->next;
	    else
		u->chans = c->next;
	    free(c);
	}
    }
}

/*************************************************************************/

/* do_mode()
 * Handle incomming command: MODE
 * Parameters:
 *  av[0] = nick to change mode for
 *  av[1] = modes */

void do_umode(const char *source, int ac, char **av)
{
   User *u;
   NickInfo *ni;
   ServerInfo *si;
   char *s;
   int add = 1; /* Set to 1 if adding modes, set to 0 of deleting modes. */

   if (stricmp(source, av[0]) != 0) {
      globops(NULL, "\2%s\2 attempted to change user mode \2%s\2 for \2%s\2.",
         source, av[1], av[0]);
      return;
   }
   if (!(u = finduser(source))) {
      return;
   }
   s = av[1];
   while (*s) {
      switch (*s++) {
	   case '+': add = 1; break;
	   case '-': add = 0; break;
	   case 'i':
            /* +i and non +i users are calculated by themselves the sum of both actually
             * equal the entire network's users, thus when mode set -- and ++, mode unset
             * ++ and -- as such to keep counts the same - Drakan */
            if (add) {
               u->mode |= UMODE_I;
               lu_invisible++;
               --lu_users;
            } else {
               u->mode &= ~UMODE_I;
               lu_users++;
               --lu_invisible;
            }
            break;
	   case 'w': add ? (u->mode |= UMODE_W) : (u->mode &= ~UMODE_W); break;
	   case 'g': add ? (u->mode |= UMODE_G) : (u->mode &= ~UMODE_G); break;
	   case 's': add ? (u->mode |= UMODE_S) : (u->mode &= ~UMODE_S); break;
	   case 'f': add ? (u->mode |= UMODE_F) : (u->mode &= ~UMODE_F); break;
	   case 'k': add ? (u->mode |= UMODE_K) : (u->mode &= ~UMODE_K); break;
	   case 'a':
	      if (add) {
               u->mode |= UMODE_A;
#ifdef ELF
		   if ((ni = findnick(u->nick))) {
		      if (is_services_root(u->nick)) {
                     send_cmd(NULL, "SVSMODE %s +y", u->nick);
			}
                  if (is_services_op(u->nick))
                     send_cmd(NULL, "SVSMODE %s +x", u->nick);
                  }
#endif
            } else {
		   u->mode &= ~UMODE_A;
#ifdef ELF
		   send_cmd(NULL, "SVSMODE %s -xy", u->nick);
#endif
		}
            break;
	    case 'o':
             if (add) {
                /* Abra Y:Line support.  Abra can control and limit the number of IRC Operators
                 * set mode +o on the network at any given time.  Abra filters excessive IRC
                 * Operators by killing them off. */
		    if ((lu_opers + 1) > conf.yomax) {
		       kill_user(NULL, u->nick, rpl_numeric_lang(u->nick, CR_KILL_YLINE_OPER));
		       return;
		    }
		    /* Is this user set to no oper?  If so reverse the modes and notice the user. */
		    if (u->flags & U_NOOP) {
		       send_cmd(NULL, "SVSMODE %s -ogskf", u->nick);
		       notice(s_OperServ, u->nick, rpl_numeric_lang(u->nick, CR_NOOP_DEOP));
		       return;
		    }
		    /* Same thing as above but for an entire server, notice the user as above. */
                if ((si = findserver(u->server)) && (si->flags & SI_NOOP)) {
		       send_cmd(NULL, "SVSMODE %s -ogskf", u->nick);
		       notice(s_OperServ, u->nick, rpl_numeric_lang(u->nick, CR_NOOP_DEOP));
		       return;
		    }
		    u->mode |= UMODE_O;
 	          if (conf.opermsg1) {
                   notice(s_GlobalNoticer, u->nick, "%s", conf.opermsg1);
	 	    }
                if (conf.opermsg2) {
                   notice(s_GlobalNoticer, u->nick, "%s", conf.opermsg2);
		    }
                if (conf.opermsg3) {
                   notice(s_GlobalNoticer, u->nick, "%s", conf.opermsg3);
                }
		    ++lu_opers;
                /* Send globops announcing this successful oper attempt. */
		    globops(NULL, "\2%s\2 (%s!%s@%s) is now an IRC Operator.", u->nick, u->nick, u->username, u->host);

                /* Check for the mandatory setting of the IRC Operator nick flag, if set ignore this
                 * otherwise send notice to the IRC Operator to set it. */
		    if ((ni = findnick(u->nick)) && !(ni->flags & NI_IRCOP)) {
                   notice(s_NickServ, u->nick, rpl_numeric_lang(u->nick, NS_NOT_OPER_ON_USER));
                   notice(s_NickServ, u->nick, rpl_numeric_lang(u->nick, NS_PLEASE_SET_IRCOP), s_NickServ);
		    }
		} else {
               u->mode &= ~UMODE_O;
#ifdef ELF
		   send_cmd(NULL, "SVSMODE %s -xy", u->nick);
#endif
		   --lu_opers;
		}
		break;
      }
   }
}

/*************************************************************************/

/* do_quit()
 * Handle a incomming command: QUIT
 * Paramters:
 *  av[0] = quit reason */

void do_quit(const char *source, int ac, char **av)
{
   User *u;
   NickInfo *ni;
   ServerInfo *si;

   if (!(u = finduser(source))) {
      return;
   }
   /* Update the last seen time just before they quit, since this would technically
    * be the last time we saw this user online. */
   if ((ni = u->ni) && (!(ni->flags & NI_VERBOTEN)) &&
       (ni->flags & (NI_IDENTIFIED | NI_RECOGNIZED))) {
      ni->last_seen = time(NULL);
   }
   delete_user(u);
}

/*************************************************************************/

/* do_kill()
 * Handle incomming command: KILL
 * Paramters:
 *  av[0] = target nick being killed
 *  av[1] = kill reason */

void do_kill(const char *source, int ac, char **av)
{
   User *u;
   NickInfo *ni;

   if (!(u = finduser(av[0]))) {
      return;
   }
   /* Update the last seen time just before they are dropped off the network, since this
    * would technically be the last time we saw this user online. */
   if ((ni = u->ni) && (!(ni->flags & NI_VERBOTEN)) &&
       (ni->flags & (NI_IDENTIFIED | NI_RECOGNIZED))) {
      ni->last_seen = time(NULL);
   }
   delete_user(u);
}

/*************************************************************************/

/* is_oper()
 * Is the given nick an IRC Operator (mode +o)? */

int is_oper(const char *nick)
{
   User *u = finduser(nick);

   if (u && (u->mode & UMODE_O)) {
      return C_TRUE;
   }
   return C_FALSE;
   /* v4 -> return u && (u->mode & UMODE_O); */
}

/*************************************************************************/

/* is_sadmin()
 * Is the given nick a Services Administrator (mode +a)? */

int is_sadmin(const char *nick)
{
   User *u = finduser(nick);

   if (u && (u->mode & UMODE_A)) {
      return C_TRUE;
   }
   return C_FALSE;
   /* v4 -> return u && (u->mode & UMODE_A); */
}

/*************************************************************************/

/* is_invisible()
 * Is the given nick a invisible (mode +i)? */

int is_invisible(const char *nick)
{
   User *u = finduser(nick);

   if (u && (u->mode & UMODE_I)) {
      return C_TRUE;
   }
   return C_FALSE;
}

/*************************************************************************/

/* is_on_chan()
 * Is the given nick on the given channel? */

int is_on_chan(const char *nick, const char *chan)
{
   User *u = finduser(nick);
   struct u_chanlist *c;

   if (!u) {
      return C_FALSE;
   }
   for (c = u->chans; c; c = c->next) {
      if (stricmp(c->chan->name, chan) == 0) {
         return C_TRUE;
      }
   }
   return C_FALSE;
}

/*************************************************************************/

/* is_chanop()
 * Is the given nick a channel operator on the given channel? */

int is_chanop(const char *nick, const char *chan)
{
   Channel *c = findchan(chan);
   struct c_userlist *u;

   if (!c) {
      return C_FALSE;
   }
   for (u = c->chanops; u; u = u->next) {
      if (stricmp(u->user->nick, nick) == 0) {
         return C_TRUE;
      }
   }
   return C_FALSE;
}

/*************************************************************************/

#ifdef V_UNREAL

/* is_halfop()
 * Is the given nick a half channel operator on the given channel? */

int is_halfop(const char *nick, const char *chan)
{
   Channel *c = findchan(chan);
   struct c_userlist *u;

   if (!c) {
      return C_FALSE;
   }
   for (u = c->halfops; u; u = u->next) {
      if (stricmp(u->user->nick, nick) == 0) {
         return C_TRUE;
      }
   }
   return C_FALSE;
}

#endif

/*************************************************************************/

/* is_voiced()
 * Is the given nick voiced (channel mode +v) on the given channel? */

int is_voiced(const char *nick, const char *chan)
{
   Channel *c = findchan(chan);
   struct c_userlist *u;

   if (!c) {
      return C_FALSE;
   }
   for (u = c->voices; u; u = u->next) {
      if (stricmp(u->user->nick, nick) == 0) {
         return C_TRUE;
      }
   }
   return C_FALSE;
}

/*************************************************************************/

/* match_usermask()
 * Does the user's usermask match the given mask (either nick!user@host or
 * just user@host)? */

int match_usermask(const char *mask, User *u)
{
    char *mask2 = sstrdup(mask);
    char *nick, *username, *host, *nick2, *host2;
    int result;

    if (strchr(mask2, '!')) {
	nick = strlower(strtok(mask2, "!"));
	username = strtok(NULL, "@");
    } else {
	nick = NULL;
	username = strtok(mask2, "@");
    }
    host = strtok(NULL, "");
    if (!username || !host) {
	free(mask2);
	return C_FALSE;
    }
    strlower(host);
    if (nick)
	nick2 = strlower(sstrdup(u->nick));
    host2 = strlower(sstrdup(u->host));
    result = (nick ? match_wild(nick, nick2) : 1) &&
		match_wild(username, u->username) &&
		match_wild(host, host2);
    free(mask2);
    if (nick)
	free(nick2);
    free(host2);
    return result;
}

/*************************************************************************/

/* split_usermask()
 * Split a usermask up into its constitutent parts.  Returned strings are
 * malloc()'d, and should be free()'d when done with.  Returns "*" for
 * missing parts. */

void split_usermask(const char *mask, char **nick, char **user, char **host)
{
    char *mask2 = sstrdup(mask);

    *nick = strtok(mask2, "!");
    *user = strtok(NULL, "@");
    *host = strtok(NULL, "");
    /* Handle special case: mask == user@host */
    if (*nick && !*user && strchr(*nick, '@')) {
	*nick = NULL;
	*user = strtok(mask2, "@");
	*host = strtok(NULL, "");
    }
    if (!*nick)
	*nick = "*";
    if (!*user)
	*user = "*";
    if (!*host)
	*host = "*";
    *nick = sstrdup(*nick);
    *user = sstrdup(*user);
    *host = sstrdup(*host);
    free(mask2);
}

/*************************************************************************/

/* create_mask()
 * Given a user, return a mask that will most likely match any address the
 * user will have from that location.  For IP addresses, wildcards the
 * appropriate subnet mask (e.g. 35.1.1.1 -> 35.*; 128.2.1.1 -> 128.2.*);
 * for named addresses, wildcards the leftmost part of the name unless the
 * name only contains two parts.  If the username begins with a ~, delete
 * it.  The returned character string is malloc'd and should be free'd
 * when done with. */

char *create_mask(User *u)
{
    char *mask, *s, *end;

    /* Get us a buffer the size of the username plus hostname.  The result
     * will never be longer than this (and will often be shorter). */
    end = mask = smalloc(strlen(u->username) + strlen(u->host) + 2);
    end += sprintf(end, "%s@", u->username);
    if (strspn(u->host, "0123456789.") == strlen(u->host)
		&& (s = strchr(u->host, '.'))
		&& (s = strchr(s+1, '.'))
		&& (s = strchr(s+1, '.'))
		&& (   !strchr(s+1, '.'))) {	/* IP addr */
	s = sstrdup(u->host);
	*strrchr(s, '.') = 0;
	if (atoi(u->host) < 192)
	    *strrchr(s, '.') = 0;
	if (atoi(u->host) < 128)
	    *strrchr(s, '.') = 0;
	sprintf(end, "%s.*", s);
	free(s);
    } else {
	if ((s = strchr(u->host, '.')) && strchr(s+1, '.')) {
	    s = sstrdup(strchr(u->host, '.')-1);
	    *s = '*';
	} else {
	    s = sstrdup(u->host);
	}
	strcpy(end, s);
	free(s);
    }
    return mask;
}

/*************************************************************************/

#endif /* __users_source__ */

/*************************************************************************/
